/*
** Copyright (C) 2001-2006 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwfilterio.c
**
** Suresh L Konda
**      routines to do io stuff with filter records.
*/

#include "silk.h"

RCSIDENT("$SiLK: rwfilterio.c 5883 2006-12-08 21:54:58Z mthomas $");

#include "librw_priv.h"


/* Version to use when SK_FILE_VERSION_ANY is specified */
#define DEFAULT_FILE_VERSION 5

/*
 * Version where data compression is enabled.  We support full
 * compression as of this version.  For files one version below this,
 * we only support reading compressed files---files created with
 * SiLK-0.10.0 through SiLK-0.10.4
 */
#define COMPMETHOD_SUPPORTED 5


/* ********************************************************************* */

/*
**  RWFILTER VERSION 1
**
**    ipUnion       sIP;             //  0- 3  Source IP
**    ipUnion       dIP;             //  4- 7  Destination IP
**
**    uint16_t      sPort;           //  8- 9  Source port
**    uint16_t      dPort;           // 10-11  Destination port
**
**    uint8_t       proto;           // 12     IP protocol
**    uint8_t       flags;           // 13     OR of all TCP flags on all pkts
**    uint8_t       input;           // 14     Router incoming SNMP interface
**    uint8_t       output;          // 15     Router outgoing SNMP interface
**
**    ipUnion       nhIP;            // 16-19  Router Next Hop IP
**    uint32_t      sTime;           // 20-23  Start time of flow-epoch secs
**
**    uint32_t      pef;             // 24-27
**    // uint32_t     pkts      :20; //        Count of packets
**    // uint32_t     elapsed   :11; //        Duration of flow
**    // uint32_t     pflag     : 1; //        'pkts' requires multiplier?
**
**    uint32_t      bbs;             // 28-31
**    // uint32_t     bPPkt     :14; //        Whole bytes-per-packet
**    // uint32_t     bPPFrac   : 6; //        Fractional bytes-per-packet
**    // uint32_t     sensorID  : 6; //        Sensor ID
**    // uint32_t     pad       : 6; //        Padding
**
**
**  32 bytes on disk.
*/

#define RECLEN_RWFILTER_V1  32


/*
 *  Unpack the array of bytes 'ar' into a record 'rwRP'
 */
static int _filterioRecordUnpack_V1V2(
    rwIOStruct_t       *rwIOS,
    rwGenericRec_V3    *rwRP,
    const uint8_t      *ar)
{
    uint32_t pef, sbb;
    uint32_t bpp, pkts, pflag;

    /* sIP, dIP, sPort, dPort */
    memcpy(&rwRP->sIP,     &ar[0],  4);
    memcpy(&rwRP->dIP,     &ar[4],  4);
    memcpy(&rwRP->sPort,   &ar[8],  2);
    memcpy(&rwRP->dPort,   &ar[10], 2);

    /* proto, flags, input, output */
    rwRP->proto          =  ar[12];
    rwRP->flags          =  ar[13];
    rwRP->input          =  ar[14];
    rwRP->output         =  ar[15];

    /* nhIP */
    memcpy(&rwRP->nhIP,    &ar[16], 4);

    /* sTime */
    memcpy(&rwRP->sTime,   &ar[20], 4);

    /* pef: uint32_t pkts:20; uint32_t elapsed :11; uint32_t pflag:1; */
    memcpy(&pef,           &ar[24], 4);
    pkts = pef >> 12;
    rwRP->elapsed = (pef >> 1) & MASKARRAY_11;
    pflag = pef & MASKARRAY_01;

    /* bytes, bpp (ignoring sensor in last 12 bits of sbb) */
    memcpy(&sbb,           &ar[28], 4);
    bpp = (sbb >> 12) & MASKARRAY_20;

    _packUnpackBytesPackets(rwRP, bpp, pkts, pflag);

    if (rwGetFileVersion(rwIOS) == 1) {
        /* handle sensor: sbb is (bPPkt:14; bPPFrac:6; sID:6; pad:6) */
        rwRP->sID = (uint8_t)((sbb >> 6) & MASKARRAY_06);
    } else {
        assert(rwGetFileVersion(rwIOS) == 2);
        /* handle sensor: sbb is (bPPkt:14; bPPFrac:6; pad: 4; sID:8) */
        rwRP->sID = (sensorID_t)(sbb & MASKARRAY_08);
    }

    /* everything else not in this format */
    rwRP->flow_type = SK_INVALID_FLOWTYPE;

    return LIBRW_OK;
}


/*
 *  Pack the record 'rwRP' into an array of bytes 'ar'
 */
static int _filterioRecordPack_V1V2(
    rwIOStruct_t           *rwIOS,
    const rwGenericRec_V3  *rwRP,
    uint8_t                *ar)
{
    int rv = LIBRW_OK; /* return value */
    uint32_t bbs;
    uint32_t pef;
    uint32_t pflag;

    /* Check sizes of fields we've expanded in later versions */
    if (rwRP->input > 255 || rwRP->output > 255) {
        rv = LIBRW_ERR_SNMP_OVRFLO;
        goto END;
    }
    if (rwRP->sID > 255) {
        rv = LIBRW_ERR_SENSORID_OVRFLO;
        goto END;
    }
    if (rwRP->elapsed >= MAX_ELAPSED_TIME_OLD) {
        rv = LIBRW_ERR_ELPSD_OVRFLO;
        goto END;
    }

    /* bytes-per-packet, packets, packets-flag */
    rv = _packPackBytesPackets(&bbs, &pef, &pflag, rwRP);
    if (rv) { goto END; }

    /* The bbs value we have has the bytes-per-packet in the least
     * significant bits.  Move it to the most significant bits, and
     * add the sensor id. */
    /* bbs: uint32_t bPPkt:14;  uint32_t bPPFrac:6; pad:4; sID:8 */
    bbs = ((bbs << 12) | (rwRP->sID & MASKARRAY_08));

    if (rwGetFileVersion(rwIOS) == 1) {
        /* For v1 of FT_RWFILTER files, only 6 bits allowed of sensor
         * is allowed.  Check that our value fits */
        if (rwRP->sID > MASKARRAY_06) {
            rv = LIBRW_ERR_SENSORID_OVRFLO;
            goto END;
        }

        /* To convert from version 2 to version 1, shrink the sID to six
         * bits and shift it six bits to the left. */
        bbs = ((bbs & 0xFFFFF000) /* top 20 bits are the same */
               | ((bbs & MASKARRAY_06) << 6));
    }

    /* The pef value we have has the packets value in the least
     * significant bits.  Move it to the most significant bits and add
     * the elapsed time and pflag. */
    /* pef: uint32_t pkts:20; uint32_t elapsed:11; uint32_t pflag:1; */
    pef = ((pef << 12) | (rwRP->elapsed << 1) | pflag);

    /* sIP, dIP, sPort, dPort */
    memcpy(&ar[0],  &rwRP->sIP,   4);
    memcpy(&ar[4],  &rwRP->dIP,   4);
    memcpy(&ar[8],  &rwRP->sPort, 2);
    memcpy(&ar[10], &rwRP->dPort, 2);

    /* proto, flags, input, output */
    ar[12] = rwRP->proto;
    ar[13] = rwRP->flags;
    ar[14] = (uint8_t)rwRP->input;
    ar[15] = (uint8_t)rwRP->output;

    /* nhIP, sTime, pef, bbs  */
    memcpy(&ar[16], &rwRP->nhIP,  4);
    memcpy(&ar[20], &rwRP->sTime, 4);
    memcpy(&ar[24], &pef,         4);
    memcpy(&ar[28], &bbs,         4);

  END:
    return rv;
}


/*
 * _filterioRecordSwap_V1(dataPtr);
 *
 * Treat dataPtr as an RWFILTER_V{1,2} record and byte-swap it in place.
 */
static void _filterioRecordSwap_V1(uint8_t *dataPtr)
{
    SWAP_DATA32(dataPtr +  0);   /* sIP */
    SWAP_DATA32(dataPtr +  4);   /* dIP */
    SWAP_DATA16(dataPtr +  8);   /* sPort */
    SWAP_DATA16(dataPtr + 10);   /* dPort */
    /* Four single bytes: (12)proto, (13)flags, (14)input, (15)output */
    SWAP_DATA32(dataPtr + 16);   /* nhIP */
    SWAP_DATA32(dataPtr + 20);   /* sTime */
    SWAP_DATA32(dataPtr + 24);   /* pef */
    SWAP_DATA32(dataPtr + 28);   /* bpp/sensorId */
}


/* ********************************************************************* */

/*
**  RWFILTER VERSION 2
**
**    Only the final 12 bits of the final four bytes differ from
**    VERSION 1: the sID is moved all the way to the right and is now
**    8 bits wide, with 4 bits of padding.
**
**    ipUnion       sIP;             //  0- 3  Source IP
**    ipUnion       dIP;             //  4- 7  Destination IP
**
**    uint16_t      sPort;           //  8- 9  Source port
**    uint16_t      dPort;           // 10-11  Destination port
**
**    uint8_t       proto;           // 12     IP protocol
**    uint8_t       flags;           // 13     OR of all TCP flags on all pkts
**    uint8_t       input;           // 14     Router incoming SNMP interface
**    uint8_t       output;          // 15     Router outgoing SNMP interface
**
**    ipUnion       nhIP;            // 16-19  Router Next Hop IP
**    uint32_t      sTime;           // 20-23  Start time of flow-epoch secs
**
**    uint32_t      pef;             // 24-27
**    // uint32_t     pkts      :20; //        Count of packets
**    // uint32_t     elapsed   :11; //        Duration of flow
**    // uint32_t     pflag     : 1; //        'pkts' requires multiplier?
**
**    uint32_t      bbs;             // 28-31
**    // uint32_t     bPPkt     :14; //        Whole bytes-per-packet
**    // uint32_t     bPPFrac   : 6; //        Fractional bytes-per-packet
**    // uint32_t     pad       : 4; //        Padding
**    // uint32_t     sensorID  : 8; //        Sensor ID
**
**
**  32 bytes on disk.
*/

#define RECLEN_RWFILTER_V2  32


/* ********************************************************************* */

/*
**  RWFILTER VERSION 3
**
**    ipUnion       sIP;             //  0- 3  Source IP
**    ipUnion       dIP;             //  4- 7  Destination IP
**
**    uint16_t      sPort;           //  8- 9  Source port
**    uint16_t      dPort;           // 10-11  Destination port
**
**    ipUnion       nhIP;            // 12-15  Router Next Hop IP
**    uint16_t      input;           // 16-17  Router incoming SNMP interface
**    uint16_t      output;          // 18-19  Router outgoing SNMP interface
**
**    uint32_t      sTime;           // 20-23  Start time of flow-epoch secs
**    uint32_t      elapsed;         // 24-27  Duration of flow
**
**    uint32_t      pkts_ftype       // 28-31
**    // uint32_t     pkts      :20; //        Count of packets
**    // uint32_t     pflag     : 1; //        'pkts' requires multiplier?
**    // uint32_t     pad       : 3; //        padding/reserved
**    // uint32_t     flowtype  : 8; //        flow type (class&type)
**
**    uint32_t      bpp;             // 32-35
**    // uint32_t     bPPkt     :14; //        Whole bytes-per-packet
**    // uint32_t     bPPFrac   : 6; //        Fractional bytes-per-packet
**    // uint32_t     pad       :12; //        padding/reserved
**
**    sensorID_t    sensorID;        // 36-37  Sensor ID
**
**    uint8_t       proto;           // 38     IP protocol
**    uint8_t       flags;           // 39     OR of all TCP flags on all pkts
**
**
**  40 bytes on disk.
*/

#define RECLEN_RWFILTER_V3  40


/*
 *  Unpack the array of bytes 'ar' into a record 'rwRP'
 */
static int _filterioRecordUnpack_V3(
    rwIOStruct_t            UNUSED(*rwIOS),
    rwGenericRec_V3        *rwRP,
    const uint8_t          *ar)
{
    uint32_t pkts_ftype, bpp;
    uint32_t pkts, pflag;

    /* sIP, dIP, sPort, dPort, nhIP, input, output, sTime, elapsed */
#if 1
    memcpy(&rwRP->sIP,     &ar[0],  28);
#else
    memcpy(&rwRP->sIP,     &ar[0],   4);
    memcpy(&rwRP->dIP,     &ar[4],   4);
    memcpy(&rwRP->sPort,   &ar[8],   2);
    memcpy(&rwRP->dPort,   &ar[10],  2);
    memcpy(&rwRP->nhIP,    &ar[12],  4);
    memcpy(&rwRP->input,   &ar[16],  2);
    memcpy(&rwRP->output,  &ar[18],  2);
    memcpy(&rwRP->sTime,   &ar[20],  4);
    memcpy(&rwRP->elapsed, &ar[24],  4);
#endif

    /* pkts_ftype, bpp, sensorID */
    memcpy(&pkts_ftype,    &ar[28],  4);
    memcpy(&bpp,           &ar[32],  4);
    memcpy(&rwRP->sID,     &ar[36],  2);

    /* protocol, tcp-flags */
    rwRP->proto           = ar[38];
    rwRP->flags           = ar[39];

    /* unpack 'pkts_ftype': pkts:20; pflag:1; pad:3; flowtype:8; */
    pkts = GET_MASKED_BITS(pkts_ftype, 12, 20);
    pflag = GET_MASKED_BITS(pkts_ftype, 11, 1);
    rwRP->flow_type = GET_MASKED_BITS(pkts_ftype, 0, 8);

    /* 'bpp' has bytes-per-packet in most significant bits; move to
     * least significant bits which is where the unpack function below
     * expects them. */
    bpp = bpp >> 12;

    /* pkts, bytes, bpp */
    _packUnpackBytesPackets(rwRP, bpp, pkts, pflag);

    return LIBRW_OK;
}


/*
 *  Pack the record 'rwRP' into an array of bytes 'ar'
 */
static int _filterioRecordPack_V3(
    rwIOStruct_t            UNUSED(*rwIOS),
    const rwGenericRec_V3  *rwRP,
    uint8_t                *ar)
{
    int rv = LIBRW_OK; /* return value */
    uint32_t pkts_ftype, bpp;
    uint32_t pflag = 0;

    rv = _packPackBytesPackets(&bpp, &pkts_ftype, &pflag, rwRP);
    if (rv) {
        return rv;
    }

    /* pkts_ftype has 'pkts' in the least significant bits; move it
     * over and add the 'pflag' and flow type */
    /* pkts_ftype: pkts:20, pflag:1; pad:3; flowType:8 */
    pkts_ftype = ((pkts_ftype << 12)
                  | ((pflag == 0) ? 0 : (1 << 11))
                  | rwRP->flow_type);

    /* 'bpp' has bytes-per-pkt in the least significant bits; move it
     * over to the most significant bits */
    bpp = (bpp << 12);

    /* sIP, dIP, sPort, dPort, nhIP, input, output, sTime */
#if 1
    memcpy(&ar[0],  &rwRP->sIP,     28);
#else
    memcpy(&ar[0],  &rwRP->sIP,      4);
    memcpy(&ar[4],  &rwRP->dIP,      4);
    memcpy(&ar[8],  &rwRP->sPort,    2);
    memcpy(&ar[10], &rwRP->dPort,    2);
    memcpy(&ar[12], &rwRP->nhIP,     4);
    memcpy(&ar[16], &rwRP->input,    2);
    memcpy(&ar[18], &rwRP->output,   2);
    memcpy(&ar[20], &rwRP->sTime,    4);
    memcpy(&ar[24], &rwRP->elapased, 4);
#endif

    /* pkts_ftype, bpp, sID, protocol, tcp-flags */
    memcpy(&ar[28], &pkts_ftype,     4);
    memcpy(&ar[32], &bpp,            4);
    memcpy(&ar[36], &rwRP->sID,      2);
    ar[38]         = rwRP->proto;
    ar[39]         = rwRP->flags;

    return LIBRW_OK;
}


/*
 * _filterioRecordSwap_V3(dataPtr);
 *
 * Treat dataPtr as an RWFILTER_V3 record and byte-swap it in place.
 */
static void _filterioRecordSwap_V3(uint8_t *dataPtr)
{
    SWAP_DATA32(dataPtr +  0);   /* sIP */
    SWAP_DATA32(dataPtr +  4);   /* dIP */
    SWAP_DATA16(dataPtr +  8);   /* sPort */
    SWAP_DATA16(dataPtr + 10);   /* dPort */
    SWAP_DATA32(dataPtr + 12);   /* nhIP */
    SWAP_DATA16(dataPtr + 16);   /* input */
    SWAP_DATA16(dataPtr + 18);   /* output */
    SWAP_DATA32(dataPtr + 20);   /* sTime */
    SWAP_DATA32(dataPtr + 24);   /* elapsed */
    SWAP_DATA32(dataPtr + 28);   /* pkts_ftype */
    SWAP_DATA32(dataPtr + 32);   /* bpp */
    SWAP_DATA16(dataPtr + 36);   /* sID */
    /* Two single bytes: (38)proto, (39)flags */
}


/* ********************************************************************* */

/*
**  RWFILTER VERSION 4
**  RWFILTER VERSION 5
**
**    ipUnion       sIP;             //  0- 3  Source IP
**    ipUnion       dIP;             //  4- 7  Destination IP
**
**    uint16_t      sPort;           //  8- 9  Source port
**    uint16_t      dPort;           // 10-11  Destination port
**
**    ipUnion       nhIP;            // 12-15  Router Next Hop IP
**    uint16_t      input;           // 16-17  Router incoming SNMP interface
**    uint16_t      output;          // 18-19  Router outgoing SNMP interface
**
**    uint32_t      sTime;           // 20-23  Start time of flow-epoch secs
**    uint32_t      elapsed;         // 24-27  Duration of flow
**
**    uint32_t      pkts_stimems;    // 28-31
**    // uint32_t     pkts      :20; //        Count of packets
**    // uint32_t     pflag     : 1; //        'pkts' requires multiplier?
**    // uint32_t     is_tcp    : 1; //        1 if flow is TCP; 0 otherwise
**    // uint32_t     sTime_msec:10; //        Fractional sTime (millisec)
**
**    uint32_t      bb_elapsems;      // 32-35
**    // uint32_t     bPPkt     :14; //        Whole bytes-per-packet
**    // uint32_t     bPPFrac   : 6; //        Fractional bytes-per-packet
**    // uint32_t     padding   : 2; //        padding/reserved
**    // uint32_t     elaps_msec:10; //        Fractional elapsed (millisec)
**
**    sensorID_t    sensorID;        // 36-37  Sensor ID
**
**    uint8_t       flowtype;        // 38     flow type (class&type)
**    uint8_t       prot_flags;      // 39     is_tcp==0: IP protocol
**                                   //        is_tcp==1 &&
**                                   //          tcp_state==0:TCPflags/all pkts
**                                   //          tcp_state==1:TCPflags/1st pkt
**
**    uint16_t      application;     // 40-41  Generator of traffic
**
**    uint8_t       tcp_state;       // 42     TCP state machine info
**    uint8_t       rest_flags;      // 43     is_tcp==0: Flow's reported flags
**                                   //        is_tcp==1 &&
**                                   //          tcp_state==0:Empty
**                                   //          tcp_state==1:TCPflags/!1st pkt
**
**
**  44 bytes on disk.
*/

#define RECLEN_RWFILTER_V4  44
#define RECLEN_RWFILTER_V5  44


/*
 *  Unpack the array of bytes 'ar' into a record 'rwRP'
 */
static int _filterioRecordUnpack_V4(
    rwIOStruct_t            UNUSED(*rwIOS),
    rwGenericRec_V3        *rwRP,
    const uint8_t          *ar)
{
    uint32_t pkts_stimems, bb_elapsems;
    uint32_t pkts, bpp, pflag;
    uint8_t is_tcp;

    /* sIP, dIP, sPort, dPort, nhIP, input, output, sTime, elapsed */
#if 1
    memcpy(&rwRP->sIP,         &ar[ 0], 28);
#else
    memcpy(&rwRP->sIP,         &ar[ 0],  4);
    memcpy(&rwRP->dIP,         &ar[ 4],  4);
    memcpy(&rwRP->sPort,       &ar[ 8],  2);
    memcpy(&rwRP->dPort,       &ar[10],  2);
    memcpy(&rwRP->nhIP,        &ar[12],  4);
    memcpy(&rwRP->input,       &ar[16],  2);
    memcpy(&rwRP->output,      &ar[18],  2);
    memcpy(&rwRP->sTime,       &ar[20],  4);
    memcpy(&rwRP->elapsed,     &ar[24],  4);
#endif

    /* pkts_stimems, bb_elapsems, sensorID */
    memcpy(&pkts_stimems,      &ar[28],  4);
    memcpy(&bb_elapsems,       &ar[32],  4);
    memcpy(&rwRP->sID,         &ar[36],  2);

    /* flow type, application */
    rwRP->flow_type           = ar[38];
    memcpy(&rwRP->application, &ar[40],  2);

    /* unpack 'pkts_stimems': pkts:20; pflag:1; pad:1; sTime_msec:10; */
    pkts = GET_MASKED_BITS(pkts_stimems, 12, 20);
    pflag = GET_MASKED_BITS(pkts_stimems, 11, 1);
    is_tcp = (uint8_t)GET_MASKED_BITS(pkts_stimems, 10, 1);
    rwRP->sTime_msec = GET_MASKED_BITS(pkts_stimems, 0, 10);

    /* protocol, tcp-flags */
    _packUnpackProtoFlags(rwRP, is_tcp, ar[39], ar[42], ar[43]);

    /* unpack 'bb_elapsems': bpp:20 (bPPkt:14; bPPFrac:6); pad:2;
     * elapsed_msec:10; */
    bpp = GET_MASKED_BITS(bb_elapsems, 12, 20);
    rwRP->elapsed_msec = GET_MASKED_BITS(bb_elapsems, 0, 10);

    /* pkts, bytes, bpp */
    _packUnpackBytesPackets(rwRP, bpp, pkts, pflag);

    return LIBRW_OK;
}


/*
 *  Pack the record 'rwRP' into an array of bytes 'ar'
 */
static int _filterioRecordPack_V4(
    rwIOStruct_t            UNUSED(*rwIOS),
    const rwGenericRec_V3  *rwRP,
    uint8_t                *ar)
{
    int rv = LIBRW_OK; /* return value */
    uint32_t pkts_stimems, bb_elapsems;
    uint32_t pflag = 0;
    uint8_t is_tcp;

    rv = _packPackBytesPackets(&bb_elapsems, &pkts_stimems, &pflag, rwRP);
    if (rv) {
        return rv;
    }

    /* protocol, tcp-flags */
    _packPackProtoFlags(&is_tcp, &ar[39], &ar[42], &ar[43], rwRP);

    /* pkts_stimems has 'pkts' in the least significant bits---move it
     * over and insert the remaining values. */
    /* pkts_stimems: pkts:20, pflag:1; is_tcp:1; sTime_msec:10 */
    pkts_stimems = ((pkts_stimems << 12)
                    | ((pflag == 0) ? 0 : (1 << 11))
                    | ((is_tcp == 0) ? 0 : (1 << 10))
                    | (rwRP->sTime_msec & MASKARRAY_10));

    /* bb_elapsems has 'bpp' in the least significant bits; move it
     * over and add elapsed_msec */
    bb_elapsems = ((bb_elapsems << 12)
                   | (rwRP->elapsed_msec & MASKARRAY_10));

    /* sIP, dIP, sPort, dPort, nhIP, input, output, sTime */
#if 1
    memcpy(&ar[ 0], &rwRP->sIP,         28);
#else
    memcpy(&ar[ 0], &rwRP->sIP,          4);
    memcpy(&ar[ 4], &rwRP->dIP,          4);
    memcpy(&ar[ 8], &rwRP->sPort,        2);
    memcpy(&ar[10], &rwRP->dPort,        2);
    memcpy(&ar[12], &rwRP->nhIP,         4);
    memcpy(&ar[16], &rwRP->input,        2);
    memcpy(&ar[18], &rwRP->output,       2);
    memcpy(&ar[20], &rwRP->sTime,        4);
    memcpy(&ar[24], &rwRP->elapased,     4);
#endif

    /* pkts_stimems, bb_elapsems, sID, flow-type, application */
    memcpy(&ar[28], &pkts_stimems,       4);
    memcpy(&ar[32], &bb_elapsems,        4);
    memcpy(&ar[36], &rwRP->sID,          2);

    /* flow type, application */
    ar[38]         = rwRP->flow_type;
    memcpy(&ar[40], &rwRP->application,  2);

    return LIBRW_OK;
}


/*
 * _filterioRecordSwap_V4(dataPtr);
 *
 * Treat dataPtr as an RWFILTER_V4 record and byte-swap it in place.
 */
static void _filterioRecordSwap_V4(uint8_t *dataPtr)
{
    SWAP_DATA32(dataPtr +  0);   /* sIP */
    SWAP_DATA32(dataPtr +  4);   /* dIP */
    SWAP_DATA16(dataPtr +  8);   /* sPort */
    SWAP_DATA16(dataPtr + 10);   /* dPort */
    SWAP_DATA32(dataPtr + 12);   /* nhIP */
    SWAP_DATA16(dataPtr + 16);   /* input */
    SWAP_DATA16(dataPtr + 18);   /* output */
    SWAP_DATA32(dataPtr + 20);   /* sTime */
    SWAP_DATA32(dataPtr + 24);   /* elapsed */
    SWAP_DATA32(dataPtr + 28);   /* pkts_stimems */
    SWAP_DATA32(dataPtr + 32);   /* bb_elapsems */
    SWAP_DATA16(dataPtr + 36);   /* sID */
    /* Two single bytes: (38)flowtype, (39)prot_flags */
    SWAP_DATA16(dataPtr + 40);   /* application */
    /* Two single bytes: (42)tcp_state, (43)rest_flags */
}


/* ********************************************************************* */

/*
 *  Return length of record of specified version, or 0 if no such
 *  version exists.  See librw_priv.h for details.
 */
uint16_t _filterioGetRecLen(fileVersion_t vers)
{
    switch (vers) {
      case 1:
        return RECLEN_RWFILTER_V1;
      case 2:
        return RECLEN_RWFILTER_V2;
      case 3:
        return RECLEN_RWFILTER_V3;
      case 4:
        return RECLEN_RWFILTER_V4;
      case 5:
        return RECLEN_RWFILTER_V5;
      default:
        return 0;
    }
}


/*
 *  status = _filterioSetFuncs(&rwIOS);
 *
 *  Checks that the file's version is allowed by this type, and sets
 *  the record length and the read and write functions for this type.
 */
static int _filterioSetFuncs(rwIOStruct *rwIOS)
{
    int rv = LIBRW_OK; /* return value */

    assert(rwGetFileType(rwIOS) == FT_RWFILTER);

    /* version check; set values based on version */
    switch (rwGetFileVersion(rwIOS)) {
      case 0:
        /*
         * because of the change in rwRec, we can no longer read
         * version 0 files without major hacking
         */
        rv = LIBRW_ERR_UNSUPPORT_VERSION;
        goto END;
      case 1:
      case 2:
        rwIOS->rwUnpackFn = &_filterioRecordUnpack_V1V2;
        rwIOS->rwPackFn   = &_filterioRecordPack_V1V2;
        rwIOS->rwSwapFn   = &_filterioRecordSwap_V1;
        rwIOS->writeHdrFn = &_headersWriteV1;
        break;
      case 3:
        rwIOS->rwUnpackFn = &_filterioRecordUnpack_V3;
        rwIOS->rwPackFn   = &_filterioRecordPack_V3;
        rwIOS->rwSwapFn   = &_filterioRecordSwap_V3;
        rwIOS->writeHdrFn = &_headersWriteV1;
        break;
      case 4:
      case 5:
        /* V4 and V5 differ only in that V5 supports compression on
         * read and write; V4 supports compression only on read */
        rwIOS->rwUnpackFn = &_filterioRecordUnpack_V4;
        rwIOS->rwPackFn   = &_filterioRecordPack_V4;
        rwIOS->rwSwapFn   = &_filterioRecordSwap_V4;
        rwIOS->writeHdrFn = &_headersWriteV1;
        break;
      default:
        rv = LIBRW_ERR_UNSUPPORT_VERSION;
        goto END;
    }

    /* check compression */
    if ((rwGetFileVersion(rwIOS) < COMPMETHOD_SUPPORTED)
        && (rwGetFileCompMethod(rwIOS) != SK_COMPMETHOD_NONE))
    {
        if ((rwGetFileVersion(rwIOS) < (COMPMETHOD_SUPPORTED-1))
            || (rwIOS->io_mode != SK_RWIO_READ))
        {
            rv = LIBRW_ERR_UNSUPPORT_COMPRESS;
            goto END;
        }
    }

    rwIOS->recLen = _filterioGetRecLen(rwGetFileVersion(rwIOS));
    if (rwIOS->recLen == 0) {
        skAppPrintErr("Record length not set for RWFILTER version %u",
                      (unsigned int)rwGetFileVersion(rwIOS));
        assert(rwIOS->recLen > 0);
        abort();
    }

  END:
    return rv;
}


int _filterioPrepareWrite(rwIOStruct *rwIOS)
{
    int rv = LIBRW_OK;

    /* set version if none was selected by caller */
    if (rwGetFileVersion(rwIOS) == SK_FILE_VERSION_ANY) {
        ((genericHeader*)rwIOS->hdr)->version = DEFAULT_FILE_VERSION;
    }

    /* check version; set reader/writer functions */
    rv = _filterioSetFuncs(rwIOS);
    if (rv) { goto END; }

    /* create the header */
    rv = _headersCreateV1(rwIOS);
    if (rv) { goto END; }

  END:
    return rv;
}


/*
 *  Set sizes and functions by version; read header after
 *  genericHeader.  See librw_priv.h for details.
 */
int _filterioPrepareRead(rwIOStruct *rwIOS)
{
    int rv = LIBRW_OK;

    /* check version; set reader/writer functions */
    rv = _filterioSetFuncs(rwIOS);
    if (rv) { goto END; }

    /* read the header */
    rv = _headersReadV1(rwIOS);
    if (rv) { goto END; }

  END:
    return rv;
}
