/*
** Copyright (C) 2001-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  sku-times.c
**
**  various utility functions for dealing with time
**
**  Suresh L Konda
**  1/24/2002
*/


#include "silk.h"

RCSIDENT("$SiLK: sku-times.c 6394 2007-02-20 20:56:36Z mthomas $");

#include "utils.h"


/* MAX_YEARS max number of years we use for fglobbing */
#define MAX_YEARS 20


/* Default form for the timestamps */
static sktimestamp_mode_t ts_mode = SK_TIMESTAMP_DEFAULT;
static sktime_fracsec_mode_t fracsec_mode = SK_TIME_FRACSEC_DEFAULT;


/* Whether we are currently suppressing fractional seconds */
#ifdef SK_ENABLE_LEGACY_TIMESTAMP
#  define SK_TIMESTAMP_SUPPRESSING_FRACSEC()    \
    ((fracsec_mode == SK_TIME_FRACSEC_DEFAULT)  \
     ? 1                                        \
     : ((fracsec_mode == SK_TIME_FRACSEC_OFF)   \
        ? 1                                     \
        : 0))
#else
#  define SK_TIMESTAMP_SUPPRESSING_FRACSEC()    \
    ((fracsec_mode == SK_TIME_FRACSEC_DEFAULT)  \
     ? 0                                        \
     : ((fracsec_mode == SK_TIME_FRACSEC_OFF)   \
        ? 1                                     \
        : 0))
#endif


/* Whether to print fractional seconds. (0 == yes, 1 == no) */
int sktimestampSuppressingFracSecs(void)
{
    return SK_TIMESTAMP_SUPPRESSING_FRACSEC();
}


/* Determine how the time will be displayed. */
void sktimestampSetMode(
    sktimestamp_mode_t      tsm,
    sktime_fracsec_mode_t   fsm)
{
    switch (tsm) {
      case SK_TIMESTAMP_CURRENT:
        /* no change */
        break;
      case SK_TIMESTAMP_DEFAULT:
      case SK_TIMESTAMP_MMDDYYYY:
      case SK_TIMESTAMP_YYYYMMDD:
        ts_mode = tsm;
        break;
    }

    switch (fsm) {
      case SK_TIME_FRACSEC_CURRENT:
        /* no change */
        break;
      case SK_TIME_FRACSEC_ON:
      case SK_TIME_FRACSEC_OFF:
      case SK_TIME_FRACSEC_DEFAULT:
        fracsec_mode = fsm;
        break;
    }
}


/* Time to ASCII */
char *sktimestamp_r(
    char                   *outbuf,
    const struct timeval   *t,
    int                     suppress_fracsec)
{
    struct tm ts;
    struct tm *rv;

#ifdef SK_ENABLE_LOCALTIME
    rv = localtime_r(&(t->tv_sec), &ts);
#else
    rv = gmtime_r(&(t->tv_sec), &ts);
#endif
    if (NULL == rv) {
        memset(&ts, 0, sizeof(struct tm));
    }

    switch (ts_mode) {
      case SK_TIMESTAMP_MMDDYYYY:
#ifdef SK_ENABLE_LEGACY_TIMESTAMP
      case SK_TIMESTAMP_DEFAULT:
      case SK_TIMESTAMP_CURRENT:
#endif
        /* legacy "MM/DD/YYYY HH:MM:SS" */
        snprintf(outbuf, 20, "%02d/%02d/%04d %02d:%02d:%02d",
                 ts.tm_mon + 1, ts.tm_mday, ts.tm_year + 1900,
                 ts.tm_hour, ts.tm_min, ts.tm_sec);
        break;
      case SK_TIMESTAMP_YYYYMMDD:
#ifndef SK_ENABLE_LEGACY_TIMESTAMP
      case SK_TIMESTAMP_DEFAULT:
      case SK_TIMESTAMP_CURRENT:
#endif
        /* new style: "YYYY/MM/DDTHH:MM:SS" */
        snprintf(outbuf, 20, "%04d/%02d/%02dT%02d:%02d:%02d",
                 ts.tm_year + 1900, ts.tm_mon + 1, ts.tm_mday,
                 ts.tm_hour, ts.tm_min, ts.tm_sec);
    }

    if (!suppress_fracsec) {
        snprintf(&outbuf[19], 5, ".%03ld", (long int)(t->tv_usec)/1000);
    }

    return outbuf;
}


char *sktimestamp(
    const struct timeval   *t,
    int                     suppress_fracsec)
{
    static char t_buf[SK_TIMESTAMP_STRLEN];
    return sktimestamp_r(t_buf, t, suppress_fracsec);
}


char *timestamp_r(uint32_t t_sec, char *outbuf)
{
    struct timeval tv;

    tv.tv_sec = (time_t)t_sec;
    tv.tv_usec = 0;
    return sktimestamp_r(outbuf, &tv, SK_TIMESTAMP_SUPPRESSING_FRACSEC());
}


char *timestamp(uint32_t t_sec)
{
    static char t_buf[SK_TIMESTAMP_STRLEN];
    struct timeval tv;

    tv.tv_sec = (time_t)t_sec;
    tv.tv_usec = 0;
    return sktimestamp_r(t_buf, &tv, SK_TIMESTAMP_SUPPRESSING_FRACSEC());
}


/*
 *  max_day = maxDayInMonth(year, month);
 *
 *    Return the maximum number of days in 'month' in the specified
 *    'year'.
 *
 *    NOTE:  Months are in the 1..12 range and NOT 0..11
 *
 */
int maxDayInMonth(int yr, int mo)
{
    static int month_days[12] = {31,28,31,30,31,30,31,31,30,31,30,31};

    /* use a 0-based array */
    assert(1 <= mo && mo <= 12);

    /* if not February or not a potential leap year, return fixed
     * value from array */
    if ((mo != 2) || ((yr % 4) != 0)) {
        return month_days[mo-1];
    }
    /* else year is divisible by 4, need more tests */

    /* year not divisible by 100 is a leap year, and year divisible by
     * 400 is a leap year */
    if (((yr % 100) != 0) || ((yr % 400) == 0)) {
        return 1 + month_days[mo-1];
    }
    /* else year is divisible by 100 but not by 400; not a leap year */

    return month_days[mo-1];
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
