/*
** Copyright (C) 2003-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfileinfo
**
**  Prints information about an rw-file:
**    file type
**    file version
**    byte order
**    compression level
**    header size
**    record size
**    record count
**    file size
**    command line args used to generate file
**
*/


#include "silk.h"

RCSIDENT("$SiLK: rwfileinfo.c 6656 2007-03-14 15:10:48Z mthomas $");

#include "sksite.h"
#include "utils.h"
#include "rwpack.h"
#include "skstringmap.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to write --help output */
#define USAGE_FH stdout

/* This is used to store information about a files */
typedef struct _info_fileinfo {
    /* the compression method */
    sk_compmethod_t     compression;
    /* file version */
    fileVersion_t       version;
    /* file format, (RWSPLIT, etc) as integer */
    fileFormat_t        format;
    /* (uncompressed) file size */
    off_t               file_size;
    /* size of one record */
    size_t              record_length;
    /* size required for all headers */
    size_t              header_length;
    /* count of records */
    off_t               count_records;
    /* byte order */
    const char         *byte_order;
    /* filesystem path to file */
    const char         *path;
    /* command lines */
    char               *commandLines;
    /* file type as string "RWSPLIT" */
    char                format_string[SK_MAX_STRLEN_FILE_FORMAT+1];
    /* compression method as string */
    char                compress_name[SK_MAX_STRLEN_FILE_FORMAT+1];
} info_fileinfo_t;

/* A list of the properties that can be printed in the order that they
 * will be printed. */
enum info_fileinfo_id {
    RWINFO_FORMAT,
    RWINFO_VERSION,
    RWINFO_BYTE_ORDER,
    RWINFO_COMPRESSION,
    RWINFO_HEADER_LENGTH,
    RWINFO_RECORD_LENGTH,
    RWINFO_COUNT_RECORDS,
    RWINFO_FILE_SIZE,
    RWINFO_COMMAND_LINES,
    /* Last item is used to get a count of the above; it must be last */
    RWINFO_PROPERTY_COUNT
};

/* A property */
typedef struct _info_property {
    /* property's label */
    const char   *label;
    /* nonzero if printed */
    unsigned      will_print :1;
} info_property_t;


/* LOCAL VARIABLES */

static info_property_t info_props[RWINFO_PROPERTY_COUNT] = {
    {"format(id)",      1},
    {"version",         1},
    {"byte-order",      1},
    {"compression(id)", 1},
    {"header-length",   1},
    {"record-length",   1},
    {"count-records",   1},
    {"file-size",       1},
    {"command-lines",   1}
};

/* whether to print the summary */
static int print_summary = 0;

/* whether to not print titles (0==print titles, 1==no titles) */
static int no_titles = 0;

/* index into argv of first non-option argument */
static int arg_index;

/* OPTIONS SETUP */

/* Create constants for the option processor */
typedef enum rwinfoOptionIds {
    OPT_FIELDS,
    OPT_SUMMARY,
    OPT_NO_TITLES
} appOptionsEnum;

static struct option appOptions[] = {
    {"fields",          REQUIRED_ARG, 0, OPT_FIELDS},
    {"summary",         NO_ARG,       0, OPT_SUMMARY},
    {"no-titles",       NO_ARG,       0, OPT_NO_TITLES},
    {0,0,0,0}           /* sentinel entry */
};

static const char *appHelp[] = {
    NULL, /* built dynamically */
    "Print a summary of total files, file sizes, and records",
    ("Do not print file names or field names; only print the\n"
     "\tvalues, one per line"),
    (char *)NULL /* sentinel entry */
};


/* LOCAL FUNCTION DECLARATIONS */

static void appUsageLong(void);
static void appTeardown(void);
static void appSetup(int argc, char **argv);
static void appOptionsUsage(FILE *fh);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  parseFields(const char *field_str);
static int  getFileInfo(info_fileinfo_t *info);
static void printFileInfo(info_fileinfo_t *info);


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
    FILE *fh = USAGE_FH;

#define USAGE_MSG                                                       \
    ("[SWITCHES] <FILES>\n"                                             \
     "\tPrint information (type, version, etc.) about a SiLK Flow,\n"   \
     "\tIPset, or Bag file.  Use the fields switch to control what\n"   \
     "\tinformation is printed.\n")

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);
    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);
    appOptionsUsage(fh);
}


/*
 *  appOptionsUsage(FILE *fh);
 *
 *    For each option defined in this application, print its name,
 *    whether it takes an option, and its help string.  Print to the
 *    specified file handle, 'fh'.
 */
static void appOptionsUsage(FILE *fh)
{
    int i, j;

    for (i = 0; appOptions[i].name; ++i) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch (appOptions[i].val) {
          case OPT_FIELDS:
            fprintf(fh,
                    ("List of fields to print; specify by name or by value.\n"
                     "\tDef. All fields.  Available fields:"));
            for (j = 0; j < RWINFO_PROPERTY_COUNT; ++j) {
                fprintf(fh, "\n\t  %3d %s", 1+j, info_props[j].label);
            }
            break;

          default:
            fprintf(fh, "%s", appHelp[i]);
            break;
        }
        fprintf(fh, "\n");
    }
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    skAppUnregister();
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* register the options */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL))
    {
        skAppPrintErr("unable to register options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    arg_index = optionsParse(argc, argv);
    if (arg_index < 0) {
        skAppUsage();  /* never returns */
    }

    /* need at least one file */
    if (arg_index == argc) {
        skAppPrintErr("expecting file name");
        skAppUsage();  /* never returns */
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;  /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    Called by optionsParse(), this handles a user-specified switch
 *    that the application has registered, typically by setting global
 *    variables.  Returns 1 if the switch processing failed or 0 if it
 *    succeeded.  Returning a non-zero from from the handler causes
 *    optionsParse() to return a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch((appOptionsEnum)opt_index) {
      case OPT_FIELDS:
        if (parseFields(opt_arg)) {
            return 1;
        }
        break;

      case OPT_SUMMARY:
        print_summary = 1;
        break;

      case OPT_NO_TITLES:
        no_titles = 1;
        break;
    }

    return 0;
}


/*
 *  ok = parseFields(field_list);
 *
 *    Parse the user's field list, setting the appropriate flags in
 *    the info_props[] vector.
 */
static int parseFields(const char *field_str)
{
    sk_stringmap_t *str_map = NULL;
    sk_stringmap_status_t rv_map;
    sk_stringmap_entry_t *map_entry;
    sk_vector_t *field_vec = NULL;
    char *fieldnum_string = NULL;
    char *cp;
    size_t len;
    int sz;
    size_t count;
    uint32_t i;
    int rv = -1;

    /* create a char array to hold the field values as strings */
    len = 11 * RWINFO_PROPERTY_COUNT;
    fieldnum_string = calloc(len, sizeof(char));
    if (fieldnum_string == NULL) {
        skAppPrintErr("Unable to create string");
        goto END;
    }

    /* create a stringmap of the available entries */
    if (SKSTRINGMAP_OK != skStringMapCreate(&str_map)) {
        skAppPrintErr("Unable to create stringmap");
        goto END;
    }

    cp = fieldnum_string;
    for (i = 0; i < RWINFO_PROPERTY_COUNT; ++i) {
        sz = snprintf(cp, len, "%d", i+1);
        if ((size_t)sz > len) {
        }
        if (skStringMapAddID(str_map, info_props[i].label, i)
            != SKSTRINGMAP_OK)
        {
            goto END;
        }
        if (skStringMapAddID(str_map, cp, i)
            != SKSTRINGMAP_OK)
        {
            goto END;
        }
        cp += sz + 1;
        len -= sz + 1;
    }

    /* create vector to hold match results */
    field_vec = skVectorNew(sizeof(sk_stringmap_entry_t*));
    if (field_vec == NULL) {
        skAppPrintErr("allocation failure at %s:%d",
                      __FILE__, __LINE__);
        goto END;
    }

    /* attempt to match */
    rv_map = skStringMapMatch(field_vec, &cp, str_map, field_str);
    switch (rv_map) {
      case SKSTRINGMAP_OK:
        break;

      case SKSTRINGMAP_PARSE_AMBIGUOUS:
        skAppPrintErr("The --%s value '%s' is ambiguous",
                      appOptions[OPT_FIELDS].name, cp);
        goto END;

      case SKSTRINGMAP_PARSE_NO_MATCH:
        skAppPrintErr(("The --%s value '%s' is not valid"),
                      appOptions[OPT_FIELDS].name, cp);
        goto END;

      case SKSTRINGMAP_PARSE_UNPARSABLE:
        skAppPrintErr("Unable to parse the --%s value '%s'",
                      appOptions[OPT_FIELDS].name, cp);
        goto END;

      default:
        skAppPrintErr("Unexpected return value from string-map parser (%d)",
                      rv_map);
        goto END;
    }

    /* turn off printing for all fields */
    for (i = 0; i < RWINFO_PROPERTY_COUNT; ++i) {
        info_props[i].will_print = 0;
    }

    /* enable fields user listed */
    count = skVectorGetCount(field_vec);
    for (i = 0; i < count; ++i) {
        skVectorGetValue(&map_entry, field_vec, i);
        info_props[map_entry->id].will_print = 1;
    }

    rv = 0;

  END:
    if (str_map) {
        skStringMapDestroy(str_map);
    }
    if (field_vec) {
        skVectorDestroy(field_vec);
    }
    if (fieldnum_string) {
        free(fieldnum_string);
    }
    return rv;
}


/*
 *  fd = openFileGetHeader(info, &gHdr);
 *
 *    Open the current file pointed at by 'info', fill 'gHdr' with its
 *    generic header, and return the file descriptor to the opened
 *    file.  Reposition the read offset to the start of the file.
 *    Handle a file that has been completely compressed with zlib.
 *
 *    Return -1 on error.
 */
static int openFileGetHeader(
    info_fileinfo_t    *info,
    int                *fd,
    genericHeader      *ghdr)
{
    ssize_t saw = -1;

    *fd = -1;

    /* Handle compressed files */
#ifdef HAVE_ZLIB_H
    char *cp = strstr(info->path , ".gz");
    if (cp && (('\0' == *(cp + 3)) || ('.' == *(cp + 3)))) {
        /* Open the file */
        gzFile *gz = gzopen(info->path, "rb");
        if (gz == NULL) {
            skAppPrintSyserror("Cannot open file '%s'", info->path);
            return -1;
        }

        /* Read the generic header */
        saw = gzread(gz, ghdr, sizeof(genericHeader));
        gzclose(gz);
        if (saw != (ssize_t)sizeof(genericHeader)) {
            goto READ_ERROR;
        }

        return 0;
    }
#endif

    /* Open the file */
    *fd = open(info->path, O_RDONLY);
    if (*fd == -1) {
        skAppPrintSyserror("Cannot open file '%s'", info->path);
        return -1;
    }

    /* Read the generic header unless we read it above */
    saw = read(*fd, ghdr, sizeof(genericHeader));
    if (saw != (ssize_t)sizeof(genericHeader)) {
        close(*fd);
        goto READ_ERROR;
    }

    /* seek back to start of file */
    if (lseek(*fd, 0, SEEK_SET)) {
        skAppPrintSyserror("Cannot seek to start of file '%s'",
                           info->path);
        close(*fd);
        return -1;
    }

    return 0;

  READ_ERROR:
    if (saw == -1) {
        skAppPrintSyserror("Cannot read from file '%s'", info->path);
    } else {
        skAppPrintErr("Short read on '%s'; perhaps it is not a SiLK file",
                      info->path);
    }
    return -1;
}


/*
 *  getCommandHistory(rwIOS, info);
 *
 *    Get the command history from the stream 'rwIOS'---if the file
 *    contains this information---and update the 'info' structure.
 */
static void getCommandHistory(
    rwIOStruct         *rwIOS,
    info_fileinfo_t    *info)
{
    char *bp;
    char *cp;
    uint32_t i;
    uint16_t bytecount;
    sk_cmd_history_t *hist;

    /* currently, only FT_RWFILTER files have command history */
    if (FT_RWFILTER != rwGetFileType(rwIOS)) {
        return;
    }

    /* we know the length of commands will not be larger than the
     * entire header, so allocate that much space */
    info->commandLines = calloc(info->header_length, 1);
    if (info->commandLines == NULL) {
        return;
    }

    bp = info->commandLines;
    /* pointer to the command history */
    hist = &(((filterHeaderV1*)(rwIOS->hdr))->cmdHist);
    /* loop over each command */
    for (i = 0; i < hist->cmd_count; ++i) {
        /* get length and point cp at the first character in command */
        bytecount = hist->cmd_array[i].cmd_length;
        cp = hist->cmd_array[i].cmd_info;
        while(bytecount--) {
            /* each command "phrase" is separated by a \0; change it
             * to a space */
            if (*cp) {
                *bp = *cp;
            } else {
                *bp = ' ';
            }
            ++cp;
            ++bp;
        }
        /* finish this command and move forward to start the next
         * command */
        *bp = '\0';
        ++bp;
    }
}


/*
 *  getNumberRecs(rwIOS, info);
 *
 *    Given 'rwIOS' to the opened file, read the file to determine the
 *    number of records in the file; add this * information to the
 *    'info' structure.
 */
static void getNumberRecs(
    rwIOStruct         *rwIOS,
    info_fileinfo_t    *info)
{
    while (0 == rwioSkipRecords(rwIOS, 4096))
        ; /* no-op */
    info->count_records = rwGetRecordCount(rwIOS);
}


/*
 *  getFileInfo(info_fileinfo_t info);
 *
 *    Fill in the 'info' struct for the file named in the struct.
 *    Return 0 on success, or 1 on failure.
 */
static int getFileInfo(info_fileinfo_t *info)
{
    rwIOStruct *rwIOS = NULL;
    int rv = LIBRW_OK;
    int fd = -1;
    genericHeader ghdr;

    /* Open the file and read the generic header */
    if (openFileGetHeader(info, &fd, &ghdr)) {
        rv = -1;
        goto END;
    }

    /* Verify is SiLK file */
    if (CHECKMAGIC(&ghdr)) {
        rv = -1;
        skAppPrintErr("Bad magic in '%s'; not a SiLK file", info->path);
        goto END;
    }

    /* Get the info in the generic header. */
    info->format = ghdr.type;
    info->version = ghdr.version;
    info->compression = ghdr.compMethod;
    if (ghdr.isBigEndian) {
        info->byte_order = "BigEndian";
    } else {
        info->byte_order = "littleEndian";
    }

    /* Lookup names */
    sksiteFileformatGetName(info->format_string, sizeof(info->format_string),
                            info->format);
    sksiteCompmethodGetName(info->compress_name, sizeof(info->compress_name),
                            info->compression);

    /* get the file size */
    info->file_size = fileSize(info->path);

    /* Attempt to process with librw if required to get the requested
     * fields */
    if ( !(info_props[RWINFO_HEADER_LENGTH].will_print
           || info_props[RWINFO_RECORD_LENGTH].will_print
           || info_props[RWINFO_COUNT_RECORDS].will_print
           || info_props[RWINFO_COMMAND_LINES].will_print))
    {
        goto END;
    }

    /* Create an rwIOStream for the file */
    rv = rwioCreate(&rwIOS, info->path, SK_RWIO_READ);
    if (rv != LIBRW_OK) {
        rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
        goto END;
    }

    /* this may fail with an unsuppored-type error code.  If it does
     * ignore the error but don't try to read the file.  Ignore other
     * errors as well, but print them to stderr. */
    if (fd != -1) {
        rv = rwioFDOpen(rwIOS, fd);
    } else {
        rv = rwioOpen(rwIOS);
    }
    if (rv != LIBRW_OK) {
        if (rv != LIBRW_ERR_UNSUPPORT_TYPE) {
            rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
        }
        rv = LIBRW_OK;
        goto END;
    }

    info->record_length = rwGetRecordSize(rwIOS);
    info->header_length = rwGetHeaderLength(rwIOS);
    getCommandHistory(rwIOS, info);

    if (info_props[RWINFO_COUNT_RECORDS].will_print) {
        getNumberRecs(rwIOS, info);
    }

  END:
    if (rwIOS) {
        rwioDestroy(&rwIOS);
        fd = -1;
    }
    if (fd != -1) {
        close(fd);
    }
    return (!!rv);
}


/*
 *  printFileInfo(info);
 *
 *    Given the file information in the 'info' structure, print the
 *    fields requested by the user---given in the 'info_props'
 *    global---to the standard output.
 */
static void printFileInfo(info_fileinfo_t *info)
{
    const char *indentString = "  ";
    const int labelWidth = -16;
    const char *comm;
    uint32_t count;

    if (!no_titles) {
        printf("%s:\n", info->path);
    }
    if (info_props[RWINFO_FORMAT].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_FORMAT].label);
        }
        printf("%s(0x%02x)\n", info->format_string, (unsigned)info->format);
    }
    if (info_props[RWINFO_VERSION].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_VERSION].label);
        }
        printf("%u\n", (unsigned)info->version);
    }
    if (info_props[RWINFO_BYTE_ORDER].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_BYTE_ORDER].label);
        }
        printf("%s\n", info->byte_order);
    }
    if (info_props[RWINFO_COMPRESSION].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_COMPRESSION].label);
        }
        printf("%s(%u)\n", info->compress_name, (unsigned)info->compression);
    }
    if (info_props[RWINFO_HEADER_LENGTH].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_HEADER_LENGTH].label);
        }
        printf("%lu\n", (unsigned long)info->header_length);
    }
    if (info_props[RWINFO_RECORD_LENGTH].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_RECORD_LENGTH].label);
        }
        printf("%lu\n", (unsigned long)info->record_length);
    }
    if (info_props[RWINFO_COUNT_RECORDS].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_COUNT_RECORDS].label);
        }
        printf(("%" PRId64 "\n"), (int64_t)info->count_records);
    }
    if (info_props[RWINFO_FILE_SIZE].will_print) {
        if (!no_titles) {
            printf("%s%*s", indentString, labelWidth,
                   info_props[RWINFO_FILE_SIZE].label);
        }
        printf(("%" PRId64 "\n"), (int64_t)info->file_size);
    }
    if (info_props[RWINFO_COMMAND_LINES].will_print && info->commandLines) {
        if (!no_titles) {
            printf("%s%*s\n", indentString, labelWidth,
                   info_props[RWINFO_COMMAND_LINES].label);
        }
        comm = info->commandLines;
        count = 0;
        while (*comm) {
            ++count;
            if (!no_titles) {
                printf("%*u%s", (-1*labelWidth), count, indentString);
            }
            printf("%s\n", comm);
            comm += 1+strlen(comm);
        }
    }
}


/*
 *  For each file, get the file's info then print it
 */
int main(int argc, char **argv)
{
    info_fileinfo_t info;
    int64_t total_files = 0;
    int64_t total_bytes = 0;
    int64_t total_recs = 0;

    appSetup(argc, argv);			/* never returns on error */

    for ( ; arg_index < argc; ++arg_index) {
        memset(&info, 0, sizeof(info_fileinfo_t));
        info.path = argv[arg_index];
        if (0 == getFileInfo(&info)) {
            printFileInfo(&info);
            ++total_files;
            total_bytes += info.file_size;
            total_recs += info.count_records;
        }
        if (info.commandLines) {
            free(info.commandLines);
        }
    }

    if (print_summary) {
        if (!no_titles) {
            printf("**SUMMARY**:\n  %-16s", "number-files");
            printf(("%" PRId64 "\n"), total_files);
        }
        if (info_props[RWINFO_COUNT_RECORDS].will_print) {
            if (!no_titles) {
                printf("  %-16s", "total-records");
            }
            printf(("%" PRId64 "\n"), total_recs);
        }
        if (info_props[RWINFO_FILE_SIZE].will_print) {
            if (!no_titles) {
                printf("  %-16s", "all-file-sizes");
            }
            printf(("%" PRId64 "\n"), total_bytes);
        }
    }

    /* done */
    appTeardown();

    return 0;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
