/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  A simple extendable array data type
**
**  Sep 2004  Dave Svoboda
**
*/

#include "silk.h"

RCSIDENT("$SiLK: cfg_array.c 7345 2007-05-30 20:39:09Z mthomas $");

#include "cfg_config.h"



static const char* cfg_errors[] = {
    "no error",
    "general error",
    "item not found",
    "overwritten data",
    "out of memory",
    "output error",
    "invalid input data"
};


/* Return error message corresponding to error code */
const char* cfg_error_msg(int error)
{
    return (error == CFG_ERROR) ? strerror(errno) : cfg_errors[error];
}


/** String functions **/

cfg_err_t cfg_free(void *x)
{
    FREE(x);
    return CFG_OK;
}


cfg_err_t cfg_string_clone(
    void      **result,
    const void *x)
{
    char *s = CFG_CAST( char, x);
    char *r;
    CFG_ALLOCATE( char, r, strlen(s) + 2);
    strcpy( r, s);
    *result = r;
    return CFG_OK;
}


int cfg_string_equality(
    const void *x1,
    const void *x2)
{
    return !strcmp((const char*) x1, (const char*) x2);
}


cfg_err_t cfg_string_print(
    FILE       *f,
    const void *string,
    int         UNUSED(indent))
{
    CFG_FPRINTF(( f, "%s", CFG_CAST( char, string)));
    /* CFG_FPRINTF(( f, "(%p) %s", string, CFG_CAST( char, string))); */
    return CFG_OK;
}


/** Array functions **/

cfg_err_t cfg_array_create(cfg_array_t **x)
{
    /* Give the array initial space to hold 2 elements, so an immediate
       insert is fast. */
    cfg_array_t *result;
    CFG_ALLOCATE( cfg_array_t, result, 1);
    CFG_ALLOCATE( void, result->elements, 2);
    result->length = 0;
    result->capacity = 2;
    *x = result;
    return CFG_OK;
}


cfg_err_t cfg_array_create_singleton(
    cfg_array_t   **x,
    void           *element)
{
    cfg_array_t *result = NULL;
    CFG_CHECK( cfg_array_create( &result));
    CFG_CHECK( cfg_array_add( result, element));
    *x = result;
    return CFG_OK;
}


cfg_err_t cfg_array_clone(
    cfg_array_t               **result,
    const cfg_array_t          *x,
    const cfg_clone_function_t  clone_function)
{
    void *new_element;
    int i;
    CFG_CHECK( cfg_array_create( result));
    for (i = 0; i < cfg_get_array_length(x); i++) {
        CFG_CHECK( clone_function( &new_element, cfg_get_array_element(x, i)));
        CFG_CHECK( cfg_array_add( *result, new_element));
    }
    return CFG_OK;
}


cfg_err_t cfg_array_destroy(
    cfg_array_t            *x,
    cfg_destroy_function_t  destroy_function)
{
    int i;
    if (NULL != destroy_function) {
        for (i = 0; i < cfg_get_array_length(x); i++) {
            CFG_CHECK( destroy_function( cfg_get_array_element( x, i)));
        }
    }
    FREE( x->elements);
    FREE(x);
    return CFG_OK;
}


cfg_err_t cfg_array_add(
    cfg_array_t    *x,
    void           *element)
{
    void **new_elements;
    size_t old_capacity = x->capacity;

    if (x->length == x->capacity) {
        /* Capacity is maxed. Create new array 2x size, xfer elements to it,
           & discard old array */
        x->capacity *= 2;
        CFG_ALLOCATE( void*, new_elements, x->capacity);
        memcpy( new_elements, x->elements, old_capacity * sizeof( void*));
        FREE( x->elements);
        x->elements = new_elements;
    }
    x->elements[ x->length++] = element;
    return CFG_OK;
}


cfg_err_t cfg_array_concatenate(
    cfg_array_t    *x1,
    cfg_array_t    *x2)
{
    int i;

    for (i = 0; i < x2->length; i++) {
        CFG_CHECK( cfg_array_add( x1, cfg_get_array_element(x2, i)));
    }
    FREE( x2->elements);
    FREE( x2); /* don't free elements, they belong to x1 */
    return CFG_OK;
}


int cfg_array_find(
    const cfg_array_t              *x,
    const void                     *element,
    const cfg_equality_function_t   equality_function)
{
    /* FIXME: Uses slow, boring & easy linear search */
    int i;
    for (i = 0; i < x->length; i++) {
        if (equality_function( element, cfg_get_array_element( x, i))) {
            return i;
        }
    }
    return -1;
}


/* Prints out an appropriate whitespace separator between elements May
 * include a newline. Part of print routine. Indent indicates how many
 * spaces to print after a newline.  Only prints out newline if
 * threhold is big enough. */
static cfg_err_t cfg_separator_print(
    FILE   *out,
    int     indent,
    int     threshold)
{
    /* Arrays are printed all on 1 line if it has less elements than this */
    static int cfg_multiline_threshold = 5;

    int i;
    if (threshold > cfg_multiline_threshold) {
        CFG_FPRINTF(( out, "\n"));
        for (i = 0; i < indent; i++) {
            CFG_FPRINTF((out, "  "));
        }
    }
    return CFG_OK;
}


cfg_err_t cfg_array_print(
    FILE                       *out,
    const cfg_array_t          *x,
    int                         indent,
    const cfg_print_function_t  print_function)
{
    /* Print format is: {<element>, <element>, ... <element>} */
    int i;
    CFG_FPRINTF(( out, "{"));
    indent++;
    CFG_CHECK( cfg_separator_print( out, indent, x->length));
    for (i = 0; i < x->length; i++) {
        if (i > 0) {
            CFG_FPRINTF(( out, ", "));
            CFG_CHECK( cfg_separator_print( out, indent, x->length));
        }
        CFG_CHECK( print_function( out, cfg_get_array_element( x, i), indent));
    }
    indent--;
    CFG_FPRINTF(( out, "}"));
    return CFG_OK;
}


cfg_err_t cfg_map_print(
    FILE                       *out,
    const cfg_map_t            *x,
    int                         indent,
    const cfg_print_function_t  print_function)
{
    /* Print format is: {<key> = <value>, <key> = <value> ... <key> = <value>}
     */
    int i;
    CFG_FPRINTF(( out, "{"));
    indent++;
    CFG_CHECK( cfg_separator_print( out, indent, cfg_get_map_length(x)));
    for (i = 0; i < cfg_get_map_length(x); i++) {
        if (i > 0) {
            CFG_FPRINTF(( out, ", "));
            CFG_CHECK(cfg_separator_print(out, indent, cfg_get_map_length(x)));
        }
        CFG_CHECK( cfg_string_print( out, cfg_get_map_key( x, i), indent));
        CFG_FPRINTF(( out, " = "));
        CFG_CHECK( print_function( out, cfg_get_map_value( x, i), indent));
    }
    indent--;
    CFG_FPRINTF(( out, "}"));
    return CFG_OK;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
