/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/
#ifndef _CFG_CONFIG_H
#define _CFG_CONFIG_H

#include "silk.h"

RCSIDENTVAR(rcsID_CFG_CONFIG_H, "$SiLK: cfg_config.h 7345 2007-05-30 20:39:09Z mthomas $");


/*
**  cfg_config.h
**
**  Declarations/Vars shared among all the cfg_*.c files.
**
*/



/* ************************************************************************ */
/* Simple error reporting */

typedef enum cfg_err_ {
    CFG_OK = 0,
    CFG_ERROR = 1, /* in this case, errno indicates the real error */
    CFG_NOT_FOUND = 2,
    CFG_DATA_OVERWRITTEN = 3,
    CFG_OUT_OF_MEMORY = 4,
    CFG_OUTPUT_ERROR = 5,
    CFG_INVALID_INPUT = 6
} cfg_err_t;

/* Handy for debugging purposes */
#define MALLOC(x) malloc(x)
#define FREE(x) free(x)

/* Return error message corresponding to error code */
const char *cfg_error_msg(int error);


/* Macros that wrap C functions around our own error recovery code */

/* Makes typecasts much easier to grep. Note the arg becomes *type! */
#define CFG_CAST( type, value)                  \
    ((type*) value)

/* Makes typecasts much easier to grep. Note the arg becomes **type! */
#define CFG_CAST_PTR( type, value)              \
    ((type**) &(value))


/* Var must be previously declared to be a type*. */
#define CFG_ALLOCATE( type, var, degree)                                \
    {                                                                   \
        (var) = CFG_CAST( type, MALLOC( (degree) * sizeof( type)));     \
        if (NULL == (var)) {return CFG_OUT_OF_MEMORY;}                  \
    }

/* Put two parens (( )) around your fprintf args in this call */
#define CFG_FPRINTF( args )                             \
    if (0 > fprintf args) {return CFG_OUTPUT_ERROR;}

/* Returns the error if procedure returns an error */
#define CFG_CHECK( procedure)                             \
    {                                                     \
        int err = CFG_OK;                                 \
        if (CFG_OK != (err = (procedure))) {return err;}  \
    }


/* ************************************************************************ */
/* A simple extendable array data type */

/** For arrays to void* to work well, we need to know something about the
    void*'s we'll be using. So we pass type-specific functions to several
    array functions; these type-specific functions do the appropriate thing
    for our array elements.
**/

/* A destroy function type; used to destroy elements */
typedef cfg_err_t (*cfg_destroy_function_t)(void *x);

/* A clone function type; used for array/map clones */
typedef cfg_err_t (*cfg_clone_function_t)(
    void      **result,
    const void *x);

/* A comparison function type; used for array search operations.
   Should return nonzero iff elements are equal. */
typedef int (*cfg_equality_function_t)(
    const void *x1,
    const void *x2);

/* A print function type, used for printing array elements */
typedef cfg_err_t (*cfg_print_function_t)(
    FILE       *out,
    const void *x,
    int         indent);

/* The extendable array datatype */
typedef struct cfg_array_ {
     /* Actual C-style array of elements */
    void  **elements;
    /* how many elements are in array now */
    int     length;
    /* how much space for more, always >= elements */
    int     capacity;
} cfg_array_t;

/** Some useful getter macros **/

/* Returns the Nth element (void*) of the array
   (cfg_array_t*). Changing pointed-to value changes the array. */
#define cfg_get_array_element( array, index)    \
    CFG_CAST( void, (array)->elements[(index)])

/* Returns the length (int) of the array (cfg_array_t*). */
#define cfg_get_array_length( array)            \
    ((array)->length)


/** Array manipulation functions **/

/* Creates a new, empty array. Array must be destroyed with
   cfg_array_destroy */
cfg_err_t cfg_array_create(
    cfg_array_t **x);

/* Creates an array with 1 element. Array then 'owns' element; element
   is not cloned. Array must be destroyed with cfg_array_destroy. */
cfg_err_t cfg_array_create_singleton(
    cfg_array_t                     **x,
    void                     *element);

/* Returns a clone of x. Also clones elements using
   clone_function. Both Result and original array can be
   destroyed with cfg_array_destroy. */
cfg_err_t cfg_array_clone(
    cfg_array_t               **result,
    const cfg_array_t          *x,
    const cfg_clone_function_t  clone_function);

/* Destroys an array. Also applies the destroy function to each element */
cfg_err_t cfg_array_destroy(
    cfg_array_t            *x,
    cfg_destroy_function_t  destroy_function);

/* Adds element to the end of array. Array then 'owns' element;
   element is not cloned. */
cfg_err_t cfg_array_add(
    cfg_array_t    *x,
    void           *element);

/* Concatenates array 2's elements onto array 1, then destroys array
   2. Array 2's elements now belong to array 1. (Doesn't clone any
   elements) */
cfg_err_t cfg_array_concatenate(
    cfg_array_t    *x1,
    cfg_array_t    *x2);

/* Finds element in array. Returns index of element if found. or -1 if
   not found. */
int cfg_array_find(
    const cfg_array_t              *x,
    const void                     *element,
    const cfg_equality_function_t   equality_function);


/* Prints out array. Elements are printed using print function.
   Indent indicates how many spaces to indent each element */
cfg_err_t cfg_array_print(
    FILE                       *out,
    const cfg_array_t          *x,
    int                         indent,
    const cfg_print_function_t  print_function);

/*** Internal functions ***/

/** Here are some suitable element functions for operating on char*'s **/

/* Destroys an object (merely a wrapper around free() */
cfg_err_t cfg_free(
    void *x);

/* Clones a string, stores clone in result, wrapper around strcpy */
cfg_err_t cfg_string_clone(
    void       **result,
    const void  *x);

/* Returns true if strings are equal, wrapper around strcmp */
int cfg_string_equality(
    const void *x1,
    const void *x2);

/* Prints a string. Returns CFG_OK if successful */
cfg_err_t cfg_string_print(
    FILE       *out,
    const void *x,
    int         indent);


/* ************************************************************************ */
/* A simple map data type */

/** These maps are essentially a Map< char*, void*>. We know our keys
    are strings, but our values can be any type. We pass type-specific
    functions to several map functions to deal with map values.
**/


/* The map type; maps keys to values. All keys are strings. */
typedef struct cfg_map_ {
    cfg_array_t    *keys;     /* Keys. All are char* */
    cfg_array_t    *values;   /* Values. May be any type */
} cfg_map_t;


/** Some useful getter macros **/

/* Returns the Nth key (char*) of the map
   (cfg_map_t*). Do not change the map key. */
#define cfg_get_map_key( map, index)                            \
    CFG_CAST( char, cfg_get_array_element((map)->keys, index))

/* Returns the Nth value (void*) of the map
   (cfg_map_t*). Changing pointed-to value changes the map. */
#define cfg_get_map_value( map, index)                                  \
    CFG_CAST( void, cfg_get_array_element((map)->values, index))

/* Returns the length (int) of the map (cfg_map_t*). */
#define cfg_get_map_length( map)                \
    cfg_get_array_length((map)->keys)


/** Map manipulation functions **/

/* Creates an empty map. Map must be destroyed with cfg_map_destroy */
cfg_err_t cfg_map_create(
    cfg_map_t **x);

/* Destroys a map. Destroys all keys, and applies the destroy function
   to each value */
cfg_err_t cfg_map_destroy(
    cfg_map_t              *x,
    cfg_destroy_function_t  destroy_function);

/* Adds element to map. If key already existed, overwrites old value
   and returns CFG_DATA_OVERWRITTEN. (Doesn't clone or free elements,
   so map now owns key and value.) */
cfg_err_t cfg_map_add(
    cfg_map_t  *x,
    char       *key,
    void       *value);

/* Returns actual value associated with key in map x. or NULL if key
   not found. Changing this value changes the map's contents. Map
   still owns the returned value. */
void *cfg_map_assoc(
    const cfg_map_t    *x,
    const char         *key);

/* Prints out map. Print function is used for values. */
cfg_err_t cfg_map_print(
    FILE                       *out,
    const cfg_map_t            *x,
    int                         indent,
    const cfg_print_function_t  print_function);


/** These are helper functions and vars for maps **/

/** Here are some suitable element functions for operating on arrays
    of char*s **/

/* A destroy wrapper function to destroy an array of pointers that can be
   merely free()'d. (eg char*). Can be used to destroy an array of char*s. */
cfg_err_t cfg_array_free_destroy(
    void *x);

/* A sample print wrapper function to print an array of strings (eg char*) */
cfg_err_t cfg_array_string_print(
    FILE       *f,
    const void *x,
    int         indent);


/* ************************************************************************ */
/* A simple date data type */

/** Our date type is merely a pair of integers. */

/* The date type. */
typedef struct cfg_date_ {
    /* stores the day, month, year */
    time_t date;
} cfg_date_t;

#define UNSPECIFIED_HOUR -1

/* No date comes before this date */
extern cfg_date_t date_min;

/* No date comes after this date */
extern cfg_date_t date_max;


/** Date manipulation functions **/

/* Creates a date. Must be destroyed with cfg_date_destroy. */
cfg_err_t cfg_date_create(
    cfg_date_t    **x,
    time_t          date);

/* Clones a date. Both result and original date must be destroyed with
   cfg_date_destroy. */
cfg_err_t cfg_date_clone(
    void      **result,
    const void *x);

/* Destroys a date. */
cfg_err_t cfg_date_destroy(
    cfg_date_t     *x);

/* Compares two dates, returns -1 if date1 precedes date2, +1 if date1
   succeeds date2, and 0 if they are equal. (A date w/o an hour
   precedes the same date with an hour.) */
int cfg_date_compare(
    const cfg_date_t   *d1,
    const cfg_date_t   *d2);


/* Prints out date. */
cfg_err_t cfg_date_print(
    FILE       *out,
    const void *x,
    int         indent);


/* ************************************************************************ */
/* A simple date-range data type */

/** Our date range type is merely a pair of dates. We are assuming:
 *  dates employ the same ordering as integers
 *  0 = minimum date, max_int = maximum date
 **/


/* The date range type. */
typedef struct cfg_daterange_ {
    cfg_date_t *minimum;
    cfg_date_t *maximum;    /* minimum <= maximum */
} cfg_daterange_t;


/** Some useful getter macros **/

/* Returns the maximum (cfg_date_t*) of the daterange
   (cfg_daterange_t*). Changing pointed-to value changes the
   daterange. */
#define cfg_get_daterange_minimum( date) ((date)->minimum)

/* Returns the minimum (cfg_date_t*) of the daterange
   (cfg_daterange_t*). Changing pointed-to value changes the
   daterange. */
#define cfg_get_daterange_maximum( date) ((date)->maximum)


/** Date range manipulation functions **/

/* Creates a daterange. Daterange owns components; does not clone
   them. Daterange must be destroyed with cfg_daterange_destroy. */
cfg_err_t cfg_daterange_create(
    cfg_daterange_t   **x,
    cfg_date_t         *minimum,
    cfg_date_t         *maximum);

/* Creates a daterange. Daterange owns components; does not clone
   them. Daterange must be destroyed with cfg_daterange_destroy. */
cfg_err_t cfg_daterange_create_before(
    cfg_daterange_t   **x,
    cfg_date_t         *maximum);

/* Creates a daterange. Daterange owns components; does not clone
   them. Daterange must be destroyed with cfg_daterange_destroy. */
cfg_err_t cfg_daterange_create_after(
    cfg_daterange_t   **x,
    cfg_date_t         *minimum);

/* Clones a daterange, including components. Both result and original
   daterange must be destroyed with cfg_daterange_destroy. */
cfg_err_t cfg_daterange_clone(
    void      **result,
    const void *x);

/* Destroys a daterange. */
cfg_err_t cfg_daterange_destroy(
    void       *x);

/* Returns nonzero if date lives within date range */
int cfg_date_in_range(
    const cfg_daterange_t  *x,
    const cfg_date_t       *d);

/* Prints out daterange. */
cfg_err_t cfg_daterange_print(
    FILE           *out,
    const void     *x,
    int             indent);


/* ************************************************************************ */
/* A simple path data type */

/** Our path type is merely a pair of strings + a date set (which may
    be NULL) **/


/* The path type; really a pair */
typedef struct cfg_path_ {
    char           *sensor;
    char           *flowlabel;
    cfg_array_t    *dateranges;    /* Array< date-range> */
} cfg_path_t;


/** Some useful getter macros **/

/* Returns the sensor (char*) of the path (cfg_path_t*). Changing
   pointed-to value changes the path. */
#define cfg_get_path_sensor( path) ((path)->sensor)

/* Returns the flowlabel (char*) of the path (cfg_path_t*). Changing
   pointed-to value changes the path. */
#define cfg_get_path_flowlabel( path) ((path)->flowlabel)

/* Returns the flowlabel (char*) of the path (cfg_path_t*). Changing
   pointed-to value changes the path. */
#define cfg_get_path_dateranges( path) ((path)->dateranges)


/** Path manipulation functions **/

/* Creates a path. Path will clone the components. Dates may be
 * NULL. Path must be destroyed with cfg_path_destroy. */
cfg_err_t cfg_path_create(
    cfg_path_t    **x,
    char           *sensor,
    char           *flowlabel,
    cfg_array_t    *dateranges);

/* Clones a path, including components. Both result and original path
   must be destroyed with cfg_path_destroy. */
cfg_err_t cfg_path_clone(
    void      **result,
    const void *x);

/* Destroys a path, and destroys components. */
cfg_err_t cfg_path_destroy(
    void       *x);

/* Prints out path. */
cfg_err_t cfg_path_print(
    FILE       *out,
    const void *x,
    int         indent);


/* ************************************************************************ */
/* The configuration object */

/** The configuration datatype is meant to store everything that lives in
    a configuration file. **/


/* Our config data type */
typedef struct cfg_config_ {
    cfg_map_t *properties;    /* A Map< Variable, Value > */
    cfg_map_t *sensors;       /* A Map< Variable, Array< Sensor> > */
    cfg_map_t *flowlabels;    /* A Map< Variable, Array< Flowlabel> > */
    cfg_map_t *dates;         /* A Map< Variable, Array< Date-Range> > */
    cfg_map_t *domains;       /* A Map< Domain, Map< Path-Variable,
                                 Array< Path> > > */
} cfg_config_t;

/** Some useful getter macros **/

/* Returns the properties (Map< Variable, String>) of the config
   (cfg_config_t*). */
#define cfg_get_config_properties( config)      \
    ((config)->properties)

/* Returns the sensors (Map< Variable, Array< Sensor>) of the config
   (cfg_config_t*). */
#define cfg_get_config_sensors( config)         \
    ((config)->sensors)

/* Returns the flowlabels (Map< Variable, Array< Flowlabel>) of the
   config (cfg_config_t*). */
#define cfg_get_config_flowlabels( config)      \
    ((config)->flowlabels)

/* Returns the flowlabels (Map< Variable, Array< DateRange>) of the
   config (cfg_config_t*). */
#define cfg_get_config_dates( config)           \
    ((config)->dates)

/* Returns the domains (Map< Domain, Map< Path-Variable, Array< Path>
   > >) of the config (cfg_config_t*). */
#define cfg_get_config_domains( config)         \
    ((config)->domains)


/* Returns a particular domain (Map< Path, Array< Path> >) associated
   with a config and a domain name. Returns NULL if domain doesn't
   exist. */
#define cfg_lookup_config_domain( config, domain)                       \
    CFG_CAST(cfg_map_t,                                                 \
             cfg_map_assoc(cfg_get_config_domains( config), domain))

/* Returns a particular path (Array< Path>) given a domain and a path
   name. Returns NULL if path doesn't exist in domain. */
#define cfg_lookup_domain_paths( domain, path_var)              \
    CFG_CAST( cfg_array_t, cfg_map_assoc( domain, path_var))

/* Returns a particular path (Path) given an array of paths and an index */
#define cfg_get_array_path( paths, i)                           \
    CFG_CAST( cfg_path_t, cfg_get_array_element( paths, i))


/** Config manipulation functions **/

/* Creates an empty configuration. Config must be destroyed with
   cfg_config_destroy */
cfg_err_t cfg_config_create(
    cfg_config_t  **x);

/* Destroys a configuration. Also destroys all components. */
cfg_err_t cfg_config_destroy(
    cfg_config_t   *x);

/* Prints out configuration. */
cfg_err_t cfg_config_print(
    FILE               *out,
    const cfg_config_t *x,
    int                 indent);


/* A printf-style function type, used for printing error messages */
typedef int (*cfg_printf_function_t)(const char *fmt, ...);

/* Parses a configuration file into a configuration struct.

   - config must have already been created, but need not be 'empty';
   preexisting data in config is kept.

   - error_fn gets called for every error in the code. You can give it
   NULL, in which case it defaults to printf.

   Returns CFG_OK if file is completely valid, or
   CFG_INVALID_INPUT if any parse errors occurred.
 */
cfg_err_t cfg_config_read(
    cfg_config_t               *x,
    FILE                       *input,
    const cfg_printf_function_t error_fn);


/* A quick lookup function for getting a path list. Returns value
   assoc with path in given domain, or NULL if not found. Value
   returned is an Array< Path>, and is still 'owned' by the configuration */
const cfg_array_t *cfg_config_lookup(
    const cfg_config_t *x,
    const char         *domain,
    const char         *path);

/* A quick lookup function for getting a path list. Returns value
   assoc with path in given domain, or NULL if not found. Value
   returned is an Array< Path>, and must be freed by the user using
   cfg_array_path_destroy */
cfg_array_t *cfg_config_lookup_date(
    const cfg_config_t *x,
    const char         *domain,
    const char         *path,
    time_t              time_stamp);

/* Returns the property value associated with the given name. Returns
   NULL if not found. Value returned is still owned by config. */
const char *cfg_config_get_property(
    const cfg_config_t *x,
    const char         *property_name);


/*** Internal functions ***/

/** We need a handful of support functions to handle the config datatypes **/

/* Destroys a map whose values are arrays of free()able ptrs */
cfg_err_t cfg_map_free_destroy(
    void       *x);

/* Destroys an array whose values are date-ranges */
cfg_err_t cfg_array_daterange_destroy(
    void       *x);

/* Prints out an array of paths */
cfg_err_t cfg_array_path_print(
    FILE       *f,
    const void *x,
    int         indent);

/* Destroys an array of paths */
cfg_err_t cfg_array_path_destroy(
    void       *x);

/* Prints out an array of date ranges */
cfg_err_t cfg_array_daterange_print(
    FILE       *f,
    const void *x,
    int         indent);

/* Prints out a map linking strings to an array of paths */
cfg_err_t cfg_map_path_print(
    FILE       *f,
    const void *x,
    int         indent);


/* Returns true if date is within any of the given ranges, or if the
   ranges are NULL */
int cfg_date_in_ranges(
    const cfg_date_t   *date,
    const cfg_array_t  *ranges);


#endif /* _CFG_CONFIG_H */

/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
