/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  A simple date range data type
**
**  Sep 2004  Dave Svoboda
**
*/

#include "silk.h"

RCSIDENT("$SiLK: cfg_daterange.c 7345 2007-05-30 20:39:09Z mthomas $");

#include "cfg_config.h"


/** Daterange functions **/


cfg_err_t cfg_daterange_create(
    cfg_daterange_t   **x,
    cfg_date_t         *minimum,
    cfg_date_t         *maximum)
{
    cfg_daterange_t *result;

    if (cfg_date_compare( minimum, maximum) == 1) {
        return CFG_INVALID_INPUT;
    }
    CFG_ALLOCATE( cfg_daterange_t, result, 1);
    cfg_get_daterange_maximum( result) = maximum;
    cfg_get_daterange_minimum( result) = minimum;
    *x = result;
    return CFG_OK;
}


cfg_err_t cfg_daterange_create_before(
    cfg_daterange_t   **x,
    cfg_date_t         *maximum)
{
    cfg_date_t *d;

    /* Clone date_min b/c it gets owned by the date getting it */
    CFG_CHECK( cfg_date_clone( CFG_CAST_PTR( void, d), &date_min));
    return cfg_daterange_create( x, d, maximum);
}


cfg_err_t cfg_daterange_create_after(
    cfg_daterange_t   **x,
    cfg_date_t         *minimum)
{
    cfg_date_t *d;
    /* Clone date_max b/c it gets owned by the date getting it */
    CFG_CHECK( cfg_date_clone( CFG_CAST_PTR( void, d), &date_max));
    return cfg_daterange_create( x, minimum, d);
}


cfg_err_t cfg_daterange_clone(
    void      **result,
    const void *x)
{
    cfg_daterange_t *p = (cfg_daterange_t*) x;
    cfg_date_t *new_minimum;
    cfg_date_t *new_maximum;

    CFG_CHECK( cfg_date_clone( CFG_CAST_PTR( void, new_minimum),
                              cfg_get_daterange_minimum(p)));
    CFG_CHECK( cfg_date_clone( CFG_CAST_PTR( void, new_maximum),
                              cfg_get_daterange_maximum(p)));
    return cfg_daterange_create( CFG_CAST_PTR( cfg_daterange_t, *result),
                                 new_minimum, new_maximum);
}


cfg_err_t cfg_daterange_destroy(void *x)
{
    cfg_daterange_t *p = CFG_CAST(cfg_daterange_t, x);

    FREE( cfg_get_daterange_maximum(p));
    FREE( cfg_get_daterange_minimum(p));
    FREE( x);
    return CFG_OK;
}


int cfg_date_in_range(
    const cfg_daterange_t  *x,
    const cfg_date_t       *d)
{
    return (cfg_date_compare( cfg_get_daterange_minimum(x), d) != 1 &&
            cfg_date_compare( cfg_get_daterange_maximum(x), d) != -1);
}

cfg_err_t cfg_daterange_print(
    FILE       *out,
    const void *x,
    int         indent)
{
    /* Print format is: <date> - <date> or MAXIMUM <date> or MINIMUM <date> */
    cfg_daterange_t *d = CFG_CAST( cfg_daterange_t, x);

    if (cfg_date_compare( cfg_get_daterange_maximum(d), &date_max) == 0) {
        CFG_FPRINTF(( out, "AFTER "));
        CFG_CHECK( cfg_date_print( out, cfg_get_daterange_minimum(d), indent));
    } else if (cfg_date_compare(cfg_get_daterange_minimum(d), &date_min) == 0){
        CFG_FPRINTF(( out, "BEFORE "));
        CFG_CHECK( cfg_date_print( out, cfg_get_daterange_maximum(d), indent));
    } else {
        CFG_CHECK( cfg_date_print( out, cfg_get_daterange_minimum(d), indent));
        CFG_FPRINTF(( out, " TO "));
        CFG_CHECK( cfg_date_print( out, cfg_get_daterange_maximum(d), indent));
    }
    return CFG_OK;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
