%{
/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Parser data for config file parser
**
**  Sep 2004  Dave Svoboda
**
*/

#include "silk.h"

RCSIDENT("$SiLK: cfg_parser.y 7345 2007-05-30 20:39:09Z mthomas $");

#include "utils.h"
#include "cfg_lexer.h"
#include "cfg_config.h"


/* The current config structure we should add stuff to. */
static cfg_config_t *current_config = NULL;

/* The current domain we should add stuff to; part of the current config */
static cfg_map_t *current_domain = NULL;

/* Function to use for reporting errors */
static cfg_printf_function_t cfg_default_printf_function;

/* Set during grammar parsing */
static cfg_err_t cfg_yerrno;

/* Set of variables (only used to see what's been defined */
static cfg_map_t *cfg_vars_map = NULL;


/* Macros that wrap C functions around our own error recovery code */

/* Put two parens (( )) around your fprintf args in this call */
#define CFG_YPRINTF( args)                                      \
    if (0 > cfg_default_printf_function args) {                 \
        fprintf( stderr, "Error in reporting grammar error\n"); \
        cfg_yerrno = CFG_OUTPUT_ERROR;                          \
    }

/* Var must be previously declared to be a type*. */
#define CFG_YALLOCATE( type, var, degree)                     \
    var = CFG_CAST( type, MALLOC( (degree) * sizeof( type))); \
    if (NULL == var) {                                        \
        CFG_YPRINTF(( cfg_error_msg( CFG_OUT_OF_MEMORY)));    \
        cfg_yerrno = CFG_OUT_OF_MEMORY;                       \
    }

/* Sets cfg_yerrno if procedure returns an error */
#define CFG_YCHECK( procedure)                                          \
    {                                                                   \
        int err = CFG_OK;                                               \
        if (CFG_OK != (err = (procedure))) {                            \
            CFG_YPRINTF(( "Grammar Error: %s\n", cfg_error_msg( err))); \
            cfg_yerrno = err;                                           \
        }                                                               \
    }

/** YACC requires us to declare these fns here, but define them after
    grammar **/

static void cfg_map_add_from_grammar(
    cfg_map_t      *map,
    char           *key,
    cfg_array_t    *value,
    const char     *label);


static void cfg_map_variable_values(
    cfg_array_t               **result,
    const cfg_map_t            *map,
    const char                 *key,
    const char                 *label,
    const cfg_clone_function_t  clone_function);


static void cfg_path_cross_product(
    cfg_array_t       **result,
    const cfg_array_t  *sensors,
    const cfg_array_t  *flowlabels,
    cfg_array_t        *dateranges);


cfg_err_t cfg_config_read(
    cfg_config_t               *x,
    FILE                       *input,
    const cfg_printf_function_t error_fn);


%}
%union {
    char *string;
    cfg_array_t *array;
    cfg_date_t *date;
    cfg_daterange_t *daterange;
}
%token SENSOR_GROUP_T
%token FLOWLABEL_GROUP_T
%token DATERANGE_GROUP_T
%token PATH_GROUP_T
%token PARTITION_GROUP_T
%token DOMAIN_T
%token PROPERTY_T
%token BEFORE_T
%token AFTER_T
%token WITH_T
%token TO_T
%token EQUALS
%token SEMICOLON
%token COLON
%token DOLLAR
%token COMMA
%token PERIOD
%token OPENBRACE
%token CLOSEBRACE
%token OPENBRACKET
%token CLOSEBRACKET
%token <string> ID
%token <string> DATE
%type <date> date
%type <daterange> daterange
%type <string> daterange_var flowlabel_var path_var sensor_var
%type <string> flowlabel sensor
%type <array> flowlabel_set flowlabel_entries flowlabel_entry
%type <array> sensor_set sensor_entries sensor_entry
%type <array> daterange_set daterange_entries daterange_entry
%type <array> path_set path_entries path_entry


%% /* Grammar rules and actions follow */
input:             /* empty */
                 | decl input
;


/* We don't do any structure adding up here; each decl adds its decl
   to current_config independently. We use current_path &
   current_domain b/c we need their current contents in order to look
   up variables. */

decl:              flowlabel_decl
                 | sensor_decl
                 | property_decl
                 | daterange_decl
                 | domain_decl
/*
                 | partition_decl
                 | partition_stmt
*/
                 | error
{
    CFG_YPRINTF(( "Line %d: Syntax Error\n", cfg_lex_line_number));
    if (CFG_OK == cfg_yerrno) {cfg_yerrno = CFG_INVALID_INPUT;};
};


flowlabel_decl:    FLOWLABEL_GROUP_T flowlabel_var EQUALS flowlabel_set SEMICOLON
{
    cfg_map_add_from_grammar( cfg_get_config_flowlabels( current_config),
                              $2, $4, "Flowlabel");
};


flowlabel_var:     DOLLAR ID
{
    $$ = $2;
};


flowlabel_set:     flowlabel_entry
                 | OPENBRACE flowlabel_entries CLOSEBRACE
{
    $$ = $2;
};


flowlabel_entries: flowlabel_entry
                 | flowlabel_entry COMMA flowlabel_entries
{
    $$ = $1;
    CFG_YCHECK( cfg_array_concatenate( $$, $3));
};


flowlabel_entry:   flowlabel_var
{
    cfg_map_variable_values( &$$, cfg_get_config_flowlabels( current_config),
                             $1, "Flowlabel", cfg_string_clone);
    FREE($1);
};
                 | flowlabel
{
    CFG_YCHECK( cfg_array_create_singleton( &$$, $1));
};


flowlabel:         ID
;



sensor_decl:       SENSOR_GROUP_T sensor_var EQUALS sensor_set SEMICOLON
{
    cfg_map_add_from_grammar( cfg_get_config_sensors( current_config),
                              $2, $4, "Sensor");
};


sensor_var:        DOLLAR ID
{
    $$ = $2;
};


sensor_set:        sensor_entry
                 | OPENBRACE sensor_entries CLOSEBRACE
{
    $$ = $2;
};


sensor_entries:    sensor_entry
                 | sensor_entry COMMA sensor_entries
{
    $$ = $1;
    CFG_YCHECK( cfg_array_concatenate( $$, $3));
};


sensor_entry:      sensor_var
{
    cfg_map_variable_values( &$$, cfg_get_config_sensors( current_config),
                             $1, "Sensor", cfg_string_clone);
    FREE($1);
};
                 | sensor
{
    CFG_YCHECK( cfg_array_create_singleton( &$$, $1));
};


sensor:            ID
;



domain_decl:       DOMAIN_T ID path_decl_set SEMICOLON
{
    cfg_map_add_from_grammar( cfg_get_config_domains(current_config),
                              $2, CFG_CAST( cfg_array_t, current_domain),
                              "Domain");
    current_domain = NULL;
};

path_decl_set:        path_decl
                 | OPENBRACE path_decl_set2 CLOSEBRACE
;


path_decl_set2:       path_decl_set
                 | path_decl_set path_decl_set2
;


path_decl:         PATH_GROUP_T path_var EQUALS path_set SEMICOLON
{
    if (current_domain == NULL) {
        CFG_YCHECK( cfg_map_create( &current_domain));
    }
    CFG_YCHECK( cfg_map_add( current_domain, $2, $4));
};


path_var:          DOLLAR ID COLON ID
{
    /* We enforce the internal struct of a path var in the grammar,
       but once it parses, we can ignroe the internal struct, just turn
       it back into a simple string */
    int length = strlen( $2);
    char *name;
    CFG_YALLOCATE( char, name, (strlen( $2) + strlen( $4) + 2));
    strcpy( name, $2);
    strcpy(name + length, ":");
    strcpy(name + length + 1, $4);
    FREE( $2);
    FREE( $4);
    $$ = name;
};


path_set:          path_entry
                 | OPENBRACE path_entries CLOSEBRACE
{
    $$ = $2;
};


path_entries:      path_entry
                 | path_entry COMMA path_entries
{
    $$ = $1;
    CFG_YCHECK( cfg_array_concatenate( $$, $3));
};


path_entry:        path_var
{
    cfg_map_variable_values( &$$, current_domain, $1, "Path", cfg_path_clone);
    FREE($1);
};
                 | sensor_set PERIOD flowlabel_set
{
    cfg_path_cross_product( &$$, $1, $3, NULL);
    cfg_array_destroy( $1, cfg_free);
    cfg_array_destroy( $3, cfg_free);
}
                 | sensor_set PERIOD flowlabel_set daterange_set
{
    cfg_path_cross_product( &$$, $1, $3, $4);
    cfg_array_destroy( $1, cfg_free);
    cfg_array_destroy( $3, cfg_free);
    cfg_array_destroy( $4, cfg_daterange_destroy);
};



property_decl:     PROPERTY_T ID ID SEMICOLON
{
    CFG_YCHECK( cfg_map_add( cfg_get_config_properties( current_config),
                                                        $2, $3));
};



daterange_decl:       DATERANGE_GROUP_T daterange_var EQUALS daterange_set SEMICOLON
{
    cfg_map_add_from_grammar( cfg_get_config_dates( current_config),
                              $2, $4, "Date Range");
};


daterange_var:        DOLLAR ID
{
    $$ = $2;
};


daterange_set:        daterange_entry
                    | OPENBRACE daterange_entries CLOSEBRACE
{
    $$ = $2;
};


daterange_entries:    daterange_entry
                    | daterange_entry COMMA daterange_entries
{
    $$ = $1;
    CFG_YCHECK( cfg_array_concatenate( $$, $3));
};


daterange_entry:      daterange_var
{
    cfg_map_variable_values( &$$, cfg_get_config_dates( current_config),
                             $1, "Date Range", cfg_daterange_clone);
    FREE($1);
};
                    | daterange
{
    CFG_YCHECK( cfg_array_create_singleton( &$$, $1));
};


daterange:            date TO_T date
{
    CFG_YCHECK( cfg_daterange_create( &$$, $1, $3));
};

daterange:            BEFORE_T date
{
    CFG_YCHECK( cfg_daterange_create_before( &$$, $2));
};

daterange:            AFTER_T date
{
    CFG_YCHECK( cfg_daterange_create_after( &$$, $2));
};

date:                 DATE
{
    struct timeval tv;
    skStringParseDatetime(&tv, $1, NULL);
    CFG_YCHECK( cfg_date_create( CFG_CAST_PTR( cfg_date_t, $$), tv.tv_sec));
};


/*
partition_decl:    PARTITION_GROUP_T partition_var
                   EQUALS partition_set SEMICOLON
;

partition_var:     DOLLAR ID {
  printf("partition variable :%s:\n", (char*) $2);
};

partition_set:     partition
                 | OPENBRACE partition_set2 CLOSEBRACE
                 | partition_var
;

partition_set2:    partition_set
                 | partition_set COMMA partition_set2
;

partition:         ID EQUALS ID {
  printf("partition :%s: = :%s:\n", (char*) $1, (char*) $3);
};



partition_stmt:    WITH_T flowlabel PARTITION_GROUP_T EQUALS partition_set SEMICOLON
;

*/
%%


/*
 * Adds declaration to proper map, printing friendly error msg if item
 * already defined.
 */
static void cfg_map_add_from_grammar(
    cfg_map_t      *map,
    char           *key,
    cfg_array_t    *value,
    const char     *label)
{
  char *key_clone;
  CFG_YALLOCATE( char, key_clone, strlen( key) + 1);
  strcpy( key_clone, key);
  if (CFG_DATA_OVERWRITTEN == cfg_map_add( map, key, value) ||
      CFG_DATA_OVERWRITTEN == cfg_map_add( cfg_vars_map, key_clone, NULL)) {
        CFG_YPRINTF(( "Line %d: %s variable $%s redefined!\n",
                      cfg_lex_line_number, label, key));
        if (CFG_OK == cfg_yerrno) {cfg_yerrno = CFG_INVALID_INPUT;};
    }
}


/*
 * Returns the value of a variable, from the appropraite map, printing
 * friendly error msg if not found. Clones variable's values, so that
 * free doesn't crash later.
 */
static void cfg_map_variable_values(
    cfg_array_t               **result,
    const cfg_map_t            *map,
    const char                 *key,
    const char                 *label,
    const cfg_clone_function_t  clone_function)
{
    cfg_array_t *values = cfg_map_assoc( map, key);
    CFG_YCHECK( cfg_array_create( result));
    if (values == NULL) {
        CFG_YPRINTF(( "Line %d: %s variable $%s is not defined!\n",
                       cfg_lex_line_number, label, key));
        if (CFG_OK == cfg_yerrno) {cfg_yerrno = CFG_INVALID_INPUT;};
    } else {
        cfg_array_t *clone;
        CFG_YCHECK( cfg_array_clone( &clone, values, clone_function));
        CFG_YCHECK( cfg_array_concatenate( *result, clone));
    }
}


/*
 * Returns an array of paths generated by combinatorially enumerating
 * the array of sensors and array of flowlabels. Dateranges just gets
 * assinged; and it may be NULL.
 */
static void cfg_path_cross_product(
    cfg_array_t       **result,
    const cfg_array_t  *sensors,
    const cfg_array_t  *flowlabels,
    cfg_array_t        *dateranges)
{
    int i, j;
    CFG_YCHECK( cfg_array_create( result));
    for (i = 0; i < cfg_get_array_length( sensors); i++) {
        for (j = 0; j < cfg_get_array_length( flowlabels); j++) {
            cfg_path_t *path;
            CFG_YCHECK( cfg_path_create( &path,
                                         cfg_get_array_element(sensors, i),
                                         cfg_get_array_element( flowlabels, j),
                                         dateranges));
            CFG_YCHECK( cfg_array_add( *result, path));
        }
    }
}


/*
 *   Parses a configuration file into a configuration struct.
 *
 *   - config must have already been created, but need not be 'empty';
 *   preexisting data in config is kept.
 *
 *   - error_fn gets called for every error in the code. You can give it
 *   NULL, in which case it defaults to printf.
 *
 *   Returns CFG_OK if file is completely valid, or
 *   CFG_INVALID_INPUT if any parse errors occurred.
 */
cfg_err_t cfg_config_read(
    cfg_config_t               *x,
    FILE                       *input,
    const cfg_printf_function_t error_fn)
{
    cfg_default_printf_function = printf;
    if (error_fn != NULL) {
        cfg_default_printf_function = error_fn;
    }
    current_config = x; /* purser puts data into current_config */
    cfg_map_create( &cfg_vars_map);
    cfg_yerrno = CFG_OK;
    yyin = input;
    yyparse();
    cfg_map_destroy( cfg_vars_map, cfg_free);
    return cfg_yerrno;
}


/* Yacc calls this function when it encounters an error */
int yyerror(char UNUSED(*s))
{
    return 0;
}

/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
