/*
** Copyright (C) 2001-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfilter.c
**
**  6/3/2001
**
**  Suresh L. Konda
**
**  Allows for selective extraction of records and fields from a rw
**  packed file.  This version, unlike rwcut, creates a binary file with
**  the filtered records.  A new file type is used.  The header does not
**  contain valid recCount and rejectCount values.  The other fields are
**  taken from the original input file.
**
**  A second header is also created which records the filter rules used
**  for each pass.  Thus this is a variable length header
*/


#include "silk.h"

RCSIDENT("$SiLK: rwfilter.c 6989 2007-04-19 21:54:20Z mthomas $");

#include "rwfilter.h"


/* TYPEDEFS AND DEFINES */

/*
 *  INCR_REC_COUNT(count, rec)
 *
 *    Increment the values in the rec_count_t 'count' by the values in
 *    the rwRec 'rec'.
 *
 */
#define INCR_REC_COUNT(count, rec) \
    { (count).flows++; (count).pkts+=(rec).pkts; (count).bytes+=(rec).bytes; }


/* EXPORTED VARIABLES */

/* struct to hold all the io streams */
filter_io_t filter_io;

/* index into argv of first option that does not start with a '--'.
 * This assumes getopt rearranges the options, which gnu getopt will
 * do. */
int arg_index;

/* maximum number of records the user wants to see.  zero for ALL. */
uint64_t max_records[DESTINATION_TYPES];

/* counts needed for --print-stat option */
filterStatisticsV2_t stats;

/* number of checks to preform */
int checker_count = 0;

/* function pointers to handle checking and or processing */
int (*checker[MAX_CHECKERS])(rwRec*);

/* user shared object plug-in */
dynlibInfoStruct *dlISP;

/* number of shared object plug-ins loaded */
int app_dynlib_count;

/* pointer to each plug-in */
dynlibInfoStruct *app_dynlib_list[APP_MAX_DYNLIBS];

/* names of plug-ins to attempt to load at startup */
const char *app_dynlib_names[] = {
    "libaddrtype.so",
    "libccfilter.so",
    "libpmapfilter.so",
    "libipport.so",
    NULL /* sentinel */
};



/* LOCAL VARIABLES */

/* cache of argc and argv used for dumping headers */
static int pargc;
static char **pargv;


/* LOCAL FUNCTIONS DECLARATIONS */

static int  filterFile(const char *inFName);
static int  readFileNamesFromFile(const char *stream);



/* FUNCTION DEFINITIONS */


/*
 *  int = readFileNamesFromStream(stream);
 *
 *    Read the input from 'stream' one line at a time.  Treat the line
 *    as the name of a file to be opened and filtered.  Return 0 on
 *    success and non-zero on error.
 *
 *    Spaces on the input lines are significant: leading, trailing,
 *    and embedded spaces are treated as part of the file name.
 *
 *    Lines that are too long are ignored, and a warning is printed to
 *    the stderr.
 */
static int readFileNamesFromFile(const char *stream)
{
    char line[PATH_MAX+1];
    char *eol;
    FILE *handle;
    int is_pipe = -1; /* assume stdin */
    unsigned long line_count = 0;
    int rv = 0; /* return value */

    /* Open the xargs stream */
    if (strcmp(stream, "stdin") == 0) {
        handle = stdin;
    } else if (openFile(stream, 0 /*read*/, &handle, &is_pipe)) {
        skAppPrintErr("Cannot open --xargs stream '%s'", stream);
        return -1;
    }

    /* read from handle until EOF */
    while (!feof(handle)) {
        if (!fgets(line, sizeof(line), handle)) {
            continue;
        }
        ++line_count;
        eol = strchr(line, '\n');
        if (eol == line) {
            /* empty line; ignore */
            continue;
        }

        if (NULL != eol) {
            /* expected behavior: read an entire line */
            *eol = '\0';
        } else if (feof(handle)) {
            /* okay: last line did not end in newline */
        } else {
            /* bad: line was longer than sizeof(line).  read until newline
             * or eof, then throw away the line */
            skAppPrintErr("xargs stream input line %lu too long. ignored",
                          line_count);
            while (fgets(line, sizeof(line), handle) && !strchr(line,'\n'))
                ; /* empty */
            continue;
        }
        rv |= filterFile(line);
    }

    /* close xargs */
    if (is_pipe == 1) {
        pclose(handle);
    } else if (is_pipe == 0) {
        fclose(handle);
    }

    return rv;
}


/*
 *  printHeaders(input_rwios);
 *
 *    Create and print the header to each output file; include the
 *    current command line invocation in the header.  If 'input_rwios'
 *    is non-null and it is an FT_RWFILTER file, the file history from
 *    that file is also included in the header.
 */
static void printHeaders(const rwIOStruct *in_ios)
{
    static int did_headers = 0;
    rwIOStruct *out_ios;
    int i, j;
    int rv;

    /* only print headers one time */
    if (did_headers) {
        return;
    }
    did_headers = 1;

    /* don't print anything on a dry-run */
    if (filter_io.dryrun_fp) {
        return;
    }

    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            out_ios = filter_io.dest_stream[i][j].ios;
            if ( !out_ios) {
                continue;
            }

            rv = LIBRW_OK;

            /* if 'in_ios' is provided and if it is an rwfilter file,
             * add its headers to each output file's headers */
            if (in_ios && FT_RWFILTER == rwGetFileType(in_ios)) {
                rv = rwioAppendToHistoryFromFile(out_ios, in_ios);
            }
            if ((rv != LIBRW_OK)
                || (rv = rwioAppendToHistoryArgv(out_ios, pargc, pargv))
                || (rv = rwioWriteHeader(out_ios)))
            {
                rwioPrintLastErr(out_ios, rv, &skAppPrintErr);
            }
        }
    }
}


/*
 *  closeAllDests();
 *
 *    Close all the output destinations.  Return 0 if they all closed
 *    cleanly, or non-zero if there was an error closing any stream.
 */
int closeAllDests(void)
{
    int i, j;
    int rv = 0;

    /* close all the output files (pass, fail, all-dest) */
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            if (filter_io.dest_stream[i][j].ios) {
                rv |= rwioClose(filter_io.dest_stream[i][j].ios);
                rv |= rwioDestroy(&(filter_io.dest_stream[i][j].ios));
                filter_io.dest_stream[i][j].ios = NULL;
            }
            filter_io.dest_stream[i][j].fname = NULL;
        }
    }

    return rv;
}


/*
 *  num_output_streams = closeOutputDests(dest_type, idx);
 *
 *    Close one stream or close all streams for a particular
 *    destination type (pass, fail, all-dest).
 *
 *    When idx is non-negative, close the destination stream at
 *    position 'idx' for the specified destination type 'dest_type'
 *    (e.g., DEST_ALL).
 *
 *    When idx is negative, close all the streams for the specified
 *    dest_type.
 *
 *    Adjust the count of output streams for 'dest_type'
 *    appropriately, and return the number of output streams across
 *    all destination types that are still open.
 */
static int closeOutputDests(
    const int   dest_type,
    int         idx)
{
    int i;
    int rv;
    int count;
    int num_outs;

    if (idx < 0) {
        /* close ALL streams of 'dest_type' */
        for (i = 0; i < filter_io.dest_count[dest_type]; ++i) {
            rv = rwioClose(filter_io.dest_stream[dest_type][i].ios);
            if (rv) {
                rwioPrintLastErr(filter_io.dest_stream[dest_type][i].ios,
                                 rv, &skAppPrintErr);
            }
            rwioDestroy(&(filter_io.dest_stream[dest_type][i].ios));
        }

        filter_io.dest_count[dest_type] = 0;
    } else {
        /* close stream 'idx' of 'dest_type' */
        rv = rwioClose(filter_io.dest_stream[dest_type][idx].ios);
        if (rv) {
            rwioPrintLastErr(filter_io.dest_stream[dest_type][idx].ios,
                             rv, &skAppPrintErr);
        }
        rwioDestroy(&(filter_io.dest_stream[dest_type][idx].ios));

        /* decrement dest_type's stream count and move the last stream to
         * stream idx's position---unless 'idx' is the last stream. */
        --filter_io.dest_count[dest_type];
        count = filter_io.dest_count[dest_type];
        if (idx != count) {
            filter_io.dest_stream[dest_type][idx].ios =
                filter_io.dest_stream[dest_type][count].ios;
            filter_io.dest_stream[dest_type][idx].fname =
                filter_io.dest_stream[dest_type][count].fname;
        }
    }

    /* compute and return the number of output streams that we now
     * have. */
    num_outs = 0;
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        num_outs += filter_io.dest_count[i];
    }

    return num_outs;
}


/*
 *  ok = filterFile(fpath);
 *
 *    This is the actual filtering of a file.  Also handles dumping
 *    out the header if required.  Return 0 on success; or 1 if the
 *    input file could not be opened.
 */
static int filterFile(const char *fpath)
{
    static int rwfilter_warn_printed = 0;
    static int reading_records = 1;
    static rwRec rwrec;
    int i;
    int fail;
    int rv = LIBRW_OK;

    /* nothing to do in dry-run mode but print the file names */
    if (filter_io.dryrun_fp) {
        fprintf(filter_io.dryrun_fp, "%s\n", fpath);
        return 0;
    }

    /* print filenames if requested */
    if (filter_io.filenames_fp) {
        fprintf(filter_io.filenames_fp, "%s\n", fpath);
    }

    if (!reading_records) {
        return 0;
    }

    /* open the input file */
    filter_io.input.ios = rwOpenFile(fpath, NULL);
    if (NULL == filter_io.input.ios) {
        /* some error. the library would have dumped a msg.  treat as
         * non-fatal error so processing can continue. */
        return 1;
    }

    ++stats.fileCount;

    if (stats.fileCount == 1) {
        /* first file, print the headers to the output files */
        printHeaders(filter_io.input.ios);
    } else if ((FT_RWFILTER == rwGetFileType(filter_io.input.ios))
               && !rwfilter_warn_printed)
    {
        rwfilter_warn_printed = 1;
        skAppPrintErr("Command line history from RWFILTER file is being lost;"
                      "\n\thistory from first file only is maintained");
    }

    /* read and process each record */
    while (reading_records && rwRead(filter_io.input.ios, &rwrec)) {

        /* increment number of read records */
        INCR_REC_COUNT(stats.rCount, rwrec);

        /* the all-dest */
        if (filter_io.dest_count[DEST_ALL]) {
            i = 0;
            while (i < filter_io.dest_count[DEST_ALL]) {
                rv = rwWrite(filter_io.dest_stream[DEST_ALL][i].ios, &rwrec);
                if ( !LIBRW_ERROR_IS_FATAL(rv)) {
                    ++i;
                } else {
                    /* close the one stream and see if we need to
                     * continue reading.  DO NOT increment i here,
                     * since the stream at position i will be
                     * different after calling closeOutputDests. */
                    reading_records = closeOutputDests(DEST_ALL, i);
                }
            }

            /* close all streams for this destination type if we are
             * at user's requested max.  If max_records[] is 0, this
             * will never be true, and all records will be
             * processed. */
            if (stats.rCount.flows == max_records[DEST_ALL]) {
                reading_records = closeOutputDests(DEST_ALL, -1);
            }
        }


        /* run all checker()'s until end or one fails */
        for (i=0, fail=0; i < checker_count && !fail; ++i) {
            fail = (*(checker[i]))(&rwrec);
        }

        if (!fail) {

            /* increment number of record that pass */
            INCR_REC_COUNT(stats.pCount, rwrec);

            /* the pass-dest */
            if (filter_io.dest_stream[DEST_PASS]) {
                i = 0;
                while (i < filter_io.dest_count[DEST_PASS]) {
                    rv = rwWrite(filter_io.dest_stream[DEST_PASS][i].ios,
                                 &rwrec);
                    if ( !LIBRW_ERROR_IS_FATAL(rv)) {
                        ++i;
                    } else {
                        /* close the one stream and see if we need to
                         * continue reading.  DO NOT increment i here,
                         * since the stream at position i will be
                         * different after calling closeOutputDests. */
                        reading_records = closeOutputDests(DEST_PASS, i);
                    }
                }

                /* close all streams for this destination type if we are
                 * at user's requested max.  If max_records[] is 0, this
                 * will never be true, and all records will be
                 * processed. */
                if (stats.pCount.flows == max_records[DEST_PASS]) {
                    reading_records = closeOutputDests(DEST_PASS, -1);
                }
            }
        } else {

            /* the fail-dest */
            if (filter_io.dest_stream[DEST_FAIL]) {
                i = 0;
                while (i < filter_io.dest_count[DEST_FAIL]) {
                    rv = rwWrite(filter_io.dest_stream[DEST_FAIL][i].ios,
                                 &rwrec);
                    if ( !LIBRW_ERROR_IS_FATAL(rv)) {
                        ++i;
                    } else {
                        /* close the one stream and see if we need to
                         * continue reading.  DO NOT increment i here,
                         * since the stream at position i will be
                         * different after calling closeOutputDests. */
                        reading_records = closeOutputDests(DEST_FAIL, i);
                    }
                }

                /* close all streams for this destination type if we are
                 * at user's requested max.  If max_records[] is 0, this
                 * will never be true, and all records will be
                 * processed. */
                if ((stats.rCount.flows - stats.pCount.flows)
                    == max_records[DEST_FAIL])
                {
                    reading_records = closeOutputDests(DEST_FAIL, -1);
                }
            }
        }

    } /* while (reading_records && rwRead()) */

    /* close input */
    rwioDestroy(&(filter_io.input.ios));
    filter_io.input.ios = NULL;

    return 0;
}



int main(int argc, char **argv)
{
    int rv = 0;
    int i;

    skAppRegister(argv[0]);


    appSetup(argc, argv);

    pargc = argc;
    pargv = argv;

    /* Get the files.  Only one of these should be active */
    if (fglobValid()) {
        const char * curFName;

        while (NULL != (curFName = fglobNext())) {
            rv |= filterFile(curFName);
        }
    } else if (filter_io.input.fname) {
        rv = filterFile(filter_io.input.fname);
    } else if (filter_io.xargs) {
        rv = readFileNamesFromFile(filter_io.xargs);
    } else {
        for (i = arg_index; i < argc; ++i) {
            rv |= filterFile(argv[i]);
        }
    }

    /*
     * We don't write the header to the destination file(s) until
     * we've read the first record.  However, if no files were read,
     * the destination files are empty, so dump the header to them
     * now.
     */
    if (stats.fileCount == 0) {
        printHeaders(NULL);
    }


    if (filter_io.stats.fp) {
        /* stats.fp should never be set on the nodes */
        filterStatisticsDump(filter_io.stats.fp, &stats);
    }

    appTeardown();


    return rv;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
