/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/
#ifndef _RWFILTER_H
#define _RWFILTER_H

#include "silk.h"

RCSIDENTVAR(rcsID_RWFILTER_H, "$SiLK: rwfilter.h 6989 2007-04-19 21:54:20Z mthomas $");

/*
**  rwfilter.h
**
**  Privater header information for the rwfilter application
**
*/

#include "rwpack.h"
#include "utils.h"
#include "dynlib.h"
#include "rwfilterstats.h"
#include "sksite.h"
#include "iptree.h"


/* TYPEDEFS AND DEFINES */

/* where to write --help output */
#define USAGE_FH stdout

/* where to send --dry-run output */
#define DRY_RUN_FH stdout

/* where to send file names when --print-filenames is active */
#define PRINT_FILENAMES_FH stderr

/* maximum number of dynamic libraries that we support */
#define APP_MAX_DYNLIBS 8

/* maximum number of filter checks */
#define MAX_CHECKERS (APP_MAX_DYNLIBS + 2)

/*
 *  The number and types of rwIOStruct output streams: pass, fail, all
 */
#define DESTINATION_TYPES 3
enum {
    DEST_PASS = 0, DEST_FAIL, DEST_ALL
};


/* max number of any destination we support.  If we only wanted one of
 * each, it could greatly simplify the code since we could nix the
 * arrays. */
#define MAX_DESTINATIONS 2


/* A wrapper around the rwIOStruct for reading and writing RW files;
 * allows us to keep the file name. */
typedef struct _rwios_wrapper {
    const char *fname;
    rwIOStruct *ios;
} rwios_wrapper_t;

/* A wrapper around a FILE* for reading and writing non-RW files */
typedef struct _file_wrapper {
    const char *fname;
    FILE       *fp;
    int       (*closefn)(FILE*);
} file_wrapper_t;

/* All of the inputs and outputs for rwfilter */
typedef struct _filter_io {
    /* current input file */
    rwios_wrapper_t input;
    /* count of each type of destination */
    int             dest_count[DESTINATION_TYPES];
    /* array to hold each destination */
    rwios_wrapper_t dest_stream[DESTINATION_TYPES][MAX_DESTINATIONS];
    /* support for the --print-statistics switch */
    file_wrapper_t  stats;
    /* where to print output during a --dry-run.  NULL when the switch
     * has not been provided. */
    FILE           *dryrun_fp;
    /* where to print output for --print-filenames.  NULL when the switch
     * has not been provided. */
    FILE           *filenames_fp;
    /* support for the --xargs switch */
    const char     *xargs;
} filter_io_t;


/* EXTERNAL VARIABLES (rwfilter.c) */

/* struct to hold all the io streams */
extern filter_io_t filter_io;

/* index into argv of first option that does not start with a '--'.
 * This assumes getopt rearranges the options, which gnu getopt will
 * do. */
extern int arg_index;

/* maximum number of records the user wants to see.  zero for ALL. */
extern uint64_t max_records[DESTINATION_TYPES];

/* counts needed for --print-stat option */
extern filterStatisticsV2_t stats;

/* number of checks to preform */
extern int checker_count;

/* function pointers to handle checking and or processing */
extern int (*checker[MAX_CHECKERS])(rwRec*);

/* user shared object plug-in */
extern dynlibInfoStruct *dlISP;

/* number of shared object plug-ins loaded */
extern int app_dynlib_count;

/* pointer to each plug-in */
extern dynlibInfoStruct *app_dynlib_list[APP_MAX_DYNLIBS];

/* names of plug-ins to attempt to load at startup */
extern const char *app_dynlib_names[];


/* FUNCTION DECLARATIONS */

/* fglob functions (fglob.c) */

void fglobUsage(FILE *);
int  fglobSetup(void);
void fglobTeardown(void);
char *fglobNext(void);
int  fglobValid(void);
int  fglobSetFilters(
    sk_bitmap_t   **sensor_bitmap,
    sk_bitmap_t   **flowtype_bitmap);


/* application setup functions (rwfilterutils.c) */

void appSetup(int argc, char **argv);
void appTeardown(void);


/* application functions (rwfilter.c) */

int  closeAllDests(void);


/* filtering  functions (rwfiltercheck.c) */

int  filterCheck(const rwRec *rwrec);
void filterUsage(FILE*);
int  filterGetCheckCount(void);
int  filterGetFGlobFilters(void);
skIPTree_t *filterGetIPsetByIndex(int set_idx);
int  filterSetup(void);
void filterTeardown(void);




#endif /* _RWFILTER_H */


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
