/*
** Copyright (C) 2005-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Routines to produce the output for the --print-statistics and
**  --print-volume-statistics switches.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: rwfilterstats.c 7479 2007-06-12 16:32:39Z mthomas $");

#include "rwfilterstats.h"
#include "utils.h"
#include "rwfilter.h"


/* Write the statistics in 'hdr' to the 'buf' */
int filterStatisticsAsString(
    char                       *buf,
    size_t                      buf_len,
    const filterStatisticsV2_t *hdr)
{
    int nbytes = -1;

    /* check input */
    if (buf == NULL || buf_len == 0) {
        return -1;
    }

    /* detailed or simple statistics? */
    if (hdr->volumeStats) {
        nbytes =
            snprintf(buf, buf_len,
                     ("%5s|%18s|%18s|%20s|%10s|\n"
                      "%5s|%18" PRIu64 "|%18" PRIu64 "|%20" PRIu64 "|%10u|\n"
                      "%5s|%18" PRIu64 "|%18" PRIu64 "|%20" PRIu64 "|%10s|\n"
                      "%5s|%18" PRIu64 "|%18" PRIu64 "|%20" PRIu64 "|%10s|\n"),
                     /* titles */ "", "Recs", "Packets", "Bytes", "Files",

                     "Total",
                     hdr->rCount.flows,
                     hdr->rCount.pkts,
                     hdr->rCount.bytes,
                     hdr->fileCount,

                     "Pass",
                     hdr->pCount.flows,
                     hdr->pCount.pkts,
                     hdr->pCount.bytes,
                     "",

                     "Fail",
                     (hdr->rCount.flows - hdr->pCount.flows),
                     (hdr->rCount.pkts  - hdr->pCount.pkts),
                     (hdr->rCount.bytes - hdr->pCount.bytes),
                     "");
    } else {
        nbytes = snprintf(buf, buf_len,
                          ("Files %5u.  Read %10" PRIu64 "."
                           "  Pass %10" PRIu64 ". Fail  %10" PRIu64 ".\n"),
                          hdr->fileCount, hdr->rCount.flows, hdr->pCount.flows,
                          (hdr->rCount.flows - hdr->pCount.flows));
    }

    /* make certain we didn't overflow the buffer */
    if (nbytes < 0 || nbytes >= (int)buf_len) {
        buf[buf_len-1] = '\0';
        nbytes = -1;
    }

    return nbytes;
}


/* write text version of statistics in 'hdr' to the file 'fh' */
static void filterStatisticsDumpAscii(
    FILE                       *fh,
    const filterStatisticsV2_t *hdr)
{
    char buf[1024];
    int nbytes;

    nbytes = filterStatisticsAsString(buf, sizeof(buf), hdr);
    if (nbytes == -1) {
        skAppPrintErr("Error generating statistics");
    }

    fprintf(fh, "%s", buf);
}


/* write binary version of statistics in 'hdr' to file 'fh' */
static void filterStatisticsDumpBinary(
    FILE                       *fh,
    const filterStatisticsV2_t *hdr)
{
    fwrite(hdr, sizeof(filterStatisticsV2_t), 1, fh);
}


/* print the statistics in proper form */
void filterStatisticsDump(FILE *fh, const filterStatisticsV2_t *hdr)
{
    if ( isRunningOnNode() ) {
        /* print in binary */
        filterStatisticsDumpBinary(fh, hdr);
    } else {
        filterStatisticsDumpAscii(fh, hdr);
    }
}


/* Initialize the 'hdr' structure */
int filterStatisticsInit(filterStatisticsV2_t *hdr)
{
    if (NULL == hdr) {
        return -1;
    }

    PREPHEADER( &(hdr->gHdr));
    hdr->gHdr.compMethod = SK_COMPMETHOD_NONE;
    hdr->gHdr.type = FT_RWPRINTSTATS;
    hdr->gHdr.version = 2;

    return 0;
}


/* Reader statistics from file descriptor 'fd' and add to 'sum' */
int filterStatisticsAddFromFile(
    filterStatisticsV2_t       *sum,
    const genericHeader UNUSED(*gHdr),
    int                         fd)
{
    filterStatisticsV2_t in;

    assert(gHdr->isBigEndian == IS_BIG_ENDIAN);
    assert(gHdr->type == FT_RWPRINTSTATS);
    assert(gHdr->version == 2);

    filterStatisticsInit(&in);
    if ((sizeof(rec_count_t) != read(fd, &in.rCount, sizeof(rec_count_t)))
        || (sizeof(rec_count_t) != read(fd, &in.pCount, sizeof(rec_count_t)))
        || (sizeof(uint32_t) != read(fd, &in.fileCount, sizeof(uint32_t)))
        || (sizeof(uint8_t) != read(fd, &in.volumeStats, sizeof(uint8_t))))
    {
        /* skAppPrintErr("Unable to read filterStatistics"); */
        return -1;
    }

    sum->rCount.flows += in.rCount.flows;
    sum->rCount.pkts  += in.rCount.pkts;
    sum->rCount.bytes += in.rCount.bytes;
    sum->pCount.flows += in.pCount.flows;
    sum->pCount.pkts  += in.pCount.pkts;
    sum->pCount.bytes += in.pCount.bytes;
    sum->fileCount    += in.fileCount;
    sum->volumeStats  |= in.volumeStats;

    return 0;
}




/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
