/*
** Copyright (C) 2001-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfilterutils.c
**
**  utility routines for rwfilter.c
**
**  Suresh L. Konda
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwfilterutils.c 6989 2007-04-19 21:54:20Z mthomas $");

#include "rwfilter.h"


/* TYPEDEFS AND DEFINES */



/* LOCAL FUNCTIONS */

static void appUsageLong(void);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  appAddDynlib(const char *dlpath, int warnOnError);
static int  filterCheckInputs(int argc);
static int  filterCheckOutputs(void);
static int  filterSetCheckers(void);


/* LOCAL VARIABLES */

/* the compression method to use when writing the file.
 * sksiteCompmethodOptionsRegister() will set this to the default or
 * to the value the user specifies. */
static sk_compmethod_t comp_method;


/* OPTION SETUP */

/* these values used to index into following arrays.  need to keep
 * everything in sync */
typedef enum {
    OPT_DRY_RUN, OPT_INPUT_PIPE, OPT_XARGS,
    OPT_MAX_PASS_RECORDS, OPT_PRINT_FILE, OPT_DYNLIB,
    OPT_PASS_DEST, OPT_FAIL_DEST, OPT_ALL_DEST,
    OPT_PRINT_STAT, OPT_PRINT_VOLUME
} appOptionsEnum;

static struct option appOptions[] = {
    {"dry-run",                 NO_ARG,       0, OPT_DRY_RUN},
    {"input-pipe",              REQUIRED_ARG, 0, OPT_INPUT_PIPE},
    {"xargs",                   REQUIRED_ARG, 0, OPT_XARGS},
    {"max-pass-records",        REQUIRED_ARG, 0, OPT_MAX_PASS_RECORDS},
    {"print-filenames",         NO_ARG,       0, OPT_PRINT_FILE},
    {OPT_DYNAMIC_LIBRARY,       REQUIRED_ARG, 0, OPT_DYNLIB},

    {"pass-destination",        REQUIRED_ARG, 0, OPT_PASS_DEST},
    {"fail-destination",        REQUIRED_ARG, 0, OPT_FAIL_DEST},
    {"all-destination",         REQUIRED_ARG, 0, OPT_ALL_DEST},
    {"print-statistics",        OPTIONAL_ARG, 0, OPT_PRINT_STAT},
    {"print-volume-statistics", OPTIONAL_ARG, 0, OPT_PRINT_VOLUME},
    {0,0,0,0}                   /* sentinel entry */
};

static const char *appHelp[] = {
    "Parse command line switches but do not process records",
    ("Read SiLK flow records from a pipe: 'stdin' or\n"
     "\tpath to named pipe. No default"),
    ("Read list of input file names from a file or pipe\n"
     "\tpathname or 'stdin'. No default"),
    "Print at most this many records; 0 for all.  Def 0",
    "Print names of input files during processing. Def. No",
    ("Augment processing with the specified dynamic\n"
     "\tlibrary. No default\n"
     "At least one of the following output switches is required:"),
    ("Destination for records which pass the filter(s):\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("Destination for records which fail the filter(s):\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("Destination for all records regardless of pass/fail:\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("Print a count of total flows read to named file.\n"
     "\tIf no pathname provided, use stderr. No default"),
    ("Print count of flows/packets/bytes read\n"
     "\tto named file. If no pathname provided, use stderr. No default"),
    (char *)NULL
};




/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                           \
    ("<app-opts> <partition-opts> {<selection-opts> | <inputFiles>}\n"      \
     "\tPartitions SiLK Flow records into one or more 'pass' and/or\n"      \
     "\t'fail' output streams.  The source of the SiLK records can\n"       \
     "\tbe stdin, a named pipe, files listed on the command line, or\n"     \
     "\tfiles selected from the data-store via the selection switches.\n"   \
     "\tThere is no default input or output; these must be specified.\n")

    FILE *fh = USAGE_FH;
    int i, j;


    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);

    fprintf(fh, "\nAPPLICATION SWITCHES control input and output\n\n");
    skOptionsDefaultUsage(fh);
    sksiteCompmethodOptionsUsage(fh);
    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    fglobUsage(fh);
    filterUsage(fh);

    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibOptionsUsage(app_dynlib_list[j], fh);
    }
    if (dlISP) {
        dynlibOptionsUsage(dlISP, fh);
    }
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
void appSetup(int argc, char **argv)
{
    int j;
    int input_count;
    int output_count;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char*)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    /* skAppRegister(argv[0]);  -- we do this in main() */
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize variables */
    memset(&filter_io, 0, sizeof(filter_io_t));
    memset(max_records, 0, sizeof(max_records));

    /* load fglob module */
    if (fglobSetup()) {
        skAppPrintErr("unable to setup fglob module");
        exit(EXIT_FAILURE);
    }

    /* load filter module */
    if (filterSetup()) {
        skAppPrintErr("unable to setup filter module");
        exit(EXIT_FAILURE);
    }

    /* register the options */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)
        || sksiteCompmethodOptionsRegister(&comp_method))
    {
        skAppPrintErr("unable to register options");
        exit(EXIT_FAILURE);
    }

    /* initialize stats data structure */
    filterStatisticsInit(&stats);

    /* don't print error messages on the nodes, and tell getopt_long()
     * to be quiet */

    /* add "hard-coded" dynamic libraries */
    for (j = 0; app_dynlib_names[j]; ++j) {
        appAddDynlib(app_dynlib_names[j], 0);
    }

    /* parse options */
    arg_index = optionsParse(argc, argv);
    assert(arg_index <= argc);
    if (arg_index < 0) {
        skAppUsage();
    }

    /* Check that there is one and only one source of input to process */
    input_count = filterCheckInputs(argc);
    if (input_count < 0) {
        /* fatal error. msg already printed */
        exit(EXIT_FAILURE);
    } else if (input_count > 1) {
        skAppPrintErr("Multiple input sources were specified\n"
                      "\tInput must come from only one of --input-pipe,"
                      " --xargs, file names on\n"
                      "\tthe command line,"
                      " or a combination of the file selection switches");
        skAppUsage();
    } else if (input_count == 0) {
        skAppPrintErr("No input was specified.\n"
                      "\tNo file selection switches were given,"
                      " neither --input-pipe nor --xargs\n"
                      "\twas specified, and no files are present on the"
                      " command line");
        skAppUsage();
    }

    /* Check we have a filtering rule--either built in or from a
     * dynamic library--or the --all-dest option */
    checker_count = filterSetCheckers();
    if ((checker_count <= 0) && (filter_io.dest_count[DEST_ALL] == 0)) {
        skAppPrintErr("No filtering rules given");
        skAppUsage();
    }


    /* check all the output streams */
    output_count = filterCheckOutputs();
    if (output_count < 0) {
        /* fatal error. msg already printed */
        exit(EXIT_FAILURE);
    } else if (output_count == 0) {
        skAppPrintErr("No output(s) specified. Abort");
        skAppUsage();
    }

    /* Try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names.  If fglob
     * is active, it will require the configuration file. */
    sksiteConfigure(0);

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to optionsRegister(); it will be called
 *    by optionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes optionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    int dest_id;
    int dest_count;

    switch ((appOptionsEnum)opt_index) {
      case OPT_PASS_DEST:
      case OPT_FAIL_DEST:
      case OPT_ALL_DEST:
        /* an output stream */
        dest_id = opt_index - OPT_PASS_DEST;
        assert(dest_id >= 0 && dest_id < DESTINATION_TYPES);

        dest_count = filter_io.dest_count[dest_id];
        if (dest_count == MAX_DESTINATIONS) {
            skAppPrintErr("Only %d --%s options allowed",
                          MAX_DESTINATIONS, appOptions[opt_index].name);
            return 1;
        }
        filter_io.dest_stream[dest_id][dest_count].fname = opt_arg;
        ++filter_io.dest_count[dest_id];
        break;

      case OPT_INPUT_PIPE:
        /*
         *  input-pipe. Delay check for multiple inputs sources till all
         *  options have been parsed
         */
        if (filter_io.input.fname) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        filter_io.input.fname = opt_arg;
        break;

      case OPT_DYNLIB:
        if (dlISP) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        dlISP = dynlibCreate(DYNLIB_SHAR_FILTER);
        if (dynlibLoad(dlISP, opt_arg)) {
            skAppPrintErr("fatal error loading library '%s':\n  %s",
                          opt_arg, dynlibLastError(dlISP));
            return 1;
        }
        break;

      case OPT_DRY_RUN:
        filter_io.dryrun_fp = DRY_RUN_FH;
        break;

      case OPT_PRINT_FILE:
        filter_io.filenames_fp = PRINT_FILENAMES_FH;
        break;

      case OPT_PRINT_VOLUME:
        stats.volumeStats = 1;
        /* FALLTHROUGH */
      case OPT_PRINT_STAT:
        if (filter_io.stats.fname) {
            skAppPrintErr("may only specified one of --%s or --%s",
                          appOptions[OPT_PRINT_STAT].name,
                          appOptions[OPT_PRINT_VOLUME].name);
            return 1;
        } else if (!opt_arg) {
            filter_io.stats.fname = "stderr";
        } else {
            filter_io.stats.fname = opt_arg;
        }
        break;

      case OPT_MAX_PASS_RECORDS:
        if (skStringParseUint64(&(max_records[DEST_PASS]), opt_arg, 0, 0)) {
            skAppPrintErr("Cannot parse %s value '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case OPT_XARGS:
        /* input-pipe for file names */
        if (filter_io.xargs) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        filter_io.xargs = opt_arg;
        break;
    } /* switch */

    return 0;                     /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
void appTeardown(void)
{
    static uint8_t teardownFlag = 0;
    int j;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    dynlibTeardown(dlISP);

    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibTeardown(app_dynlib_list[j]);
        app_dynlib_list[j] = NULL;
    }
    app_dynlib_count = 0;

    filterTeardown();

    fglobTeardown();

    closeAllDests();

    /* close stats */
    if (filter_io.stats.fp && filter_io.stats.closefn) {
        if (-1 == (*filter_io.stats.closefn)(filter_io.stats.fp)) {
                skAppPrintErr("Error closing --print-stat stream '%s': %s",
                              filter_io.stats.fname, strerror(errno));
        }
    }
    filter_io.stats.fname = NULL;

    /* close input */
    if (filter_io.input.ios) {
        rwioDestroy(&filter_io.input.ios);
        filter_io.input.ios = NULL;
    }
    filter_io.input.fname = NULL;


    skAppUnregister();
}


/*
 *  appAddDynlib(dlpath, warnOnError);
 *
 *    Load the dynamic library at 'dlpath', store its information in
 *    the global 'app_dynlib_list', and increment the global
 *    'app_dynlib_count'.  Return 0 on success.  Return 1 if the
 *    library cannot be loaded or if too many dynamic libraries have
 *    been loaded.
 */
static int appAddDynlib(const char *dlpath, int warnOnError)
{
    dynlibInfoStructPtr dl;

    if (app_dynlib_count == APP_MAX_DYNLIBS) {
        skAppPrintErr("Too many dynlibs.  Only %u allowed",
                      APP_MAX_DYNLIBS);
        return 1;
    }

    dl = dynlibCreate(DYNLIB_SHAR_FILTER);
    if (dynlibLoad(dl, dlpath)) {
        /* failed to find library, missing symbols, or setup failed */
        if (warnOnError) {
            skAppPrintErr("Error loading dynamic library '%s': %s",
                          dlpath, dynlibLastError(dl));
        }
        dynlibTeardown(dl);
        return 1;
    }

    app_dynlib_list[app_dynlib_count] = dl;
    ++app_dynlib_count;

    return 0;
}


/*
 *  stream_count = filterCheckInputs(argc)
 *
 *    Do basic checks for the input, and return the number of inputs
 *    selected from among: command-line files, fglob options, piped
 *    input, or xargs.  Returns -1 on error, such as trying to read
 *    binary records from a TTY.
 */
static int filterCheckInputs(int argc)
{
    int count = 0;
    int rv;

    /* Check for file names on command line */
    if (arg_index < argc) {
        ++count;
    }

    /* the input-pipe must be 'stdin' or it must be an existing FIFO.
     * If 'stdin', stdin must not be a TTY, since we expect binary. */
    if (filter_io.input.fname) {
        ++count;
        if (strcmp(filter_io.input.fname, "stdin") == 0) {
            if (FILEIsATty(stdin)) {
                skAppPrintErr("stdin is connected to a terminal. Abort.");
                return -1;
            }
        } else if ( !fileExists(filter_io.input.fname)) {
            skAppPrintErr("input-pipe '%s' does not exist",
                          filter_io.input.fname);
            return -1;
        } else if ( !isFIFO(filter_io.input.fname)) {
            skAppPrintErr("input-pipe '%s' is not a pipe",
                          filter_io.input.fname);
            return -1;
        }
    }

    /* the xargs argument must be 'stdin' (terminal is ok) or it must
     * exist. Open it. */
    if (filter_io.xargs) {
        ++count;
        if ((strcmp(filter_io.xargs, "stdin") != 0)
            && !fileExists(filter_io.xargs))
        {
            skAppPrintErr("The --xargs stream '%s' does not exist",
                          filter_io.xargs);
            return -1;
        }
    }

    if (count == 1) {
        /*
         * Let any traditional fglob options work as filters.
         */
        if (filterGetFGlobFilters()) {
            return -1;
        }
    }

    /* Check if fglob args were given */
    rv = fglobValid();
    if (rv == -1) {
        /* error with fglob options */
        return -1;
    }
    if (rv) {
        ++count;
    }

    return count;
}


/*
 *  stream_count = filterCheckOutputs()
 *
 *    Do basic checks for the output and open some outputs.  Make
 *    certain only one stream uses stdout.  Open the statistics and
 *    all-dest handles.
 *
 *    Returns -1 on error, or the number of output streams requested.
 *    Note that for this count --dry-run and a WILLPROCESS dynamic
 *    library (the global dlISP) are considered outputs.
 */
static int filterCheckOutputs(void)
{
    int count = 0;
    int stdout_used = 0;
    int i;
    int j;
    int rv;

    /* If the dynamic library is taking over all processing, warn if
     * --pass or --fail was also given. */
    if (dlISP && DYNLIB_WILLPROCESS == dynlibGetStatus(dlISP)) {
        ++count;
        if (filter_io.dest_count[DEST_PASS]) {
            skAppPrintErr("warning: dynamic library overriding "
                          "--pass-destinations");
            for (j = 0; j < filter_io.dest_count[DEST_PASS]; ++j) {
                filter_io.dest_stream[DEST_PASS][j].fname = NULL;
            }
            filter_io.dest_count[DEST_PASS] = 0;
        }
        if (filter_io.dest_count[DEST_FAIL]) {
            skAppPrintErr("warning: dynamic library overriding "
                          "--fail-destinations");
            for (j = 0; j < filter_io.dest_count[DEST_FAIL]; ++j) {
                filter_io.dest_stream[DEST_FAIL][j].fname = NULL;
            }
            filter_io.dest_count[DEST_FAIL] = 0;
        }
        if (filter_io.stats.fp || filter_io.stats.fname) {
            skAppPrintErr("warning: dynamic library overriding "
                          "--print-statistics");
            filter_io.stats.fname = NULL;
            filter_io.stats.fp = NULL;
        }
    }

    /* Basic checks: Only allow one output stream to use 'stdout'.
     * Make certain no one uses 'stderr'.  Make certain files either
     * do not exist or have no content if they exist.  For the binary
     * outputs, make certain the stream is not a terminal. */
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            if (0==strcmp("stderr", filter_io.dest_stream[i][j].fname)){
                skAppPrintErr("stderr is not a valid output device. Abort.");
                return -1;
            } else if (0==strcmp("stdout", filter_io.dest_stream[i][j].fname)){
                if (FILEIsATty(stdout)) {
                    /* Ugh!  Hack alert!!  Ignore result of this test
                     * under MPI, since it is always TRUE. */
                    skAppPrintErr("stdout is connected to a terminal. Abort.");
                    return -1;
                }
                if (stdout_used) {
                    skAppPrintErr("stdout is already allocated. Abort.");
                    return -1;
                }
                stdout_used = 1;
            } else if (fileExists(filter_io.dest_stream[i][j].fname)
                       && !isFIFO(filter_io.dest_stream[i][j].fname))
            {
                /* file exists and is regular file: error */
                skAppPrintErr(("Output file '%s' exists.\n"
                               "If you really want to overwrite this file, "
                               "remove it manually and then re-run %s."),
                              filter_io.dest_stream[i][j].fname, skAppName());
                return -1;
            }
        }
    }

    /* Check the STATISTICS stream: increment the output count and
     * check whether stdout is already used if the stats are going to
     * stdout. */
    if (filter_io.stats.fname) {
        ++count;
        if (strcmp(filter_io.stats.fname, "stdout") == 0) {
            if (stdout_used) {
                skAppPrintErr("stdout is already allocated. Abort.");
                return -1;
            }
        }
    }
    /* Actually open the STATISTICS stream unless this is a dry-run */
    if (filter_io.stats.fname
        && !filter_io.dryrun_fp)
    {
        if (strcmp(filter_io.stats.fname, "stdout") == 0) {
            /* we checked for stdout already used above */
            filter_io.stats.fp = stdout;
        } else if (strcmp(filter_io.stats.fname, "stderr") == 0) {
            filter_io.stats.fp = stderr;
        } else {
            /* try to open the file */
            filter_io.stats.fp = fopen(filter_io.stats.fname, "w");
            if ( filter_io.stats.fp == (FILE *)NULL) {
                skAppPrintErr("Can't open file '%s' for printstats output",
                              filter_io.stats.fname);
                return -1;
            }
            filter_io.stats.closefn = fclose;
        }
    }

    /* Open all the record output streams */
    for (i = 0; i < DESTINATION_TYPES; ++i) {
        for (j = 0; j < filter_io.dest_count[i]; ++j) {
            /* create and open. */
            rwIOStruct *rwIOS;
            rv = rwioCreate(&rwIOS, filter_io.dest_stream[i][j].fname,
                            SK_RWIO_WRITE);
            if (rv == LIBRW_OK) {
                rv = rwioSetFileType(rwIOS, FT_RWFILTER);
            }
            if (rv == LIBRW_OK) {
                rv = rwioSetCompression(rwIOS, comp_method);
            }

            if (!filter_io.dryrun_fp) {
                if (rv == LIBRW_OK) {
                    rv = rwioOpen(rwIOS);
                }
            }
            if (rv) {
                rwioPrintLastErr(rwIOS, rv, &skAppPrintErr);
                rwioDestroy(&rwIOS);
                closeAllDests();
                return -1;
            }
            filter_io.dest_stream[i][j].ios = rwIOS;
            ++count;
        }
    }

    return count;
}


/*
 *  count = filterSetCheckers();
 *
 *    Set the array of function pointers to the pass/fail checking
 *    routines, and return the number of pointers that were set.  A
 *    return code of 0 means no filtering rules were specified.
 */
static int filterSetCheckers(void)
{
    int j;
    int count = 0;

    if (filterGetCheckCount() > 0) {
        checker[count] = (int (*)(rwRecPtr))filterCheck;
        ++count;
    }

    for (j = 0; j < app_dynlib_count; ++j) {
        if (dynlibCheckActive(app_dynlib_list[j])) {
            checker[count] = dynlibGetRWProcessor(app_dynlib_list[j]);
            ++count;
        }
    }

    if (dynlibCheckLoaded(dlISP)) {
        checker[count] = dynlibGetRWProcessor(dlISP);
        ++count;
    }

    return count;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
