/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Simple app to the the cfg_* code.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: testTraffic.c 7345 2007-05-30 20:39:09Z mthomas $");

#include "cfg_config.h"
#include "utils.h"

/* extern int yydebug; */


static char *prog_name = NULL;



static cfg_err_t help(FILE* out)
{
    CFG_FPRINTF(( out, "Usage: %s [switches] <command> [command-args]\n",
                  prog_name));
    CFG_FPRINTF(( out, "Switches: \n"));
    CFG_FPRINTF(( out, "  -f <file>  read config from <file>; Def: stdin\n"));
    CFG_FPRINTF(( out, "  -r         pause() program after reading config\n"));
    CFG_FPRINTF(( out, "  -e         pause() program just before exit.\n"));

    CFG_FPRINTF(( out, "Valid commands: \n"));

    CFG_FPRINTF(( out, "\tgetTraffic < <file> validate\n"));
    CFG_FPRINTF(( out, "\t\tRead config file & print diagnostics.\n"));

    CFG_FPRINTF(( out, "\tgetTraffic < <file> config\n"));
    CFG_FPRINTF(( out, "\t\tPrint configuration.\n"));

    CFG_FPRINTF(( out, "\tgetTraffic < <file> property <property>\n"));
    CFG_FPRINTF(( out, "\t\tPrint value of property.\n"));

    CFG_FPRINTF(( out, "\tgetTraffic < <file> domain <domain>\n"));
    CFG_FPRINTF(( out, "\t\tPrint contents of domain.\n"));

    CFG_FPRINTF(( out, "\tgetTraffic < <file> path <domain> <path>\n"));
    CFG_FPRINTF(( out, "\t\tPrint contents of path in domain.\n"));

    CFG_FPRINTF(( out, "\tgetTraffic < <file> path-time <domain> <path> <year> <month> <day> <hour>\n"));
    CFG_FPRINTF(( out, "\t\tPrint contents of path in domain at the specific time.\n"));

    CFG_FPRINTF(( out, "\tgetTraffic help\n"));
    CFG_FPRINTF(( out, "\t\tPrint help info.\n"));
    return CFG_OK;
}

static cfg_err_t invalid(FILE* out, int argc, char** argv)
{
    int i;
    CFG_FPRINTF(( out, "Invalid parameters:"));
    for (i = 0; i < argc; i++) {
        CFG_FPRINTF(( out, " %s", argv[i]));
    }
    CFG_FPRINTF(( out, "\n"));
    return help( out);
}

static cfg_err_t validate(cfg_config_t UNUSED(*config), int UNUSED(argc), char UNUSED(**argv))
{
    /* nothing to do */
    return CFG_OK;
}

static cfg_err_t print_config(cfg_config_t *config, int UNUSED(argc), char UNUSED(**argv))
{
    CFG_CHECK( cfg_config_print( stdout, config, 0));
    return CFG_OK;
}

static cfg_err_t property(cfg_config_t *config, int argc, char **argv)
{
    char* property = argv[argc-1];
    const char* contents;
    contents = cfg_config_get_property( config, property);
    if (contents == NULL) {
        CFG_FPRINTF(( stderr, "No property %s: %s\n",
                      property, cfg_error_msg(CFG_NOT_FOUND)));
    } else {
        CFG_FPRINTF(( stdout, "Property %s has value: %s\n",
                      property, contents));
    }
    return CFG_OK;
}

static cfg_err_t domain(cfg_config_t *config, int argc, char **argv)
{
    char* domain = argv[argc-1];
    cfg_map_t* contents;
    contents = cfg_lookup_config_domain( config, domain);
    if (contents == NULL) {
        CFG_FPRINTF(( stderr, "No domain %s: %s\n",
                      domain, cfg_error_msg(CFG_NOT_FOUND)));
    } else {
        CFG_FPRINTF(( stdout, "Domain %s has value:\n", domain));
        CFG_CHECK( cfg_map_path_print( stdout, contents, 0));
        CFG_FPRINTF(( stdout, "\n"));
    }
    return CFG_OK;
}

static cfg_err_t path(cfg_config_t *config, int argc, char **argv)
{
    char* domain = argv[argc-2];
    char* path = argv[argc-1];
    const cfg_array_t* contents;
    contents = cfg_config_lookup( config, domain, path);
    if (contents == NULL) {
        CFG_FPRINTF((stderr,
                     "No path-group variable %s in domain %s: %s\n",
                     path, domain, cfg_error_msg(CFG_NOT_FOUND)));
    } else {
        CFG_FPRINTF((stdout,
                     "Path-group variable %s in domain %s has value:\n",
                     path, domain));
        CFG_CHECK( cfg_array_print( stdout, contents, 0, cfg_path_print));
        CFG_FPRINTF(( stdout, "\n"));
    }

    return CFG_OK;
}

static cfg_err_t path_time(cfg_config_t *config, int argc, char **argv)
{
    int argi = argc - 7;
    char* domain = argv[++argi];
    char* path = argv[++argi];
    int year = atoi( argv[++argi]);
    int month = atoi( argv[++argi]);
    int day = atoi( argv[++argi]);
    int hour = atoi( argv[++argi]);
    cfg_array_t* contents;
    struct tm t;

    t.tm_year = year - 1900;
    t.tm_mon = month - 1;
    t.tm_mday = day;
    t.tm_hour = hour;

    contents = cfg_config_lookup_date( config, domain, path,
                                       timegm(&t));
    if (contents == NULL) {
        CFG_FPRINTF(( stderr, "No path-group variable %s in domain %s on %.2d/%.2d/%.2d:%.2d: %s\n",
                      path, domain, year, month, day, hour,
                      cfg_error_msg(CFG_NOT_FOUND)));
    } else {
        CFG_FPRINTF(( stdout, "Path-group variable %s in domain %s on %.2d/%.2d/%.2d:%.2d has value:\n",
                      path, domain, year, month, day, hour));
        CFG_CHECK( cfg_array_print( stdout, contents, 0, cfg_path_print));
        CFG_FPRINTF(( stdout, "\n"));
        CFG_CHECK( cfg_array_path_destroy( contents));
    }
    return CFG_OK;
}


int main(int argc, char** argv)
{
    char* cmd;
    cfg_config_t* config = NULL;
    FILE *fp = stdin;
    int opt;
    int pause_at = 0;
    cfg_err_t (*fn)(cfg_config_t *config, int argc, char **argv) = NULL;

    prog_name = argv[0];

    while (-1 != (opt = getopt(argc, argv, "ref:"))) {
        switch (opt) {
          case 'f':
            /* input file */
            fp = fopen(optarg, "r");
            if (fp == NULL) {
                fprintf(stderr, "Could not open file '%s'\n", optarg);
                exit(EXIT_FAILURE);
            }
            break;

          case 'r':
          case 'e':
            /* pause() after reading file 'r' or at exit 'e' */
            pause_at = opt;
            break;

          case '?':
          default:
            help(stdout);
            exit(EXIT_FAILURE);
        }
    }

    argc -= optind;
    argv += optind;

    /* user's command; not program name */
    cmd = argv[0];

    if (argc == 0) {
        invalid( stderr, argc, argv);
    } else if (!strcmp( cmd, "help") && argc == 1) {
        help( stdout);
    } else if (!strcmp( cmd, "validate") && argc == 1) {
        fn = &validate;
    } else if (!strcmp( cmd, "config") && argc == 1) {
        fn = &print_config;
    } else if (!strcmp( cmd, "property") && argc == 2) {
        fn = &property;
    } else if (!strcmp( cmd, "domain") && argc == 2) {
        fn = &domain;
    } else if (!strcmp( cmd, "path") && argc == 3) {
        fn = &path;
    } else if (!strcmp( cmd, "path-time") && argc == 7) {
        fn = &path_time;
    } else {
        invalid( stderr, argc, argv);
    }

    if (fn != NULL) {
        /* For testing purposes, ignore any error from cfg_config_read */

        /* yydebug = 1; */

        CFG_CHECK( cfg_config_create( &config));
        cfg_config_read( config, fp, NULL);
        if (pause_at == 'r') {
            pause();
        }
        fn(config, argc, argv);
        if (config != NULL) {
            CFG_CHECK( cfg_config_destroy( config));
        }
    }

    if (pause_at == 'e') {
        pause();
    }

    return CFG_OK;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
