/*
** Copyright (C) 2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: rwipaimport.c 7494 2007-06-13 14:58:02Z mthomas $");

#include "rwipa.h"

/* LOCAL DEFINES AND TYPEDEFS */

/* where to write output from --help */
#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

/* Holds the appOptionsEnum value for the operation being executed. */
static int catalog_type = IPA_CAT_NONE;

/* Name of the IPA catalog to import to */
static char *catalog_name = NULL;

/* A short description of the catalog */
static char *catalog_desc = NULL;

/* Date/time string specifying beginning of validity interval */
static char *start_time_str = NULL;

/* Date/time string specifying end of validity interval */
static char *end_time_str = NULL;

/* Start and end times are converted into struct timeval */
struct timeval start_time, end_time;

/* index of first option that is not handled by the options handler. */
static int arg_index = 0;


/* OPTIONS SETUP */

typedef enum {
    OPT_CATALOG_NAME,
    OPT_TYPE,
    OPT_CATALOG_DESC,
    OPT_START_TIME,
    OPT_END_TIME
} appOptionsEnum;

static struct option appOptions[] = {
    {"catalog",         REQUIRED_ARG, 0, OPT_CATALOG_NAME},
    {"type",            REQUIRED_ARG, 0, OPT_TYPE},
    {"description",     REQUIRED_ARG, 0, OPT_CATALOG_DESC},
    {"start-time",      REQUIRED_ARG, 0, OPT_START_TIME},
    {"end-time",        REQUIRED_ARG, 0, OPT_END_TIME},
    {0, 0, 0, 0}        /* sentinel entry */
};

static const char *appHelp[] = {
    ("Import the data into the named IPA catalog; the catalog\n"
     "\twill be created if necessary"),
    "Specify the type of the IPA catalog; one of set, bag, or pmap",
    ("Describe the catalog's contents (for new catalogs)"),
    ("Specify the time when the data is first valid, in\n"
     "\tYYYY/MM/DD[:HH[:MM[:SS]]] format. Def. None.  Requires --end-time"),
    ("Specify end of validity interval. Def. None"),
    (char *) NULL
};


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                       \
    ("--catalog=CATALOG --type=TYPE [SWITCHES] INPUT_FILE\n"            \
     "\tImport a SiLK IPSet, Bag, or Prefix Map from the named\n"       \
     "\tINPUT_FILE to an IP Address Association (IPA) catalog.\n")

    FILE *fh = USAGE_FH;

    skAppStandardUsage(fh, USAGE_MSG, appOptions, appHelp);
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    skAppUnregister();
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    /* verify same number of options and help strings */
    assert((sizeof(appHelp) / sizeof(char *)) ==
           (sizeof(appOptions) / sizeof(struct option)));

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* register the options */
    if (optionsRegister(appOptions, (optHandler) appOptionsHandler, NULL)) {
        skAppPrintErr("unable to register options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    arg_index = optionsParse(argc, argv);
    assert(arg_index <= argc);
    if (arg_index < 0) {
        /* optionsParse() printed the error */
        skAppUsage();                            /* never returns */
    }

    /* verify that we have something to do */
    if (catalog_type == IPA_CAT_NONE) {
        skAppPrintErr("A valid catalog type must be specified");
        skAppUsage();
    }

    /* Make sure there's a filename to process */
    if (arg_index == argc) {
        skAppPrintErr("No input file given on the command line");
        skAppUsage();
    }

    /* A catalog name must be specified */
    if (catalog_name == NULL) {
        skAppPrintErr("You must specify a catalog name with the --%s option",
                      appOptions[OPT_CATALOG_NAME].name);
        skAppUsage();
    }

    /* A time period (start time and end time) is optional, but if present,
     *  both must be specified */
    if ( (start_time_str == NULL) ^ (end_time_str == NULL)  ) {
        skAppPrintErr(("Incomplete time range specified."
                       "  If the imported data is\n"
                       "\tassociated with specific dates,"
                       " you must specify both the\n"
                       "\t--%s and --%s options"),
                      appOptions[OPT_START_TIME].name,
                      appOptions[OPT_END_TIME].name);
        skAppUsage();
    } else if ( (start_time_str != NULL) && (end_time_str != NULL) ) {
        if (skStringParseDatetime(&start_time, start_time_str, NULL)) {
            skAppPrintErr("Error parsing --%s argument %s",
                          appOptions[OPT_START_TIME].name,
                          start_time_str);
            skAppUsage();
        }
        if (skStringParseDatetime(&end_time, end_time_str, NULL)) {
            skAppPrintErr("Error parsing --%s argument %s",
                          appOptions[OPT_END_TIME].name,
                          end_time_str);
            skAppUsage();
        }
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to optionsRegister(); it will be called
 *    by optionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes optionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ((appOptionsEnum) opt_index) {
      case OPT_CATALOG_NAME:
        catalog_name = opt_arg;
        break;

      case OPT_TYPE:
        if (!strcasecmp(opt_arg, "set")) {
            catalog_type = IPA_CAT_SET;
        } else if (!strcasecmp(opt_arg, "bag")) {
            catalog_type = IPA_CAT_BAG;
        } else if (!strcasecmp(opt_arg, "pmap")) {
            catalog_type = IPA_CAT_PMAP;
        } else {
            skAppPrintErr("Could not parse %s value %s",
                          appOptions[opt_index].name, opt_arg);
            return FALSE;
        }
        break;

      case OPT_CATALOG_DESC:
        catalog_desc = opt_arg;
        break;

      case OPT_START_TIME:
        start_time_str = opt_arg;
        break;

      case OPT_END_TIME:
        end_time_str = opt_arg;
        break;
    }

    return 0;                   /* OK */
}


static int import_set(IPAContext *ipa, const char *filename)
{
    skstream_t *stream = NULL;
    skIPTree_t *set = NULL;
    skIPTreeCIDRBlockIterator_t cidr_iter;
    skIPTreeCIDRBlock_t cidr;
    int rv = 0;

    /* Open the IPSet fle */
    if ((rv = skStreamCreate(&stream, SK_IO_READ, SK_CONTENT_SILK))
        || (rv = skStreamBind(stream, filename))
        || (rv = skStreamOpen(stream)))
    {
        skStreamPrintLastErr(stream, rv, &skAppPrintErr);
        rv = -1;
        goto done;
    }

    /* Read ipTree from file */
    if (0 != skIPTreeRead(stream, &set)) {
        skAppPrintErr("Error reading binary IPset from %s",
                      skStreamGetPathname(stream));
        rv = -1;
        goto done;
    }

    skIPTreeCIDRBlockIteratorBind(&cidr_iter, set);

    while (skIPTreeCIDRBlockIteratorNext(&cidr, &cidr_iter)==SK_ITERATOR_OK) {
        ipa_add_cidr(ipa, cidr.addr, cidr.mask, 0);
    }

  done:
    skStreamDestroy(&stream);
    skIPTreeDelete(&set);
        
    return rv;
}


static int import_bag(IPAContext *ipa, const char *filename)
{
    skstream_t *stream = NULL;
    skBag_header_t *bag = NULL;
    skBag_iterator_t *bag_iter = NULL;
    skBag_key_t key;
    skBag_counter_t count;

    int rv = 0;

    /* Open the Bag fle */
    if ((rv = skStreamCreate(&stream, SK_IO_READ, SK_CONTENT_SILK))
        || (rv = skStreamBind(stream, filename))
        || (rv = skStreamOpen(stream)))
    {
        skStreamPrintLastErr(stream, rv, &skAppPrintErr);
        skAppPrintErr("Could not open Bag input stream");
        rv = -1;
        goto done;
    }

    /* Read Bag from file */
    if (skBag_readBinary(&bag, stream) != SKBAG_OK) {
        skAppPrintErr("Error reading Bag from file %s",
                      skStreamGetPathname(stream));
        rv = -1;
        goto done;
    }

    if (skBag_allocIterator(bag, &bag_iter) != SKBAG_OK) {
        skAppPrintErr("Could not create Bag iterator");
        rv = -1;
        goto done;
    }

    while (SKBAG_OK == skBag_getNext(bag_iter, &key, &count)) {
        /* FIXME: do we want to CIDRize if adjacent counts are the same? */
        ipa_add_cidr(ipa, key, 32, count);
    }

  done:
    if (bag_iter) {
        skBag_freeIterator(bag_iter);
    }
    if (bag) {
        skBag_free(bag);
    }
    skStreamDestroy(&stream);
    return rv;
}


/*
 *  pmap = openMapFile(path);
 *
 *    Open the prefix map file at 'path' and verify that it can be
 *    processed by this program.
 */
static skPrefixMap_t *openMapFile(const char *map_file)
{
    skstream_t *stream = NULL;
    skPrefixMap_t *map = NULL;
    skPrefixMap_err_t map_error;
    int rv;

    if ((rv = skStreamCreate(&stream, SK_IO_READ, SK_CONTENT_SILK))
        || (rv = skStreamBind(stream, map_file))
        || (rv = skStreamOpen(stream)))
    {
        skStreamPrintLastErr(stream, rv, &skAppPrintErr);
        goto ERROR;
    }

    /* read in the prefixmap file */
    map_error = skPrefixMapRead(&map, stream);
    skStreamDestroy(&stream);

    if (SKPREFIXMAP_OK != map_error) {
        if (SKPREFIXMAP_ERR_ARGS == map_error) {
            skAppPrintErr("Failed to read pmap file '%s': Invalid arguments",
                          map_file);
        } else if (SKPREFIXMAP_ERR_MEMORY == map_error) {
            skAppPrintErr("Failed to read pmap file '%s': Out of memory",
                          map_file);
        } else if (SKPREFIXMAP_ERR_IO == map_error) {
            skAppPrintErr("Failed to read pmap file '%s': I/O error",
                          map_file);
        }
        map = NULL;
        goto ERROR;
    }

    /* verify that the prefix map is of the correct generation */
    if (skPrefixMapGetDictionaryWordCount(map) == 0) {
        skAppPrintErr("The pmap file'%s' cannot be processed by this program",
                      map_file);
        goto ERROR;
    }

    /* complain if ranges were requested and the file does not contain
     * address data.  if only ranges were requested, destroy the map. */
    if (SKPREFIXMAP_CONT_ADDR != skPrefixMapGetContentType(map)) {
        skAppPrintErr(("The pmap file '%s'"
                       " does not contain an address prefix map"),
                      map_file);
    }

    /* Success */
    return map;

  ERROR:
    if (map) {
        skPrefixMapDelete(map);
    }
    skStreamDestroy(&stream);
    return NULL;
}


static void add_pmap_labels(IPAContext *ipa, skPrefixMap_t *pmap)
{
    char label[1024];
    uint32_t label_count;
    uint32_t i;

    label_count = skPrefixMapGetDictionaryWordCount(pmap);
    for (i = 0; i < label_count; ++i) {
        skPrefixMapGetStringForVal(label, sizeof(label), pmap, i);
        ipa_add_label(ipa, label, (uint64_t) i);
    }
}


static void add_pmap_ranges(IPAContext *ipa, skPrefixMap_t *pmap)
{
    skPrefixMapIterator_t iter;
    uint32_t start_addr;
    uint32_t end_addr;
    uint32_t val;

    skPrefixMapIteratorBind(&iter, pmap);
    while (skPrefixMapIteratorNext(&iter, &start_addr, &end_addr, &val)
           != SK_ITERATOR_NO_MORE_ENTRIES)
    {
        ipa_add_range(ipa, start_addr, end_addr, val);
    }
}


static int import_pmap(IPAContext *ipa, const char *filename)
{
    skPrefixMap_t *pmap = NULL;

    /* open the prefix map file */
    pmap = openMapFile(filename);
    if (pmap == NULL) {
        exit(EXIT_FAILURE);
    }

    add_pmap_labels(ipa, pmap);

    add_pmap_ranges(ipa, pmap);

    return 0;
}


int main(int argc, char **argv)
{
    const char *filename = NULL;
    int rv = 1;
    char *ipa_db_url = NULL;
    IPAContext *ipa;

    appSetup(argc, argv);       /* never returns on error */

    filename = argv[arg_index];

#if 0
    /* FIXME: Auto-detection of file format (some day)... */
    rv = skStreamReadSilkHeader(stream, &file_header, sizeof(file_header));
    if (rv) {
        skStreamPrintLastErr(stream, rv, &skAppPrintErr);
        rv = -1;
        goto done;
    }

    format = skStreamGetSilkFormat(stream);
#endif

    ipa_db_url = get_ipa_config();

    if (ipa_db_url == NULL) {
        skAppPrintErr("Could not get IPA configuration");
        rv = EXIT_FAILURE;
        goto done;
    }

    ipa_create_context(&ipa, ipa_db_url);

    if (ipa == NULL) {
        skAppPrintErr("Could not create IPA context");
        rv = EXIT_FAILURE;
        goto done;
    }

    ipa_begin(ipa);

    if (ipa_add_dataset(ipa, catalog_name, catalog_desc, catalog_type,
                        start_time.tv_sec, end_time.tv_sec) != IPA_OK) {
        rv = -1;
        goto done;
    }


    switch (catalog_type) {
      case IPA_CAT_SET:
        rv = import_set(ipa, filename);
        break;

      case IPA_CAT_BAG:
        rv = import_bag(ipa, filename);
        break;

      case IPA_CAT_PMAP:
        rv = import_pmap(ipa, filename);
        break;
    }

    if (!rv) {
        ipa_commit(ipa);
    } else {
        skAppPrintErr("Warning: rolling back IPA transaction");
        ipa_rollback(ipa);
    }

  done:
    if (ipa_db_url) {
        free(ipa_db_url);
    }
    appTeardown();
    return rv;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
