/*
** Copyright (C) 2005-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
** rwmatch
**
** rwmatch.c is a utility which takes two files (now called in and
** out) and produces a new file where the contents are linked together
** in the form (saddress, values) on one side and (daddress, values)
** on the other side.  apart from the saddress/daddress relationship,
** any other mating rules are specified by the command line.
**
** This is necessary because there are several operations where mating
** is not direct reversal; the particular case which is driving this
** applicatin is traceroute where the only rational is close
** relationship in time between s and d address, and one of them is a
** UDP signal with an ICMP response
**
** command arguments:
**
** --relate:    takes fields as pairs for relationship.  The first
**              value is the field in file 1, the second value is the
**              field in file ** 2.
**
**              for example, the default could be --relate=1,2
**
**              multiple --relate predicates are accepted.
**              Traditional tcp/ip matching would be:
**              --relate=1,2 --relate=2,1 --relate=3,4 --relate=4,3
**
** --delta:     the number of seconds difference we'll allow between
**              two records before considering them similar.  Defaults
**              to 0.
**
** --unmated:   output path for unmatched records; defaults to
**              /dev/null.  A record must have a response before it is
**              ejected.  So we're always sending out at least two
**              records.
**
** Apart from that; rwmate will annotate the nhip field with an index
** indicating the 'mating flow' that a record belongs to.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwmatch.c 6075 2007-01-22 18:59:53Z mthomas $");

#include "utils.h"
#include "rwpack.h"
#include "sksite.h"


/* TYPEDEFS AND DEFINES */

/* Where to send --help output */
#define USAGE_FH stdout

#define RWMATE_REL_MAX 8
#define RWMATE_REL_MAX_STR "8"

#define RWMATE_QUERY 0
#define RWMATE_RESPONSE 1


typedef enum _fieldsEnum {
    RWMATE_FID_SIP = 1, RWMATE_FID_DIP, RWMATE_FID_SPORT, RWMATE_FID_DPORT,
    RWMATE_FID_PROTO, RWMATE_FID_PKTS, RWMATE_FID_BYTES, RWMATE_FID_FLAGS,
    RWMATE_FID_STIME, RWMATE_FID_ETIME
} fields_enum;

typedef struct _appState {
    const char *query_fname;
    const char *response_fname;
    const char *matched_fname;
    rwIOStruct *query_ptr;
    rwIOStruct *response_ptr;
    rwIOStruct *matched_ptr;
} rwm_app_state_t;


/* LOCAL FUNCTIONS */

static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);    /* never returns */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static uint32_t getField(fields_enum field_id, const rwRec *current_rec);
static int  compareRecords(const rwRec *query_rec, const rwRec *response_rec);


/* LOCAL VARIABLES */

static uint8_t rwm_total_pairs;
static uint8_t rwm_mate_pairs[RWMATE_REL_MAX][2];
static uint8_t verbose_flag = 0;
static uint32_t rwm_delta = 0;
static rwm_app_state_t app_state;


/* OPTIONS SETUP */

typedef enum _appOptionsEnum {
    OPT_RELATE, OPT_TIME_DELTA, OPT_UNMATCHED,
    OPT_VERBOSE
} appOptionsEnum;

static struct option appOptions[] = {
    {"relate",          REQUIRED_ARG, 0, OPT_RELATE},
    {"time-delta",      REQUIRED_ARG, 0, OPT_TIME_DELTA},
    {0,0,0,0}           /* sentinel entry */
};

static const char *appHelp[] = {
    ("A pair of fields to relate when matching data, given in\n"
     "\tthe form '<query field id>,<response field id>'.  The switch may\n"
     "\tbe repeated up to " RWMATE_REL_MAX_STR " times.  Field IDs:\n"
     "\t    1 source IP          5 protocol    \n"
     "\t    2 destination IP     6 packets     \n"
     "\t    3 source port        7 bytes       \n"
     "\t    4 destination port   8 TCP flags   "),
    ("Maximum time difference (in seconds) permitted between\n"
     "\ttwo records for relation. Def. 0"),
    (char *)NULL
};


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                       \
    ("--relate=FIELD_PAIR QUERY_FILE RESPONSE_FILE MATCHED_PATH\n"      \
     "\tGroups records as queries and responses\n")

    FILE *fh = USAGE_FH;

    skAppStandardUsage(fh, USAGE_MSG, appOptions, appHelp);
    sksiteOptionsUsage(fh);
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int arg_index;
    int rv;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize globals */
    memset(&app_state, '\0', sizeof(rwm_app_state_t));
    rwm_total_pairs = 0;
    verbose_flag = 1;

    /* register the options */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)
        || sksiteOptionsRegister(SK_SITE_FLAG_CONFIG_FILE))
    {
        skAppPrintErr("unable to register options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    arg_index = optionsParse(argc, argv);
    if (arg_index < 0) {
        skAppUsage();             /* never returns */
    }

    /* try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names */
    sksiteConfigure(0);

    /* verify that related pairs were given */
    if (rwm_total_pairs == 0) {
        skAppPrintErr("at least one --%s pair must be given",
                      appOptions[OPT_RELATE].name);
        skAppUsage();             /* never returns */
    }

    /* get the file arguments */
    if (arg_index == argc) {
        skAppPrintErr("missing QUERY_FILE argument");
        skAppUsage();             /* never returns */
    }
    app_state.query_fname = argv[arg_index];
    arg_index++;

    if (arg_index == argc) {
        skAppPrintErr("missing RESPONSE_FILE argument");
        skAppUsage(); /* never returns */
    }
    app_state.response_fname = argv[arg_index];
    arg_index++;

    if (arg_index == argc) {
        skAppPrintErr("missing MATCHED_FILE argument");
        skAppUsage(); /* never returns */
    }
    app_state.matched_fname = argv[arg_index];
    arg_index++;


    /* check for extra options */
    if (arg_index != argc) {
        skAppPrintErr("too many arguments or unrecognized switch '%s'",
                      argv[arg_index]);
        skAppUsage(); /* never returns */
    }


    /* Now, open the files */
    if ((rv = rwioCreate(&app_state.query_ptr, app_state.query_fname,
                         SK_RWIO_READ))
        || (rv = rwioOpen(app_state.query_ptr)))
    {
        rwioPrintLastErr(app_state.query_ptr, rv, &skAppPrintErr);
        skAppPrintErr("cannot open QUERY_FILE. Exiting.");
        exit(EXIT_FAILURE);
    }

    if ((rv = rwioCreate(&app_state.response_ptr, app_state.response_fname,
                         SK_RWIO_READ))
        || (rv = rwioOpen(app_state.response_ptr)))
    {
        rwioPrintLastErr(app_state.response_ptr, rv, &skAppPrintErr);
        skAppPrintErr("cannot open RESPONSE_FILE. Exiting.");
        exit(EXIT_FAILURE);
    }

    if ((rv = rwioCreate(&app_state.matched_ptr, app_state.matched_fname,
                         SK_RWIO_WRITE))
        || (rv = rwioOpen(app_state.matched_ptr)))
    {
        rwioPrintLastErr(app_state.matched_ptr, rv, &skAppPrintErr);
        skAppPrintErr("cannot open MATCHED_FILE. Exiting.");
        exit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int rv;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    if (app_state.query_ptr != NULL) {
        rv = rwCloseFile(app_state.query_ptr);
        if (rv) {
            rwioPrintLastErr(app_state.query_ptr, rv, &skAppPrintErr);
        }
    }
    if (app_state.response_ptr != NULL) {
        rv = rwCloseFile(app_state.response_ptr);
        if (rv) {
            rwioPrintLastErr(app_state.response_ptr, rv, &skAppPrintErr);
        }
    }
    if (app_state.matched_ptr != NULL) {
        rv = rwCloseFile(app_state.matched_ptr);
        if (rv) {
            rwioPrintLastErr(app_state.matched_ptr, rv, &skAppPrintErr);
        }
    }

    skAppUnregister();
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to optionsRegister(); it will be called
 *    by optionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes optionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t *parsed_list;
    uint32_t count;

    switch ((appOptionsEnum)opt_index) {

      case OPT_RELATE:
        if (rwm_total_pairs == RWMATE_REL_MAX) {
            skAppPrintErr(("Total number of relations specified"
                           " exceeds maximum (%d)"),
                          RWMATE_REL_MAX);
            return 1;
        }
        if (skStringParseNumberList(&parsed_list, &count, opt_arg,
                                    RWMATE_FID_SIP, RWMATE_FID_FLAGS, 2))
        {
            /* could not parse list or too many entries in list */
            return 1;
        }
        assert(count > 0 && count <= 2);
        if (count != 2) {
            skAppPrintErr("each relation requires two IDs. Error parsing '%s'",
                          opt_arg);
            free(parsed_list);
            return 1;
        }
        rwm_mate_pairs[rwm_total_pairs][RWMATE_QUERY] = parsed_list[0];
        rwm_mate_pairs[rwm_total_pairs][RWMATE_RESPONSE] = parsed_list[1];
        rwm_total_pairs++;
        free(parsed_list);
        break;

      case OPT_TIME_DELTA:
        if (skStringParseUint32(&rwm_delta, opt_arg, 0, 0)) {
            skAppPrintErr("error parsing %s: '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case OPT_VERBOSE:
        verbose_flag = 1;
        break;

      case OPT_UNMATCHED:
        skAppPrintErr("unmatched path not yet supported");
        break;
    }

    return 0; /* OK */
}


/*
 * uint32_t getField(fields_enum current_field, rwRec *current_rec)
 *
 * Given a record and a field ID returns the value of the field for
 * that record.
 */
static uint32_t getField(fields_enum field_id, const rwRec *current_rec)
{
    uint32_t result = 0;

    switch (field_id) {
      case RWMATE_FID_SIP:
        result = current_rec->sIP.ipnum;
        break;
      case RWMATE_FID_DIP:
        result = current_rec->dIP.ipnum;
        break;
      case RWMATE_FID_SPORT:
        result = current_rec->sPort;
        break;
      case RWMATE_FID_DPORT:
        result = current_rec->dPort;
        break;
      case RWMATE_FID_PROTO:
        result = current_rec->proto;
        break;
      case RWMATE_FID_PKTS:
        result = current_rec->pkts;
        break;
      case RWMATE_FID_BYTES:
        result = current_rec->bytes;
        break;
      case RWMATE_FID_FLAGS:
        result = current_rec->flags;
        break;
      case RWMATE_FID_STIME:
        result = current_rec->sTime;
        break;
      case RWMATE_FID_ETIME:
        result = current_rec->sTime + current_rec->elapsed;
        break;
    };
    return result;
}


/*
 * int compareRecords(rwRec *propose, rwRec *response)
 *
 * this uses the same hash information from rwm_hash to compare an
 * query and response record in order to determine whether or not.
 * Note the global variable dependence.
 *
 * Returns 1 if the query > outing, -1 if response > query, 0 if
 * equal
 */
static int compareRecords(
    const rwRec    *query_rec,
    const rwRec    *response_rec)
{
    uint8_t i;
    uint32_t qval, rval;

    for (i = 0 ; i < rwm_total_pairs; i++) {
        qval = getField(rwm_mate_pairs[i][RWMATE_QUERY], query_rec);
        rval = getField(rwm_mate_pairs[i][RWMATE_RESPONSE], response_rec);
        if (qval > rval) {
            return 1;
        } else if (qval < rval) {
            return -1;
        };
    };

    /*
     * We successfully matched all match fields that we care about.  Now we
     * compare timing.
     * The response should be within <delta> seconds of the end of the
     * message.
     */
    if (getField(RWMATE_FID_STIME, query_rec)
        > getField(RWMATE_FID_STIME, response_rec))
    {
        return 1;
    };

    /*
     * We know response.etime > outging.stime, and we know by now that
     * query.stime >= response.stime; so we just have to compare etime.
     */
    if ((getField(RWMATE_FID_ETIME, query_rec) + rwm_delta)
        < getField(RWMATE_FID_ETIME, response_rec))
    {
        return -1;
    };

    /*
     * we've passed all tests, so the two must be related.
     */
    return 0;
}


int main(int argc, char **argv)
{
    uint32_t i = 0;
    uint8_t printed_query = 0;
    rwRec query_rec, response_rec;

    appSetup(argc, argv); /* never returns on error */

    /*
     * The revised version of this application requires
     * sorted data that matches the mating.  So this means
     * that we're going to have records sorted slightly
     * differently (if we match 2/1, we sort --field=2,9 and
     * --field=1,9) on the two applications.
     *
     * The loop always begins with us
     * with a reference input record and an output record.
     * as long as the output record < the input record, we continue
     * to loop.
     */
    rwioWriteHeader(app_state.matched_ptr);
    rwRead(app_state.response_ptr, &response_rec);

    while (rwRead(app_state.query_ptr, &query_rec)) {
        /*
         * As long as the query_rec is > response rec,
         * move forward in the response queue.
         */
        while (compareRecords(&query_rec, &response_rec) < 0) {
            if (!rwRead(app_state.query_ptr, &query_rec)) {
                goto terminate;
            }
        }

        /*
         * If they're equal, print out the query records,
         * then all the matching responses
         */
        printed_query = 0;
        while (compareRecords(&query_rec, &response_rec) == 0) {
            if (!printed_query) {
                i++;
                query_rec.nhIP.ipnum = i;
                rwWrite(app_state.matched_ptr, &query_rec);
                printed_query = 1;
            }
            response_rec.nhIP.ipnum = 0xFF000000 | i;
            rwWrite(app_state.matched_ptr, &response_rec);
            if (!rwRead(app_state.response_ptr, &response_rec)) {
                goto terminate;
            }
        }
        /*
         * If the response record is greater, then advance the
         * query pointer
         */
        while (compareRecords(&query_rec, &response_rec) > 0) {
            if (!rwRead(app_state.response_ptr, &response_rec)) {
                goto terminate;
            }
        }
    }

  terminate:
    if (app_state.matched_ptr) {
        rwCloseFile(app_state.matched_ptr);
    }
    /* done */
    appTeardown();

    return 0;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
