/*
** Copyright (C) 2004-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
** 
** Use of the SILK system and related source code is subject to the terms 
** of the following licenses:
** 
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
** 
** NO WARRANTY
** 
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
** DELIVERABLES UNDER THIS LICENSE.
** 
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
** Mellon University, its trustees, officers, employees, and agents from 
** all claims or demands made against them (and any related losses, 
** expenses, or attorney's fees) arising out of, or relating to Licensee's 
** and/or its sub licensees' negligent use or willful misuse of or 
** negligent conduct or willful misconduct regarding the Software, 
** facilities, or other rights or assistance granted by Carnegie Mellon 
** University under this License, including, but not limited to, any 
** claims of product liability, personal injury, death, damage to 
** property, or violation of any laws or regulations.
** 
** Carnegie Mellon University Software Engineering Institute authored 
** documents are sponsored by the U.S. Department of Defense under 
** Contract F19628-00-C-0003. Carnegie Mellon University retains 
** copyrights in all material produced under this contract. The U.S. 
** Government retains a non-exclusive, royalty-free license to publish or 
** reproduce these documents, or allow others to do so, for U.S. 
** Government purposes only pursuant to the copyright license under the 
** contract clause at 252.227.7013.
** 
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwnetmask
**
**  Read in rw-records and output rw-records, masking the Source IP
**  and Destination IP by the prefix-lengths given on the command
**  line.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwnetmask.c 6657 2007-03-14 15:21:52Z mthomas $");

#include "utils.h"
#include "rwpack.h"


/* TYPEDEFS AND DEFINES */

/* File handle for --help output */
#define USAGE_FH stdout


/* LOCAL FUNCTIONS */

static void appUsageLong(void);			/* never returns */
static void appTeardown(void);
static void appSetup(int argc, char **argv);	/* never returns */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);


/* LOCAL VARIABLES */

/* The netmask of source/dest/next-hop IP.  For class A (/8), value
 * would be 0xFF000000. Value of 0 means to apply no mask--logically
 * the same as 0xFFFFFFFF. */
static struct {
    uint32_t  source;
    uint32_t  dest;
    uint32_t  next_hop;
} nmask_prefixes = {
    0, 0, 0
};


/* OPTIONS */

typedef enum {
    OPT_SOURCE, OPT_DEST, OPT_NEXTHOP
}  appOptionsEnum;

static struct option appOptions[] = {
    {"source-prefix-length",      REQUIRED_ARG, 0, OPT_SOURCE},
    {"destination-prefix-length", REQUIRED_ARG, 0, OPT_DEST},
    {"next-hop-prefix-length",    REQUIRED_ARG, 0, OPT_NEXTHOP},
    {0,0,0,0}			  /* sentinel entry */
};

static const char *appHelp[] = {
    /* add help strings here for the applications options */
    "High bits of the sIP to keep; Def 32",
    "High bits of the dIP to keep; Def 32",
    "High bits of the nhIP to keep; Def 32",
    (char *)NULL
};



/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("[--source-prefix=N] [--destination-prefix=N] [--next-hop-prefix=N]\n"   \
     "\tRead SiLK Flow records from STDIN, mask-off the lower (32-N) bits\n"  \
     "\tof the specified IP address(es), and write the resulting records to\n"\
     "\tstdout.  Does not process files.\n")

    FILE *fh = USAGE_FH;

    skAppStandardUsage(fh, USAGE_MSG, appOptions, appHelp);
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int arg_index;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* register the options */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL))
    {
        skAppPrintErr("unable to register options");
        exit(EXIT_FAILURE);
    }

    /* parse options */
    arg_index = optionsParse(argc, argv);
    if (arg_index < 0) {
        skAppUsage();		/* never returns */
    }

    /* check for extra options */
    if (arg_index > argc) {
        skAppPrintErr("unexpected arguments; input must come from stdin.");
        skAppUsage();
    }

    /* make certain at least one mask was specified */
    if ((0 == nmask_prefixes.source) && (0 == nmask_prefixes.dest)
        && (0 == nmask_prefixes.next_hop))
    {
        skAppPrintErr("must specify at least one prefix length option");
        skAppUsage();
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;			/* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    skAppUnregister();
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to optionsRegister(); it will be called
 *    by optionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes optionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    uint32_t n;
    int rv;
    /* address of prefix to set */
    uint32_t *netmask_p = NULL;

    /* Store the address of the prefix to set in the switch(), then
     * set the referenced value below the switch() */
    switch ((appOptionsEnum)opt_index) {
      case OPT_SOURCE:
        if (nmask_prefixes.source) {
            skAppPrintErr("warning: multiple source lengths given;"
                          " using final");
        }
        netmask_p = &nmask_prefixes.source;
        break;

      case OPT_DEST:
        if (nmask_prefixes.dest) {
            skAppPrintErr("warning: multiple destination lengths given;"
                          " using final");
        }
        netmask_p = &nmask_prefixes.dest;
        break;

      case OPT_NEXTHOP:
        if (nmask_prefixes.next_hop) {
            skAppPrintErr("warning: multiple next-hop lengths given;"
                          " using final");
        }
        netmask_p = &nmask_prefixes.next_hop;
        break;
    }

    /* we know the index is good, now set the value */
    assert(netmask_p);

    /* Parse it */
    rv = skStringParseUint32(&n, opt_arg, 1, 32);
    if (rv < -10) {
        skAppPrintErr("Prefix %lu out of range; use 1 <= n <= 32",
                      (unsigned long)n);
        return 1;
    }
    if (rv != 0) {
        skAppPrintErr("Cannot parse '%s' as a number", opt_arg);
        return 1;
    }

    *netmask_p = ~0 << (32-n);

    return 0; /* OK */
}


/*
 *  maskInput(input_file, output_file);
 *
 *    Read rw-records from the 'input_file', mask off the source,
 *    destination, and/or next-hop IP addresses, and print the records
 *    to the 'output_file'.
 */
static int maskInput(
    const char *input_file,
    const char *output_file)
{
    rwIOStruct_t *input_ios;
    rwIOStruct_t *output_ios;
    rwRec rwrec;
    int rv = LIBRW_OK;

    /* Open input */
    input_ios = rwOpenFile(input_file, NULL);
    if ( !input_ios) {
        return 1;
    }

    /* Open the output file and copy the headers from the source file */
    if ((rv = rwioCreate(&output_ios, output_file, SK_RWIO_WRITE))
        || (rv = rwioSetHeaderFromFile(output_ios, input_ios))
        || (rv = rwioOpen(output_ios)))
    {
        goto END;
    }

    /* read the records and mask the IP addresses */
    while (!LIBRW_ERROR_IS_FATAL(rv) && rwRead(input_ios, &rwrec)) {
        if (nmask_prefixes.source) {
            rwrec.sIP.ipnum &= nmask_prefixes.source;
        }
        if (nmask_prefixes.dest) {
            rwrec.dIP.ipnum &= nmask_prefixes.dest;
        }
        if (rwrec.nhIP.ipnum && nmask_prefixes.next_hop) {
            rwrec.nhIP.ipnum &= nmask_prefixes.next_hop;
        }

        rv = rwWrite(output_ios, &rwrec);
    }

 END:
    if (rv) {
        rwioPrintLastErr(output_ios, rv, &skAppPrintErr);
    }
    if (output_ios) {
        rwCloseFile(output_ios);
    }
    if (input_ios) {
        rwCloseFile(input_ios);
    }

    return rv;
}


int main(int argc, char **argv)
{
    int rv;

    appSetup(argc, argv); /* never returns on error */

    rv = maskInput("stdin", "stdout");

    /* done */
    appTeardown();

    return rv;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
