/*
** Copyright (C) 2001-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwsort reads SiLK Flow Records from the standard input or from
**  named files and sorts them on one or more user-specified fields.
**
**  A large buffer is allocated; we attempt to create a buffer of
**  MAX_IN_CORE_SIZE bytes, but back off if the allocation fails.
**
**  Records are read and stored in this buffer; if the input ends
**  before the buffer is filled, the records are sorted and printed to
**  standard out or to the named output file.
**
**  However, if the buffer fills before the input is completely read,
**  the records in the buffer are sorted and written to a temporary
**  file on disk; the buffer is cleared, and reading of the input
**  resumes, repeating the process as necessary until all records are
**  read or until we write MAX_TMP_FILES temporary files.  We then do
**  an N-way merge-sort on the temporary files, where N is either all
**  the files or the maximum number that we can open before running
**  out of descriptors (EMFILE).  If we hit EMFILE, we merge the N
**  files into a new temporary file, then add it to the list of files
**  to merge.
**
**  When the temporary files are written to the same volume (file
**  system) as the final output, the maximum disk usage will be
**  2-times the number of records read (times the size per record);
**  when different volumes are used, the disk space required for the
**  temporary files will be between 1 and 1.5 times the number of
**  records.
**
**  TO DO:
**
**    -- Should better handle the case where we reach the
**    MAX_TMP_FILES limit.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwsort.c 7493 2007-06-13 14:53:20Z mthomas $");

#include "rwpack.h"
#include "utils.h"
#include "iochecks.h"
#include "skstringmap.h"
#include "sksite.h"
#include "dynlib.h"


/* LOCAL DEFINES AND TYPEDEFS */

/* where to send --help output */
#define USAGE_FH stdout

/*
 *  For debugging.  Argument should be surrounded by double-parens:
 *      PRINTDEBUG((stderr, "My output\n"));
 */
/* #define PRINTDEBUG(args) fprintf args */
#ifndef PRINTDEBUG
#  define PRINTDEBUG(args)
#endif

/*
 *  How large of a buffer to attempt to allocate; scale back if we
 *  cannot allocate this much space
 *
 *  Try to allocate almost 2GB of RAM.
 */
#define DEFAULT_SORT_BUFFER_SIZE   0x78000000

/*
 *  How quickly to back off when allocation fails.
 */
#define IN_CORE_REDUCE_RATIO  0.75

/*
 *  If we can't allocate space for at least this many records, give
 *  up.
 */
#define MIN_IN_CORE_RECORDS     1000

/*
 *  Maximum number of temporary files to create.  We need an array of
 *  this many char*'s to point to the file names.
 */
#define MAX_TMP_FILES  2048

/*
 *  Max number of plug-ins supported
 */
#define APP_MAX_DYNLIBS 8

/*
 *  Maximum number of fields that can come from plugins.  Allow four
 *  per plug-in.
 */
#define MAX_PLUGIN_KEY_FIELDS  32

/*
 *  Maximum bytes allotted to a "node", which is the complete rwRec
 *  and the bytes required by all keys that can come from plug-ins.
 *  Allow 8 bytes per field, plus enough space for an rwRec.
 */
#define MAX_NODE_SIZE  (256 + SK_MAX_RECORD_SIZE)

/*
 *  Where to create temp files by default.  This can be overridden by
 *  the --temp-dir switch, or environment variable named in
 *  TMPDIR_ENVAR1 or the environment variable named in TMPDIR2.
 *
 *  If this is not defined, no default exists.
 */
#define DEFAULT_TEMP_DIR "/tmp"

/*
 *  Name of primary envionment variable that holds name of temp
 *  directory.  This is consulted when the --temp-dir switch is not
 *  given.
 */
#define TMPDIR_ENVAR1 "SILK_TMPDIR"

/*
 *  Name of alternate envionment variable that holds name of temp
 *  directory.  Used when the --temp-dir switch is not given and the
 *  variable named by TMPDIR_ENVAR1 is not set.
 */
#define TMPDIR_ENVAR2 "TMPDIR"

/*
 *  Function type for signal handler
 */
typedef void (*sig_fn_t)(int);

/* function that plugins must supply */
typedef int (*sortf_t)(unsigned int, uint8_t *bin_val, const rwRec *);

/* Internal interface to libdynlib */
typedef struct {
    /* the libdynlib object */
    dynlibInfoStruct   *dlisp;
    /* number of field just before this dynlib */
    unsigned int        offset;
    /* function to get output from dynlib */
    sortf_t             fxn;
} app_dynlib_t;

/* for key fields that come from plug-ins, this struct will hold
 * information about a single field */
typedef struct _key_field {
    /* the id of the column (from librw) that make up this key_field */
    uint32_t        kf_id;
    /* the byte-offset for this field */
    size_t          kf_offset;
    /* the byte-width of this field */
    size_t          kf_width;
    /* the function to call to fill in this field: used to convert a
     * member of an rwRec into a binary value. */
    sortf_t         kf_fxn;
    /* the field_id that must be passed to kf_fxn for this field */
    unsigned int    kf_fxn_id;
} key_field_t;



/* LOCAL VARIABLES */

/* available key fields; rwAsciiFieldMapAddDefaultFields() fills this */
static sk_stringmap_t *field_map = NULL;

/* number of fields to sort over; rwAsciiFieldMapParseFields() sets this */
static uint32_t num_fields = 0;

/* IDs of the fields to sort over; rwAsciiFieldMapParseFields() sets
 * it; values are from the rwrec_printable_fields_t enum and from
 * values that come from dynamic libraries. */
static uint32_t *sort_fields = NULL;

/* the maximum ID in the field_names string-map */
static uint32_t max_field_id = 0;

/* the size of a "node".  Because the output from rwsort are SiLK
 * records, the node size includes the complete rwRec, plus any binary
 * fields that we get from plug-ins to use as the key.  This node_size
 * value may increase when we parse the --fields switch. */
static uint32_t node_size = sizeof(rwRec);

/* the columns that make up the key that come from plug-ins */
static key_field_t key_fields[MAX_PLUGIN_KEY_FIELDS];

/* the number of these key_fields */
static size_t key_num_fields = 0;

/* input checker */
static iochecksInfoStruct_t *ioISP;

/* name of output stream */
static char *output_path;

/* output stream */
static rwIOStruct_t *out_rwios;

/* non-zero if we are shutting down due to a signal; controls whether
 * errors are printed in appTeardown(). */
static int g_caught_signal = 0;

/* the compression method to use when writing the file.
 * sksiteCompmethodOptionsRegister() will set this to the default or
 * to the value the user specifies. */
static sk_compmethod_t comp_method;

/* maximum amount of RAM to attempt to allocate */
static uint64_t sort_buffer_size = DEFAULT_SORT_BUFFER_SIZE;

/* temporary directory */
static char *g_user_temp_dir = NULL;

/* template used to make temporary files. */
static char temp_file_template[PATH_MAX];

/* names of temporary files */
static char *tmp_names[MAX_TMP_FILES];

/* number of shared object plug-ins loaded */
static int app_dynlib_count;

/* pointer to each plug-in */
static app_dynlib_t app_dynlib_list[APP_MAX_DYNLIBS];

/* names of plug-ins to attempt to load at startup */
static const char *app_dynlib_names[] = {
    "libaddrtype.so",
    "libccfilter.so",
    "libpmapfilter.so",
    NULL /* sentinel */
};


/* OPTIONS */

typedef enum {
    OPT_FIELDS,
    OPT_OUTPUT_PATH,
    OPT_INPUT_PIPE,
    OPT_DYNAMIC_LIB_VALUE,
    OPT_SORT_BUFFER_SIZE,
    OPT_TEMP_DIRECTORY
} appOptionsEnum;

static struct option appOptions[] = {
    {"fields",              REQUIRED_ARG, 0, OPT_FIELDS},
    {"output-path",         REQUIRED_ARG, 0, OPT_OUTPUT_PATH},
    {"input-pipe",          REQUIRED_ARG, 0, OPT_INPUT_PIPE},
    {OPT_DYNAMIC_LIBRARY,   REQUIRED_ARG, 0, OPT_DYNAMIC_LIB_VALUE},
    {"sort-buffer-size",    REQUIRED_ARG, 0, OPT_SORT_BUFFER_SIZE},
    {"temp-directory",      REQUIRED_ARG, 0, OPT_TEMP_DIRECTORY},
    {0,0,0,0}               /* sentinel entry */
};

static const char *appHelp[] = {
    NULL, /* generated dynamically */
    ("Destination for sorted output (stdout|pipe).\n"
     "\tDefault is stdout if stdout is not a terminal"),
    ("Get input byte stream from pipe (stdin|pipe).\n"
     "\tDefault is stdin if no filenames are given on the command line"),
    "Use given dynamic library to add fields. Def. None",
    NULL, /* generated dynamically */
    NULL, /* generated dynamically */
    (char *)NULL
};



/* LOCAL FUNCTION PROTOTYPES */

static void appUsageLong(void);                 /* never returns */
static void appTeardown(void);
static void appExit(int status);                /* never returns */
static void appSetup(int argc, char **argv);    /* never returns */
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  parseFields(const char *fields_string);
static void appHandleSignal(int signal_value);
static void appSetSignalHandler(sig_fn_t);
static int  appAddDynlib(const char *dlpath, int warn_on_error);
static app_dynlib_t* fieldNum2Dynlib(uint32_t field_id);

#define tempFileGetName(tmp_idx) \
    (tmp_names[(tmp_idx)] ? tmp_names[(tmp_idx)] : "NULL")
static int   tempFileMakeTemplate(char *template, size_t len);
static FILE *tempFileCreate(int tmp_idx);
static FILE *tempFileOpen(int tmp_idx);
static void  tempFileRemove(int tmp_idx);
static int   dumpRecordsToTemp(int, const void*, uint32_t);
static void  mergeFiles(rwIOStruct_t *, int, int);
static void  rwqsort(char *a, size_t n, size_t es, int (*cmp)());



/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                             \
    ("--fields=<FIELDS> [SWITCHES] [FILES]\n"                                 \
     "\tRead SiLK Flow records, sort them by the specified FIELD(S), and\n"   \
     "\twrite the records to the named output path or to the standard\n"      \
     "\toutput.  When no FILES are given on command line, flows are read\n"   \
     "\tfrom the standard input.\n")

    FILE *fh = USAGE_FH;
    int i, j;

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);
    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);
    sksiteOptionsUsage(fh);

    for (i = 0; appOptions[i].name; i++ ) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch (i) {
          case OPT_FIELDS:
            /* Dynamically build the help */
            fprintf(fh, "Field(s) to use as key:\n");
            rwAsciiFieldMapPrintUsage(fh, field_map);
            break;
          case OPT_SORT_BUFFER_SIZE:
            fprintf(fh,
                    ("Attempt to allocate this much memory for the sort\n"
                     "\tbuffer, in bytes."
                     "  Append k, m, g, for kilo-, mega-, giga-bytes,\n"
                     "\trespectively. Def. %" PRIu32 "\n"),
                    DEFAULT_SORT_BUFFER_SIZE);
            break;
          case OPT_TEMP_DIRECTORY:
            fprintf(fh,
                    ("Temporary directory to use while sorting.\n"
                     "\tDef. $" TMPDIR_ENVAR1 " or $" TMPDIR_ENVAR2));
#ifdef DEFAULT_TEMP_DIR
            fprintf(fh, " or %s", DEFAULT_TEMP_DIR);
#endif
            fprintf(fh, "\n");
            break;
          default:
            /* Simple help text from the appHelp array */
            assert(appHelp[i]);
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }

    sksiteCompmethodOptionsUsage(fh);

    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibOptionsUsage(app_dynlib_list[j].dlisp, fh);
    }

    fprintf(fh, "\nNOTE: The --fields switch is required.\n");
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int i;
    int rv;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(ioISP);

    /* close and destroy output */
    if (out_rwios) {
        rv = rwioDestroy(&out_rwios);
        if (rv && !g_caught_signal) {
            /* only print error when not in signal handler */
            rwioPrintLastErr(out_rwios, rv, &skAppPrintErr);
        }
        out_rwios = NULL;
    }

    /* remove any temporary files */
    for (i = 0; i < MAX_TMP_FILES; ++i) {
        if (tmp_names[i]) {
            tempFileRemove(i);
        }
    }

    /* Dynamic library teardown */
    for (i = 0; i < app_dynlib_count; ++i) {
        dynlibTeardown(app_dynlib_list[i].dlisp);
        app_dynlib_list[i].dlisp = NULL;
    }
    app_dynlib_count = 0;

    /* free variables */
    if (sort_fields != NULL) {
        free(sort_fields);
    }
    if (field_map) {
        skStringMapDestroy(field_map);
    }

    skAppUnregister();
}


/*
 *  appExit(status)
 *
 *  Exit the application with the given status.
 */
static void appExit(int status)
{
    appTeardown();
    exit(status);
}


/*
 *  appHandleSignal(signal_value)
 *
 *    Call appExit() to exit the program.  If signal_value is SIGPIPE,
 *    close cleanly; otherwise print a message that we've caught the
 *    signal and exit with EXIT_FAILURE.
 */
static void appHandleSignal(int sig)
{
    g_caught_signal = 1;

    if (sig == SIGPIPE) {
        /* we get SIGPIPE if something downstream, like rwcut, exits
         * early, so don't bother to print a warning, and exit
         * successfully */
        appExit(EXIT_SUCCESS);
    } else {
        skAppPrintErr("Caught signal..cleaning up and exiting");
        appExit(EXIT_FAILURE);
    }
}


/*
 *  appSetSignalHandler(sig_fn)
 *
 *    Call sig_fn to clean up any temporary files whenever we get a
 *    signal.
 */
static void appSetSignalHandler(sig_fn_t sig_fn)
{
    /* list of handled signals */
    int sigs[] = {SIGINT, SIGPIPE, SIGQUIT, SIGTERM};
    struct sigaction act;
    size_t i;

    act.sa_handler = sig_fn;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;
#endif

    for (i = 0; i < (sizeof(sigs)/sizeof(int)); ++i) {
        if (sigaction(sigs[i], &act, NULL) < 0) {
            skAppPrintErr("Cannot register handler for signal #%d",
                          sigs[i]);
            appExit(EXIT_FAILURE);
        }
    }
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int rv = LIBRW_OK;
    int j;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

#ifdef DEFAULT_TEMP_DIR
    assert(DEFAULT_TEMP_DIR != NULL);
#endif

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* Initialize variables */
    memset(tmp_names, 0, sizeof(tmp_names));
    memset(key_fields, 0, sizeof(key_fields));

    /* register the options */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)
        || sksiteCompmethodOptionsRegister(&comp_method)
        || sksiteOptionsRegister(SK_SITE_FLAG_CONFIG_FILE))
    {
        skAppPrintErr("unable to register options");
        appExit(EXIT_FAILURE);
    }

    /* allow for 0 input and 0 output pipes */
    ioISP = iochecksSetup(0, 0, argc, argv);

    /* initialize string-map of field identifiers: add default fields;
     * keep the millisecond fields so rwcut and rwsort take the same
     * switches; the seconds and milliseconds value map to the same
     * code. */
    if (rwAsciiFieldMapAddDefaultFields(&field_map)) {
        skAppPrintErr("unable to setup fields stringmap");
        exit(EXIT_FAILURE);
    }
    max_field_id = RWREC_PRINTABLE_FIELD_COUNT - 1;

    /* add hard-coded dynamic libraries */
    for (j = 0; app_dynlib_names[j]; ++j) {
        appAddDynlib(app_dynlib_names[j], 0);
    }

    /* parse options */
    ioISP->firstFile = optionsParse(argc, argv);
    if (ioISP->firstFile < 0) {
        skAppUsage();           /* never returns */
    }

    /* try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names */
    sksiteConfigure(0);

    /* Make sure the user specified at least one field */
    if (num_fields == 0) {
        skAppPrintErr("The sorting key (--%s switch) was not given",
                      appOptions[OPT_FIELDS].name);
        skAppUsage();           /* never returns */
    }

    /* initialize the template for temporary files */
    if (tempFileMakeTemplate(temp_file_template, PATH_MAX)) {
        skAppUsage();
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP) || iochecksInputs(ioISP, 0)) {
        skAppUsage();
    }

    /* Check for an output stream; or default to stdout  */
    if (NULL == output_path) {
        output_path = "stdout";
    }

    /* open output */
    if ((rv = rwioCreate(&out_rwios, output_path, SK_RWIO_WRITE))
        || (rv = rwioSetCompression(out_rwios, comp_method))
        || (rv = rwioOpen(out_rwios)))
    {
        rwioPrintLastErr(out_rwios, rv, NULL);
        skAppPrintErr("Couldn't open output file.  Exiting.");
        appExit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        appExit(EXIT_FAILURE);
    }

    /* set signal handler to clean up temp files on SIGINT, SIGTERM, etc */
    appSetSignalHandler(&appHandleSignal);

    return;                       /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to optionsRegister(); it will be called
 *    by optionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes optionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    switch ((appOptionsEnum)opt_index) {
      case OPT_FIELDS:
        assert(opt_arg);
        if (parseFields(opt_arg)) {
            return 1;
        }
        break;

      case OPT_OUTPUT_PATH:
        /* output file path: treat "stdout" as standard output---which
         * is the default; anything else as a file path */
        assert(opt_arg);
        if (0 != strcmp(opt_arg, "stdout")) {
            /* check for switch given multiple times */
            if (NULL != output_path) {
                skAppPrintErr("Multiple output files: '%s' and '%s'",
                              output_path, opt_arg);
                return 1;
            }
            output_path = opt_arg;
        }
        break;

      case OPT_INPUT_PIPE:
        assert(opt_arg);
        if (iochecksInputSource(ioISP, opt_arg)) {
            return 1;
        }
        break;

      case OPT_DYNAMIC_LIB_VALUE:
        /* dynamic library */
        if (appAddDynlib(opt_arg, 1)) {
            return 1;
        }
        break;

      case OPT_SORT_BUFFER_SIZE:
        if (skStringParseHumanUint64(&sort_buffer_size, opt_arg,
                                     SK_HUMAN_NORMAL))
        {
            skAppPrintErr("Cannot parse --%s value: %s",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        if ((sort_buffer_size < MIN_IN_CORE_RECORDS * node_size)
            || (sort_buffer_size >= UINT32_MAX))
        {
            skAppPrintErr(("The --%s value must be between %" PRIu32
                           " and %" PRIu32),
                          appOptions[opt_index].name,
                          (MIN_IN_CORE_RECORDS * node_size), UINT32_MAX);
            return 1;
        }
        break;

      case OPT_TEMP_DIRECTORY:
        assert(opt_arg);
        g_user_temp_dir = opt_arg;
        break;
    }

    return 0;                     /* OK */
}


/*
 *  status = parseFields(fields_string);
 *
 *    Parse the user's option for the --fields switch and fill in the
 *    global g_fields[].  Return 0 on success; 1 on failure.
 */
static int parseFields(const char *field_string)
{
    uint32_t i;

    /* have we been here before? */
    if (num_fields > 0) {
        skAppPrintErr("--%s switch provided multiple times",
                      appOptions[OPT_FIELDS].name);
        return 1;
    }

    /* verify user gave us input */
    if (field_string == NULL || field_string[0] == '\0') {
        skAppPrintErr("Missing a value for the --%s switch",
                      appOptions[OPT_FIELDS].name);
        return 1;
    }

    /* parse the input */
    if (rwAsciiFieldMapParseFields(&sort_fields, &num_fields, field_string,
                                   field_map, RW_ASCII_DUPES_ERROR))
    {
        /* library printed the error */
        return 1;
    }

    /* initialize any dynlibs; return error if initialization fails */
    for (i = 0; i < num_fields; ++i) {
        if (sort_fields[i] >=  RWREC_PRINTABLE_FIELD_COUNT) {
            /* field comes from a plug-in */
            app_dynlib_t *app_dl;
            key_field_t *key;

            /* field is from a plug-in */
            app_dl = fieldNum2Dynlib(sort_fields[i]);
            if (NULL == app_dl) {
                /* could not initialize the plug-in */
                return 1;
            }
            if (key_num_fields == MAX_PLUGIN_KEY_FIELDS) {
                skAppPrintErr("Too many fields specified %lu > %u max",
                              (unsigned long)key_num_fields,
                              MAX_PLUGIN_KEY_FIELDS);
                return 1;
            }

            key = &(key_fields[key_num_fields]);
            key->kf_fxn    = app_dl->fxn;
            key->kf_id     = sort_fields[i];
            key->kf_fxn_id = key->kf_id - app_dl->offset;
            key->kf_offset = node_size;
            key->kf_width  = (size_t)key->kf_fxn(key->kf_fxn_id, NULL, NULL);

            ++key_num_fields;
            node_size += key->kf_width;
            if (node_size > MAX_NODE_SIZE) {
                skAppPrintErr("Sort key is too large %lu bytes > %d max",
                              (unsigned long)node_size, MAX_NODE_SIZE);
                return 1;
            }
        }
    }

    return 0;
}


/*
 *  appAddDynlib(dlpath, warnOnError);
 *
 *    Load the dynamic library at 'dlpath', store its information in
 *    the global 'app_dynlib_list', and increment the global
 *    'app_dynlib_count'.  Return 0 on success.  Return 1 if the
 *    library cannot be loaded or if too many dynamic libraries have
 *    been loaded.
 */
static int appAddDynlib(const char *dlpath, int warn_on_error)
{
    dynlibInfoStruct *dlisp = NULL;
    sortf_t sortf;
    sk_stringmap_entry_t *entries = NULL;
    int i;
    uint32_t offset;
    int rv = 1; /* return value; assume error */

    /* fields from the plug-in are added above this value.  All fields
     * from the plug-in must have an ID > 0. */
    offset = max_field_id;

    /* too many plug-ins? */
    if (app_dynlib_count == APP_MAX_DYNLIBS) {
        skAppPrintErr("Too many dynlibs.  Only %u allowed",
                      APP_MAX_DYNLIBS);
        goto END;
    }

    /* load plug-in */
    dlisp = dynlibCreate(DYNLIB_SORT);
    if (dynlibLoad(dlisp, dlpath)) {
        /* failed to find library, missing symbols, or setup failed */
        if (warn_on_error) {
            skAppPrintErr("Error loading dynamic library '%s': %s",
                          dlpath, dynlibLastError(dlisp));
        }
        goto END;
    }
    sortf = (sortf_t)dynlibGetRWProcessor(dlisp);
    /* dynlibCheck() should have made this check already */
    assert(sortf != NULL);

    /* get the fields for the application */
    if (dynlibConfigure(dlisp, (void*)(&entries)) != 0) {
        skAppPrintErr("Error getting fields from plug-in %s",
                      dynlibGetPath(dlisp));
        goto END;
    }
    if (entries == NULL) {
        /* We didn't get any fields from dynlibConfigure().  That's
         * the only way this application supports fields, so ignore
         * this dynlib. */
        if (warn_on_error) {
            skAppPrintErr(("Did not get fields from configure() method in\n"
                           "plug-in '%s'"),
                          dynlibGetPath(dlisp));
        }
        goto END;
    }

    /* the dynlibConfigure() routine gave us the fields */
    for (i = 0; entries[i].name; ++i) {
        /* check that value is not too small or too big */
        if (entries[i].id == 0) {
            skAppPrintErr("Invalid id '0' for field '%s' in %s plugin",
                          entries[i].name, dynlibGetPath(dlisp));
            goto END;
        }
        if (offset >= (UINT32_MAX - entries[i].id)) {
            skAppPrintErr("Invalid id '%u' for field '%s' in %s plugin",
                          entries[i].id, entries[i].name,
                          dynlibGetPath(dlisp));
            goto END;
        }

        /* add it */
        if (SKSTRINGMAP_OK != skStringMapAddID(field_map, entries[i].name,
                                               offset + entries[i].id))
        {
            skAppPrintErr("Cannot add entries to map");
            goto END;
        }
        /* keep track of the maximum field id */
        if (max_field_id < (offset + entries[i].id)) {
            max_field_id = offset + entries[i].id;
        }
    }

    /* success, add the dynamic library */
    app_dynlib_list[app_dynlib_count].dlisp  = dlisp;
    app_dynlib_list[app_dynlib_count].fxn    = sortf;
    app_dynlib_list[app_dynlib_count].offset = offset;
    ++app_dynlib_count;

    rv = 0;

  END:
    if ((rv != 0) && (dlisp != NULL)) {
        sk_link_err_t link_rv;
        sk_link_item_t *node;
        sk_stringmap_entry_t *entry;

        dynlibTeardown(dlisp);

        /* roll back the field_map to those whose IDs are <= offset */
        link_rv = skLinkGetHead(&node, field_map);
        while (link_rv == SKLINK_OK) {
            skLinkGetData((void**)&entry, node);
            link_rv = skLinkGetNext(&node, node);
            if (entry->id > offset) {
                /* remove the entry */
                skStringMapRemoveByName(field_map, entry->name);
            }
        }
        max_field_id = offset;
    }

    return rv;
}


/*
 *  fieldNum2Dynlib(field_id);
 *
 *    Given a field number 'field_id' from the global 'field_map',
 *    return a pointer to a app_dynlib_t from the app_dynlib_list[]
 *    that points to the dynamic library used to print the field_id
 *    field.  Initializes the dynamic library.
 *
 *    Return NULL if the field_id is not related to any dynamic
 *    library or if the initialization of dynamic library failed.
 */
static app_dynlib_t *fieldNum2Dynlib(uint32_t field_id)
{
    int j;
    app_dynlib_t *app_dl;

    if (!app_dynlib_count || (field_id <= app_dynlib_list[0].offset)) {
        return NULL;
    }

    for (j = 1; j < app_dynlib_count; ++j) {
        if (field_id <= app_dynlib_list[j].offset) {
            break;
        }
    }
    app_dl = &(app_dynlib_list[j-1]);

    /* since we will be using the dynamic library, initialize it */
    if (dynlibInitialize(app_dl->dlisp)) {
        /* did not find the field; go through a lot of trouble to
         * print an error message, then return */
        sk_vector_t *entries_vec;
        sk_stringmap_entry_t *entry = NULL;

        entries_vec = skVectorNew(sizeof(sk_stringmap_entry_t*));
        if (NULL == entries_vec) {
            skAppPrintErr(("field not available--cannot initialize %s"),
                          dynlibGetPath(app_dl->dlisp));
            return NULL;
        }
        skStringMapGetNames(entries_vec, field_map, field_id);
        skVectorGetValue(&entry, entries_vec, 0);
        if (NULL != entry) {
            skAppPrintErr(("field '%s' not available--cannot initialize %s"),
                          entry->name, dynlibGetPath(app_dl->dlisp));
        }
        skVectorClear(entries_vec);
        return NULL;
    }

    return app_dl;
}


/* Define our raw sorting functions */
#define RETURN_IF_SORTED(a, b)                  \
    if ((a) < (b)) { return -1; }               \
    else { if ((a) > (b)) { return 1; } }

/*
 *  recsort(a, b);
 *
 *     Returns an ordering on the recs pointed to `a' and `b' by
 *     comparing the fields listed in the sort_fields[] array.
 */
static int recsort(const uint8_t *a, const uint8_t *b)
{
    uint32_t i;
    size_t plugin_field = 0;
    int rv;

    for (i = 0; i < num_fields; i++) {
        switch (sort_fields[i]) {
          case RWREC_FIELD_SIP:
            RETURN_IF_SORTED(((rwRec*)a)->sIP.ipnum, ((rwRec*)b)->sIP.ipnum);
            break;

          case RWREC_FIELD_DIP:
            RETURN_IF_SORTED(((rwRec*)a)->dIP.ipnum, ((rwRec*)b)->dIP.ipnum);
            break;

          case RWREC_FIELD_SPORT:
            RETURN_IF_SORTED(((rwRec*)a)->sPort, ((rwRec*)b)->sPort);
            break;

          case RWREC_FIELD_DPORT:
            RETURN_IF_SORTED(((rwRec*)a)->dPort, ((rwRec*)b)->dPort);
            break;

          case RWREC_FIELD_PROTO:
            RETURN_IF_SORTED(((rwRec*)a)->proto, ((rwRec*)b)->proto);
            break;

          case RWREC_FIELD_PKTS:
            RETURN_IF_SORTED(((rwRec*)a)->pkts, ((rwRec*)b)->pkts);
            break;

          case RWREC_FIELD_BYTES:
            RETURN_IF_SORTED(((rwRec*)a)->bytes, ((rwRec*)b)->bytes);
            break;

          case RWREC_FIELD_FLAGS:
            RETURN_IF_SORTED(((rwRec*)a)->flags, ((rwRec*)b)->flags);
            break;

          case RWREC_FIELD_STIME:
          case RWREC_FIELD_STIME_MSEC:
            RETURN_IF_SORTED(((rwRec*)a)->sTime, ((rwRec*)b)->sTime);
            RETURN_IF_SORTED(((rwRec*)a)->sTime_msec,
                             ((rwRec*)b)->sTime_msec);
            break;

          case RWREC_FIELD_ELAPSED:
          case RWREC_FIELD_ELAPSED_MSEC:
            RETURN_IF_SORTED(((rwRec*)a)->elapsed, ((rwRec*)b)->elapsed);
            RETURN_IF_SORTED(((rwRec*)a)->elapsed_msec,
                             ((rwRec*)b)->elapsed_msec);
            break;

          case RWREC_FIELD_ETIME:
          case RWREC_FIELD_ETIME_MSEC:
            {
                uint32_t a_msec;
                uint32_t b_msec;
                uint32_t a_carry = 0;
                uint32_t b_carry = 0;

                a_msec = ((rwRec*)a)->sTime_msec + ((rwRec*)a)->elapsed_msec;
                if (a_msec > 1000) {
                    a_carry = 1;
                    a_msec -= 1000;
                }
                b_msec = ((rwRec*)b)->sTime_msec + ((rwRec*)b)->elapsed_msec;
                if (b_msec > 1000) {
                    b_carry = 1;
                    b_msec -= 1000;
                }

                RETURN_IF_SORTED((((rwRec*)a)->sTime
                                  + ((rwRec*)a)->elapsed + a_carry),
                                 (((rwRec*)b)->sTime
                                  + ((rwRec*)b)->elapsed + b_carry));
                RETURN_IF_SORTED(a_msec, b_msec);
            }
            break;

          case RWREC_FIELD_SID:
            RETURN_IF_SORTED(((rwRec*)a)->sID, ((rwRec*)b)->sID);
            break;

          case RWREC_FIELD_INPUT:
            RETURN_IF_SORTED(((rwRec*)a)->input, ((rwRec*)b)->input);
            break;

          case RWREC_FIELD_OUTPUT:
            RETURN_IF_SORTED(((rwRec*)a)->output, ((rwRec*)b)->output);
            break;

          case RWREC_FIELD_NHIP:
            RETURN_IF_SORTED(((rwRec*)a)->nhIP.ipnum,
                             ((rwRec*)b)->nhIP.ipnum);
            break;

          case RWREC_FIELD_INIT_FLAGS:
            RETURN_IF_SORTED(((rwRec*)a)->init_flags,
                             ((rwRec*)b)->init_flags);
            break;

          case RWREC_FIELD_REST_FLAGS:
            RETURN_IF_SORTED(((rwRec*)a)->rest_flags,
                             ((rwRec*)b)->rest_flags);
            break;

          case RWREC_FIELD_TCP_STATE:
            RETURN_IF_SORTED(((rwRec*)a)->tcp_state,
                             ((rwRec*)b)->tcp_state);
            break;

          case RWREC_FIELD_APPLICATION:
            RETURN_IF_SORTED(((rwRec*)a)->application,
                             ((rwRec*)b)->application);
            break;

          case RWREC_FIELD_FTYPE_CLASS:
          case RWREC_FIELD_FTYPE_TYPE:
            RETURN_IF_SORTED(((rwRec*)a)->flow_type,
                             ((rwRec*)b)->flow_type);
            break;

          case RWREC_FIELD_ICMP_TYPE_CODE:
            {
                /* Figure out whether sPort or dPort has the ICMP
                 * values.  We must check each record's values. */
                uint16_t icmp_a = ((((rwRec*)a)->dPort == 0)
                                   ? (((rwRec*)a)->sPort)
                                   : (((rwRec*)a)->dPort));
                uint16_t icmp_b = ((((rwRec*)b)->dPort == 0)
                                   ? (((rwRec*)b)->sPort)
                                   : (((rwRec*)b)->dPort));
                RETURN_IF_SORTED(icmp_a, icmp_b);
            }
            break;

          default:
            /* we go through the fields in the same way they were
             * added, and plugin_field should always be pointing to
             * the current plugin. */
            assert(plugin_field < key_num_fields);
            assert(key_fields[plugin_field].kf_id == sort_fields[i]);
            rv = memcmp(&(a[key_fields[plugin_field].kf_offset]),
                        &(b[key_fields[plugin_field].kf_offset]),
                        key_fields[plugin_field].kf_width);
            ++plugin_field;
            if (rv != 0) {
                return rv;
            }
            break;
        }
    }

    return 0;
}


/*
 *  ok = tempFileMakeTemplate(template, len);
 *
 *    Fill in the C string 'template' of size 'len' with the temp-file
 *    template.  Attempts to find a temporary directory, either from
 *    the user's input, from the TMPDIR environment variable, or from
 *    the DEFAULT_TEMP_DIR macro.  Also uses the application name.
 *    'template' is suitable for passing to mkstemp().  Returns 0 on
 *    success, or -1 if no directory could be found.
 */
static int tempFileMakeTemplate(char *template, size_t len)
{
    char *tmp_dir = NULL;
    int rv;

    /* First, figure out the directory to use.  Use the user's option
     * if given, or SILK_TMPDIR, TMPDIR, or the default */
    if (!tmp_dir) {
        tmp_dir = g_user_temp_dir;
    }
    if (!tmp_dir) {
        tmp_dir = getenv(TMPDIR_ENVAR1);
    }
    if (!tmp_dir) {
        tmp_dir = getenv(TMPDIR_ENVAR2);
    }
#ifdef DEFAULT_TEMP_DIR
    if (!tmp_dir) {
        tmp_dir = DEFAULT_TEMP_DIR;
        if (tmp_dir && FILEIsATty(stderr)) {
            skAppPrintErr("Warning: Using default temporary directory %s",
                          tmp_dir);
        }
    }
#endif /* DEFAULT_TEMP_DIR */
    if (!tmp_dir) {
        skAppPrintErr("Cannot find a value for the temporary directory.");
        return -1;
    }
    if ( !dirExists(tmp_dir)) {
        skAppPrintErr("Temporary directory '%s' does not exist", tmp_dir);
        return -1;
    }

    rv = snprintf(template, len, "%s/%s_tmp.XXXXXXXX", tmp_dir, skAppName());
    if ((size_t)rv >= len || rv < 0) {
        skAppPrintErr("snprintf error");
        return -1;
    }

    return 0;
}


/*
 *  filep = tempFileCreate(tmp_idx);
 *
 *    Create and open a new temporary file indexed by 'tmp_idx.'  The
 *    name of the file can be found by tempFileGetName(), and the file
 *    can be re-opened with the tempFileOpen() function.  Returns a
 *    file pointer to the new file, or NULL on error: 'tmp_idx' out of
 *    range, mkstemp() fails.
 *
 */
static FILE *tempFileCreate(int tmp_idx)
{
    FILE *fp = NULL;
    int fd;

    /* check that tmp_idx is in range */
    if (tmp_idx < 0 || tmp_idx >= MAX_TMP_FILES) {
        skAppPrintErr("tempFileCreate index %d out of range 0 < n < %d",
                      tmp_idx, MAX_TMP_FILES);
        return NULL;
    }

    /* copy template name */
    tmp_names[tmp_idx] = strdup(temp_file_template);
    if (NULL == tmp_names[tmp_idx]) {
        skAppPrintErr("Out of memory!");
        return NULL;
    }

    /* open file */
    fd = mkstemp(tmp_names[tmp_idx]);
    if (fd == -1) {
        skAppPrintErr("Cannot create temp file %s: %s",
                      tmp_names[tmp_idx], strerror(errno));
        free(tmp_names[tmp_idx]);
        tmp_names[tmp_idx] = NULL;
        return NULL;
    }

    /* convert descriptor to pointer */
    fp = fdopen(fd, "w");
    if (fp == NULL) {
        skAppPrintErr("Cannot convert descriptor to pointer for %s: %s",
                      tmp_names[tmp_idx], strerror(errno));
        close(fd);
        return NULL;
    }

    PRINTDEBUG((stderr, "new tmp %d => '%s'\n", tmp_idx, tmp_names[tmp_idx]));

    return fp;
}


/*
 *  fp = tempFileOpen(tmp_idx);
 *
 *    Returns a FILE* pointing to a unique temporary filename
 *    associated with 'tmp_idx' that has been previously created with
 *    tempFileCreate().  Returns NULL if file cannot be opened.
 */
static FILE *tempFileOpen(int tmp_idx)
{
    char *fn;

    assert (0 <= tmp_idx && tmp_idx < MAX_TMP_FILES);
    assert (tmp_names[tmp_idx]);

    fn = tempFileGetName(tmp_idx);
    PRINTDEBUG((stderr, "existing tmp %d => '%s'\n", tmp_idx, fn));

    return fopen(fn, "r");
}


/*
 *  tempFileRemove(tmp_idx);
 *
 *    Remove the temporary file indexed by 'tmp_idx'.
 */
static void tempFileRemove(int tmp_idx)
{
    char *fn;

    assert (0 <= tmp_idx && tmp_idx < MAX_TMP_FILES);
    assert (tmp_names[tmp_idx]);

    fn = tempFileGetName(tmp_idx);

    PRINTDEBUG((stderr, "Removing file %s of size %lu\n",
                fn, (unsigned long)fileSize(fn)));
    if (-1 == remove(fn)) {
        if (fileExists(fn)) {
            skAppPrintErr("Cannot remove file '%s': %s",
                          fn, strerror(errno));
        }
    }

    tmp_names[tmp_idx][0] = '\0';
    free(tmp_names[tmp_idx]);
    tmp_names[tmp_idx] = NULL;
}


/*
 *  ok = dumpRecordsToTemp(tmp_idx, buffer, count);
 *
 *    Print 'count' records from 'buffer' into a newly created
 *    temporary file.  Index the temp file by 'tmp_idx'.  Returns 0 on
 *    success, or -1 if a temp file cannot be created, or the write or
 *    close calls fail.
 */
static int dumpRecordsToTemp(
    int             tmp_idx,
    const void     *rec_buffer,
    uint32_t        rec_count)
{
    FILE* temp_filep = NULL;
    int rv = -1; /* return value */

    temp_filep = tempFileCreate(tmp_idx);
    if (temp_filep == NULL) {
        goto END;
    }

    PRINTDEBUG((stderr, "Writing %lu records to '%s'...",
                (unsigned long)rec_count, tempFileGetName(tmp_idx)));
    if (rec_count != fwrite(rec_buffer, node_size, rec_count, temp_filep)) {
        /* error writing, errno may or may not be set */
        PRINTDEBUG((stderr, "ERROR!\n"));
        skAppPrintErr("Error writing records to temp file %s: %s",
                      tempFileGetName(tmp_idx), strerror(errno));
        goto END;
    }

    /* Success so far */
    PRINTDEBUG((stderr, "done\n"));
    rv = 0;

  END:
    /* close the file if open */
    if (temp_filep) {
        if (fclose(temp_filep) == EOF) {
            /* error closing file; only report error if status so far
             * is good. */
            if (rv == 0) {
                skAppPrintErr("Error closing temp file %s: %s",
                              tempFileGetName(tmp_idx), strerror(errno));
                rv = -1;
            }
        }
    }

    return rv;
}


/*
 *  mergeFiles(result_temp_id, temp_id_1, temp_id_2)
 *
 *    Merge the temporary files numbered from 'temp_id_1' to
 *    'temp_id_2' into the output file 'out_ios', maintaining sorted
 *    order.  Return 0 on success, or -1 if an error occurs opening
 *    the temporary files or writing records to the output file.
 */
static void mergeFiles(rwIOStruct_t *out_rwios, int tmp_idx_1, int tmp_idx_2)
{
    FILE *fps[MAX_TMP_FILES];
    uint8_t recs[MAX_TMP_FILES][MAX_NODE_SIZE];
    size_t has_data[MAX_TMP_FILES];
    uint8_t lowest_rec[MAX_NODE_SIZE];
    FILE *fp_result = NULL;
    int i, j, count = 0;
    size_t remain = 0;
    int lowest = 0;
    int tmp_idx_a = tmp_idx_1;
    int tmp_idx_b = tmp_idx_2;
    FILE *fp_intermediate = NULL;
    int tmp_idx_intermediate = tmp_idx_2;
    int last = 0;
    int rv;

    PRINTDEBUG((stderr, "merging %d through %d to %s\n",
                tmp_idx_1, tmp_idx_2, rwGetFileName(out_rwios)));

    memset(has_data, 0, sizeof(has_data));

    /* This loop repeats as long as we haven't read all of the temp
     * files generated in the qsort stage. */
    do {
        /* open intermediate tmp file in case we need it (we might not) */
        ++tmp_idx_intermediate;
        fp_intermediate = tempFileCreate(tmp_idx_intermediate);
        if (fp_intermediate == NULL) {
            goto END;
        }

        count = 0;

        /* Repeat for as many temp files as we can open before we hit
         * EMFILE, or all of them if we can */
        for (i = 0, j = tmp_idx_a; j <= tmp_idx_b; i++, j++) {
            if (i == MAX_TMP_FILES) {
                PRINTDEBUG((stderr, "too many temporary files opened\n"));
                goto END;
            }
            fps[i] = tempFileOpen(j);
            if (fps[i]) {
                ++count;
            } else if (errno == EMFILE) {
                /* Blast!  We can't open any more temp files.  So, we
                 * rewind by one to catch this one the next time
                 * around. */
                tmp_idx_b = j - 1;
                PRINTDEBUG((stderr,
                            "EMFILE limit hit--merging %d through %d to %d\n",
                            tmp_idx_a, tmp_idx_b, tmp_idx_intermediate));
                break;
            } else {
                PRINTDEBUG((stderr, "couldn't open tmp file %s\n",
                            tempFileGetName(j)));
                goto END;
            }
        }

        /* Here, we check to see if we've opened all temp files.  If
         * so, set a flag so we break out of the loop after we're
         * done. */
        if (tmp_idx_b == tmp_idx_2) {
            last = 1;
        }
        PRINTDEBUG((stderr,"count: %d\n", count));

        /* Read the first record from each temp file into the work buffer. */
        for (i = 0; i < count; i++) {
            remain |= has_data[i] = fread(recs[i], node_size, 1, fps[i]);
        }

        /* Ensure lowest_rec has a value (not necessarily the lowest yet) */
        memcpy(lowest_rec, recs[0], node_size);
        lowest = 0;

        /* "remain" is a flag which indicates that there is data left
         * to read from one or more of the temp files. */
        while (remain) {
            /* Loop through each of the temp files */
            for (i = 0; i < count; i++) {
                /* If the temp file has not reached EOF yet... */
                if ( has_data[i] ) {
                    /* If this temp file's record is the lowest */
                    if ((lowest < 0) || (recsort(lowest_rec, recs[i]) > 0)) {
                        /* Copy the record to the lowest_rec, and
                         * record the index of the temp file with the
                         * lowest record */
                        memcpy(lowest_rec, recs[i], node_size);
                        lowest = i;
                    }
                }
            }

            if (last) {
                /* if we've opened all tmp files successfully, write
                 * to output */
                rv = rwWrite(out_rwios, (rwRec*)lowest_rec);
                if (0 != rv) {
                    rwioPrintLastErr(out_rwios, rv, &skAppPrintErr);
                    if (LIBRW_ERROR_IS_FATAL(rv)) {
                        goto END;
                    }
                }
            } else {
                /* otherwise, write to intermediate tmp file */
                if (fwrite(lowest_rec, node_size, 1, fp_intermediate) != 1) {
                    skAppPrintErr("Could not write to tmpfile %s",
                                  tempFileGetName(tmp_idx_intermediate));
                    goto END;
                }
            }

            /* replace the record we just wrote */
            has_data[lowest] = fread(recs[lowest], node_size, 1, fps[lowest]);
            memcpy(lowest_rec, recs[lowest], node_size);

            /* Set the "remain" flag to zero if we're done. */
            remain = 0;
            for (i = 0; i < count; i++) {
                remain |= has_data[i];
                if (! has_data[i] ) {
                    /* Force recalculation of the lowest temp file
                     * index if any of the files has hit EOF. */
                    lowest = -1;
                }
            }
        }

        /* Close and delete all of the temp files that we processed
         * this time. */
        for (i = 0; i < count; i++) {
            if (fps[i]) {
                fclose(fps[i]);
            }
            tempFileRemove(tmp_idx_a + i);
        }

        /* Start the next merge with the next input temp file */
        tmp_idx_a = tmp_idx_b + 1;

        /* End the next merge with the temp file we just created */
        tmp_idx_2 = tmp_idx_b = tmp_idx_intermediate;

        /* Close the intermediate temp file. */
        if (fp_intermediate) {
            if (EOF == fclose(fp_intermediate)) {
                skAppPrintErr("Error closing file %s: %s",
                              tempFileGetName(tmp_idx_intermediate),
                              strerror(errno));
                goto END;
            }
            fp_intermediate = NULL;
        }
    } while (! last );

  END:
    if (fp_intermediate) {
        fclose(fp_intermediate);
    }
    if (fp_result) {
        fclose(fp_result);
    }
}



/*
 *  rwIOS = nextFile();
 *
 *    Return a IO Structure to the next input file to read.  Return
 *    NULL if no more input files.
 */
static rwIOStruct_t *nextFile(void)
{
    static int counter = -1;
    rwIOStruct_t *rwios = NULL;

    if (counter < 0) {
        counter = ioISP->firstFile;
    }

    /* Open next file; keep trying until good one found. */
    for (; (rwios == NULL && counter < ioISP->fileCount); counter++) {
        rwios = rwOpenFile(ioISP->fnArray[counter], ioISP->inputCopyFD);
        if (rwios == NULL) {
            skAppPrintErr("Ignoring input file '%s': Could not open",
                          ioISP->fnArray[counter]);
        }
    }

    return rwios;
}


int main(int argc, char **argv)
{
    int temp_file_no = 0;
    rwIOStruct_t *input_rwios;      /* input stream */
    void *record_buffer = NULL;     /* Region of memory for records */
    uint8_t *cur_node = NULL;       /* Ptr into record_buffer */
    uint32_t buffer_max_recs;       /* buffer size (in number of records) */
    uint32_t record_count = 0;      /* Number of records read */
    double multiplier = 1.0;
    size_t i;
    int rv;

    appSetup(argc, argv);                 /* never returns on error */

    /* Allocate the buffer: We'll try to allocate a really BIG
     * chunk...if that fails, halve the chunk until we succeed. */
    buffer_max_recs = sort_buffer_size / node_size;
    while (1) {
        PRINTDEBUG((stderr, "buffer_max_recs = %d\n", buffer_max_recs));
        record_buffer = malloc(node_size * buffer_max_recs);
        if (record_buffer) {
            /* malloc was successful */
            break;
        } else if (buffer_max_recs < MIN_IN_CORE_RECORDS) {
            /* give up at this point */
            skAppPrintErr("Could not allocate space for %d records; abort",
                          MIN_IN_CORE_RECORDS);
            appExit(EXIT_FAILURE);
        } else {
            /* reduce it and try again */
            multiplier *= IN_CORE_REDUCE_RATIO;
            buffer_max_recs = (sort_buffer_size * multiplier) / node_size;
        }
    }

    /* open first file */
    input_rwios = nextFile();

    record_count = 0;
    cur_node = record_buffer;
    while (input_rwios != NULL) {
        /* read record */
        if (rwRead(input_rwios, (rwRec*)cur_node) == 0) {
            /* end of file: close current and open next */
            rwioDestroy(&input_rwios);
            input_rwios = nextFile();
            continue;
        }

        /* lookup data from plug-in */
        for (i = 0; i < key_num_fields; ++i) {
            key_field_t *key = &(key_fields[i]);
            key->kf_fxn(key->kf_fxn_id, &(cur_node[key->kf_offset]),
                        (rwRec*)cur_node);
        }

        record_count++;
        cur_node += node_size;
        if (record_count == buffer_max_recs) {
            /* Sort */
            rwqsort(record_buffer, record_count, node_size, &recsort);

            /* Write to temp file */
            if (dumpRecordsToTemp(temp_file_no++,record_buffer,record_count)){
                appExit(EXIT_FAILURE);
            }

            /* Reset record buffer to 'empty' */
            record_count = 0;
            cur_node = record_buffer;
        }
    }

    /* Sort (and maybe store) last batch of records */
    if (record_count > 0) {
        rwqsort(record_buffer, record_count, node_size, &recsort);

        if (temp_file_no > 0) {
            /* Write last batch to temp file */
            if (dumpRecordsToTemp(temp_file_no++,record_buffer,record_count)) {
                appExit(EXIT_FAILURE);
            }
        }
    }

    /* Generate the output */

    if (record_count == 0 && temp_file_no == 0) {
        /* No records were read at all; print a warning and write a
         * header to the output file */
        free(record_buffer);
        skAppPrintErr("No input records");
        if (0 != (rv = rwioWriteHeader(out_rwios))) {
            rwioPrintLastErr(out_rwios, rv, &skAppPrintErr);
        }
    } else if (temp_file_no == 0) {
        /* No temp files written, just output batch of records */
        uint32_t c;

        PRINTDEBUG((stderr, "Writing %lu records to '%s'\n",
                    (unsigned long)record_count, rwGetFileName(out_rwios)));
        for (c = 0, cur_node = record_buffer;
             c < record_count;
             ++c, cur_node += node_size)
        {
            rv = rwWrite(out_rwios, (rwRec*)cur_node);
            if (0 != rv) {
                rwioPrintLastErr(out_rwios, rv, &skAppPrintErr);
                if (LIBRW_ERROR_IS_FATAL(rv)) {
                    free(record_buffer);
                    appExit(EXIT_FAILURE);
                }
            }
        }
        free(record_buffer);
    } else {
        free(record_buffer);

        /* now merge all the temp files */
        mergeFiles(out_rwios, 0, temp_file_no - 1);
    }

    /* close the file */
    if ((rv = rwioClose(out_rwios))
        || (rv = rwioDestroy(&out_rwios)))
    {
        rwioPrintLastErr(out_rwios, rv, &skAppPrintErr);
        appExit(EXIT_FAILURE);
    }
    out_rwios = NULL;

    appTeardown();
    return (0);
}


/*
  qsort()

  A la Bentley and McIlroy in Software - Practice and Experience,
  Vol. 23 (11) 1249-1265. Nov. 1993


*/

typedef long WORD;
#define W               sizeof(WORD)    /* must be a power of 2 */
#define SWAPINIT(a, es) swaptype = \
                        (( (a-(char *)0) | es) % W) ? 2 : (es > W) ? 1 : 0
#define exch(a, b, t)   (t = a, a = b, b = t)
#define swap(a, b)      swaptype != 0 ? swapfunc(a, b, es, swaptype) : \
                        (void )exch(*(WORD*)(a), *(WORD*)(b), t)
/* swap sequences of records */
#define vecswap(a, b, n)        if ( n > 0) swapfunc(a, b, n, swaptype)

#define PVINIT(pv, pm)  if (swaptype != 0) pv = a, swap(pv, pm); \
                        else pv = (char *)&v, v = *(WORD*)pm

#define min(a, b)       a <= b ? a : b

static char *med3(char *a, char *b, char *c, int (*cmp)())
{
    return(   cmp(a, b) < 0 ?
              (cmp(b, c) < 0 ? b : cmp(a, c) < 0 ? c : a)
              : (cmp(b, c) > 0 ? b : cmp(a, c) > 0 ? c : a)
              );
}


static void swapfunc(char *a, char *b, size_t n, int swaptype)
{
    if( swaptype <= 1)
        {
            WORD t;
            for( ; n > 0; a += W, b += W, n -= W)
                exch(*(WORD*)a, *(WORD*)b, t);
        }
    else
        {
            char t;
            for( ; n > 0; a += 1, b += 1, n -= 1)
                exch(*a, *b, t);
        }
}


static void rwqsort(char *a, size_t n, size_t es, int (*cmp)())
{
    char *pa, *pb, *pc, *pd, *pl, *pm, *pn, *pv;
    int r, swaptype;
    WORD t, v;
    size_t s;

    SWAPINIT(a, es);
    if(n < 7)
        {
            /* use insertion sort on smallest arrays */
            for (pm = a + es; pm < a + n*es; pm += es)
                for(pl = pm; pl > a && cmp(pl-es, pl) > 0; pl -= es)
                    swap(pl, pl-es);
            return;
        }

    pm = a + (n/2)*es;            /* small arrays middle element */
    if ( n > 7)                   /* SLK What about n == 7 ? */
        {
            pl = a;
            pn = a + (n - 1)*es;
            if( n > 40 )
                {
                    /* big arays.  Pseudomedian of 9 */
                    s = (n / 8) * es;
                    pl = med3(pl, pl + s, pl + 2 * s, cmp);
                    pm = med3(pm - s, pm, pm + s, cmp);
                    pn = med3(pn - 2 * s, pn - s, pn, cmp);
                }
            pm = med3(pl, pm, pn, cmp); /* mid-size, med of 3 */
        }
    PVINIT(pv, pm);               /* pv points to the partition value */
    pa = pb = a;
    pc = pd = a + (n - 1) * es;
    for(;;)
        {
            while(pb <= pc && (r = cmp(pb, pv)) <= 0)
                {
                    if (r == 0)
                        {
                            swap(pa, pb);
                            pa += es;
                        }
                    pb += es;
                }
            while (pc >= pb && (r = cmp(pc, pv)) >= 0)
                {
                    if( r == 0)
                        {
                            swap(pc, pd);
                            pd -= es;
                        }
                    pc -= es;
                }
            if( pb > pc)
                break;
            swap(pb, pc);
            pb += es;
            pc -= es;
        }
    pn = a + n * es;
    s = min(pa - a, pb - pa);
    vecswap(a, pb - s, s);
    s = min(pd - pc, pn -pd -(int32_t)es);
    vecswap(pb, pn - s, s);
    if (( s = pb - pa) > es) {
        rwqsort(a, s/es, es, cmp);
    }
    if (( s = pd - pc) > es) {
        rwqsort(pn - s, s/es, es, cmp);
    }

    return;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
