/*
** Copyright (C) 2001-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwuniq.c
**
**  Implementation of the rwuniq application.
**
**  rwuniq reads SiLK flow records---from files listen on the command
**  line or from the standard input when no filenames are given---and
**  bins those flows by a key composed of user-selected fields of an
**  rwRec, or by fields generated from a plug-in---run time dynamic
**  library.  For each bin, a user-selected combination of bytes,
**  packets, flows, earliest start-time, latest end-time, distinct
**  sIPs, and/or distinct dIPs may be computed.
**
**  Once the input is read or rwuniq's hash table runs out of memory,
**  the keys fields and computed values are printed for each bin that
**  meets the user-specified minimum and maximum.
**
*/


#include "silk.h"

RCSIDENT("$SiLK: rwuniq.c 7294 2007-05-25 14:49:53Z mthomas $");

#include "utils.h"
#include "rwpack.h"
#include "iochecks.h"
#include "hashlib.h"
#include "iptree.h"
#include "dynlib.h"
#include "sksite.h"


/* TYPEDEFS AND DEFINES */

/* #define GIVE_META_INFO 1 */

/* Whether to destroy the IPsets in unpackValues() (==1) or at
 * shutdown (==0).  Set to 1 makes output faster when RAM is
 * exhausted, since we free() memory. */
#define UNPACK_VALUES_DESTROY_IPSET  1

/* For the compound key version of rwuniq, the maximum number of bytes
 * we allow in the key.  An arbitary number that should be big enough
 * for almost all cases. */
#define MAX_KEY_BYTE_WIDTH 64

/* Initializer for hash table creation */
#define HASH_INITIAL_SIZE 500000

/* Value to use as exit code when we run out of memory */
#define RWUNIQ_NO_MEMORY_EXIT_CODE 16

/* For ASCII output, maximum size of a single value column. Must be
 * big enough to hold printed uint64_t, timestamp(), class, type, or
 * sensor. */
#define VALUE_OUT_BUF_SIZE 64

/* default sTime bin size to use when --bin-time is requested */
#define DEFAULT_TIME_BIN  60
#define DEFAULT_TIME_BIN_STR  "60"

/* file handle for --help usage message */
#define USAGE_FH stdout

/* Where to write filenames if --print-file specified */
#define PRINT_FILENAMES_FH  stderr

/* function that plugins must supply */
typedef int (*uniqf_t)(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);

/* struct to hold the columns/fields that make up our key for this run */
typedef struct _key_field {
    /* the id of the column (from librw) that make up this key_field */
    uint32_t        kf_id;
    /* the byte-offset for this field */
    size_t          kf_offset;
    /* the byte-width of this field */
    size_t          kf_width;
    /* address of member in global rwrec used to fill this field.
     * When this is set, it is used to convert the binary value back
     * into a text value, regardless of whether kf_fxn is non-null. */
    uint8_t        *kf_addr;
    /* the function to call to fill in this field.  When this is set,
     * it is used to convert a member of an rwRec into a binary value,
     * regardless of whether kf_addr is non-null. */
    uniqf_t         kf_fxn;
    /* the field_id that must be passed to kf_fxn for this field */
    unsigned int    kf_fxn_id;
    /* a buffer to hold the textual value for this field */
    char            kf_textbuf[VALUE_OUT_BUF_SIZE];
    /* width of above for columnar output */
    int             kf_text_len;
} key_field_t;

/*
 *  Possible value fields to count w.r.t. the uniq key.  Keep this in
 *  sync with the value_fields[] array, and with the OPT_COUNT_*
 *  values in appOptionsEnum.
 */
typedef enum {
    COUNT_BYTES = 0,
    COUNT_PACKETS,
    COUNT_FLOWS,
    COUNT_STIME,
    COUNT_ETIME,
    COUNT_SIP,
    COUNT_DIP,
    _TOTAL_COUNTS_ /* must be last item */
} value_types_t;

/* The number of types; should be equal to _TOTAL_COUNTS_ */
#define TOTAL_COUNTS 7

/* struct to hold information about potential value fields */
typedef struct _value_col {
    /* the title of this field */
    const char     *vc_title;
    /* the user's parameter to this field (if any) */
    const char     *vc_user_param;
    /* the byte-width of this field */
    size_t          vc_byte_width;
    /* the text width of the field for columnar output */
    int             vc_text_len;
    /* Packing order: The various values are packed into a single byte
     * array, and we dereference elements of this array directly.
     * Pack things from biggest to smallest avoid mis-alignment
     * issues: Put bytes (64bit) first, and put pointer values between
     * the 64bit and 32bit values. */
    uint8_t         vc_pack_order;
    /* whether this column is used for --all-counts, 1==yes */
    unsigned        vc_include_in_all :1;
    /* whether the user requested this field */
    unsigned        vc_iscounted :1;
} value_col_t;

/* struct to hold the values that we are calculating for this run */
typedef struct _value_field {
    /* only print sums if the minimum value is at least this value */
    uint64_t        vf_min;
    /* only print sums if the maximum value is no more than this value */
    uint64_t        vf_max;
    /* the byte-offset for each value field */
    size_t          vf_offset;
    /* the byte-width of this field */
    size_t          vf_width;
    /* the text width of the field for columnar output */
    int             vf_text_len;
    /* the id for this column */
    value_types_t   vf_id;
} value_field_t;

/* Internal interface to libdynlib */
typedef struct {
    dynlibInfoStruct   *dlisp;  /* the libdynlib object */
    unsigned int        offset; /* number of field just before this dynlib */
    uniqf_t             fxn;    /* function to get output from dynlib */
} app_dynlib_t;

/* Structure describing the output functions for printing the
 * key. Fields that are part of the standard rwrec are printed with an
 * rwAsciiStream_t; those that come from a plug-in must be determined
 * by calling the plug-in's uniq() function, which the key_field_t has
 * a reference to. */
typedef struct _out_stream {
    /* Call rwAsciiPrintRec() with this stream as first arg */
    rwAsciiStream_t *a_stream;
    /* Use the buffer from this key-field */
    key_field_t     *k_field;
} out_stream_t;

/* Maximum number of dynamic libraries that rwuniq can open */
#define APP_MAX_DYNLIBS 8


/* LOCAL FUNCTIONS */

static void appUsageLong(void);                 /* never returns */
static void appSetup(int argc, char **argv);    /* never returns */
static void appTeardown(void);
static int  appOptionsHandler(clientData, int, char *);

/* Adding key and value fields */
static int  parseKeyFields(const char *field_string);
static key_field_t *addKeyField(
    uint32_t        id,
    size_t          width,
    uint8_t        *addr,
    uniqf_t         uniqf,
    unsigned int    fn_id);
static int setValueFields(void);

/* I/O Functions */
#ifdef  ENABLE_BINARY
static void writeBinary(void);
#endif
static void writeColTitles(void);
static void writeAsciiHash(void);
static void outStreamsSetProperties(void);
static int  fillRecordAndKey(uint8_t *key);

/* Functions to manipulate and test the count values in the hash
 * table */
static int  packValue(
    uint8_t        *out_val_ptr,
    const rwRec    *rec_ptr);
static void unpackValues(
    uint64_t        out_value[],
    uint8_t        *packed_val);
static int  checkValues(
    const uint64_t  values[]);
static void bzeroValue(
    uint8_t        *packed_val);

/* Functions to handle plug-ins */
static int appAddDynlib(const char *dlpath, int warn_on_error);
static app_dynlib_t* fieldNum2Dynlib(uint32_t fieldNum);

/* Sometimes key fields must be modified---e.g., sTime and eTime when
 * the --bin-time switch is given.  These functions have the same
 * signature as the uniq() function and do the necessary
 * modification. */
static int setKeyStimeFloor(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);
static int setKeyEtime(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);
static int setKeyAdjustElapsed(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);


/* LOCAL VARIABLES */

/* hash table to count tuples */
static HashTable *g_hash_counter;

/* record to get members of. we use a global in order to reference the
 * address of the members of the record */
static rwRec g_rwrec;

/* number of records read */
static uint64_t g_record_count = 0;

/* delimiter between output columns */
static char g_delimiter = '|';

/* available key fields */
static sk_stringmap_t *field_map = NULL;

/* the maximum ID in the field_names string-map */
static uint32_t max_field_id = 0;

/* the text the user entered for the --fields switch */
static char *fields_arg = NULL;

/* the columns/fields that make up our key for this run */
static key_field_t *key_fields = NULL;

/* the number of actual columns/fields that make up the key */
static size_t key_num_fields = 0;

/* the total byte count of the key */
static size_t key_byte_width = 0;

/* Information about each potential value field. */
static value_col_t value_col[TOTAL_COUNTS] = {
    {"Bytes",      NULL, sizeof(uint64_t),     20, 0, 1, 0},
    {"Packets",    NULL, sizeof(uint32_t),     10, 3, 1, 0},
    {"Records",    NULL, sizeof(uint32_t),     10, 4, 1, 0},
    {"min_sTime",  NULL, sizeof(uint32_t),     19, 5, 1, 0},
    {"max_eTime",  NULL, sizeof(uint32_t),     19, 6, 1, 0},
    {"Unique_SIP", NULL, sizeof(skIPTree_t**), 10, 1, 0, 0},
    {"Unique_DIP", NULL, sizeof(skIPTree_t**), 10, 2, 0, 0}
};

/* Information about each value field that is being computed. */
static value_field_t value_fields[TOTAL_COUNTS];

/* number of value fields the user has requested */
static size_t value_num_fields = 0;

/* total byte count for all value_fields that are being computed */
static size_t value_byte_width = 0;

/* Array of out_stream_t* */
static out_stream_t **outputs;

/* Count of 'outputs' */
static size_t output_count;

/* name of program to run to page output */
static char *pager;

/* whether to use legacy (1), new-style (0), or default (-1) timestamps */
static int legacy_timestamps = -1;

/* the floor of the sTime and/or eTime */
static uint32_t time_bin_size = 0;

/* group input/output variables into a single struct */
static struct _uniq_io {
    FILE                 *outF;
    const char           *outFPath;
    int                   outIsPipe;
    iochecksInfoStruct_t *ISP;
} uniq_io;

/* flags set by the user options */
static struct _uniq_flags {
    unsigned binary_output  :1;      /* Default to ascii */
    unsigned presorted_input:1;      /* Assume input is sorted */
    unsigned print_filenames:1;
    unsigned no_columns     :1;
    unsigned no_titles      :1;
    unsigned epoch_time     :1;
    unsigned integer_sensors:1;
    unsigned integer_ips    :1;
    unsigned zero_pad_ips   :1;
} uniq_flags = {0, 0, 0, 0, 0, 0, 0, 0, 0};

/* List of dynamic libraries to attempt to open at startup */
static const char *app_dynlib_names[] = {
    "libaddrtype.so",
    "libccfilter.so",
    "libpmapfilter.so",
    NULL /* sentinel */
};

/* Dynamic libraries we actually opened */
static app_dynlib_t app_dynlib_list[APP_MAX_DYNLIBS];

/* Number of dynamic libraries actually opened */
static int app_dynlib_count;


/* OPTIONS */

typedef enum {
    OPT_THRESHOLD,
    OPT_FIELDS,
    OPT_PRESORTED_INPUT,
    OPT_ALL_COUNTS,
    OPT_COUNT_BYTES,
    OPT_COUNT_PACKETS,
    OPT_COUNT_FLOWS,
    OPT_COUNT_STIME,
    OPT_COUNT_ETIME,
    OPT_COUNT_SIP,
    OPT_COUNT_DIP,
    OPT_DYNAMIC_LIB_VALUE,
    OPT_BIN_TIME,
    OPT_EPOCH_TIME,
    OPT_INTEGER_IPS,
    OPT_ZERO_PAD_IPS,
    OPT_INTEGER_SENSORS,
    OPT_NO_TITLES,
    OPT_NO_COLUMNS,
    OPT_COLUMN_SEPARATOR,
    OPT_DELIMITED,
    OPT_PRINT_FILENAMES,
    OPT_COPY_INPUT,
    OPT_OUTPUT_PATH,
    OPT_PAGER,
    OPT_LEGACY_TIMESTAMPS
#ifdef ENABLE_BINARY
    , OPT_ENABLE_BINARY
#endif
} appOptionsEnum;

/* Skip deprecated switches in --help output */
static int app_option_first_index = OPT_FIELDS;

static struct option appOptions[] = {
    {"threshold",           REQUIRED_ARG, 0, OPT_THRESHOLD},
    {"fields",              REQUIRED_ARG, 0, OPT_FIELDS},
    {"presorted-input",     NO_ARG,       0, OPT_PRESORTED_INPUT},
    {"all-counts",          NO_ARG,       0, OPT_ALL_COUNTS},
    {"bytes",               OPTIONAL_ARG, 0, OPT_COUNT_BYTES},
    {"packets",             OPTIONAL_ARG, 0, OPT_COUNT_PACKETS},
    {"flows",               OPTIONAL_ARG, 0, OPT_COUNT_FLOWS},
    {"stime",               NO_ARG,       0, OPT_COUNT_STIME},
    {"etime",               NO_ARG,       0, OPT_COUNT_ETIME},
    {"sip-distinct",        OPTIONAL_ARG, 0, OPT_COUNT_SIP},
    {"dip-distinct",        OPTIONAL_ARG, 0, OPT_COUNT_DIP},
    {OPT_DYNAMIC_LIBRARY,   REQUIRED_ARG, 0, OPT_DYNAMIC_LIB_VALUE},
    {"bin-time",            OPTIONAL_ARG, 0, OPT_BIN_TIME},
    {"epoch-time",          NO_ARG,       0, OPT_EPOCH_TIME},
    {"integer-ips",         NO_ARG,       0, OPT_INTEGER_IPS},
    {"zero-pad-ips",        NO_ARG,       0, OPT_ZERO_PAD_IPS},
    {"integer-sensors",     NO_ARG,       0, OPT_INTEGER_SENSORS},
    {"no-titles",           NO_ARG,       0, OPT_NO_TITLES},
    {"no-columns",          NO_ARG,       0, OPT_NO_COLUMNS},
    {"column-separator",    REQUIRED_ARG, 0, OPT_COLUMN_SEPARATOR},
    {"delimited",           OPTIONAL_ARG, 0, OPT_DELIMITED},
    {"print-filenames",     NO_ARG,       0, OPT_PRINT_FILENAMES},
    {"copy-input",          REQUIRED_ARG, 0, OPT_COPY_INPUT},
    {"output-path",         REQUIRED_ARG, 0, OPT_OUTPUT_PATH},
    {"pager",               REQUIRED_ARG, 0, OPT_PAGER},
    {"legacy-timestamps",   OPTIONAL_ARG, 0, OPT_LEGACY_TIMESTAMPS},
#ifdef ENABLE_BINARY
    {"binary",              NO_ARG,       0, OPT_ENABLE_BINARY},
#endif
    {0,0,0,0}               /* sentinel entry */
};

static char *appHelp[] = {
    "DEPRECATED.  Use --flows instead",
    NULL, /* built dynamically */
    ("Assume input has been presorted using\n"
     "\trwsort invoked with the exact same --fields value. Def. No"),
    ("Enable the next five switches--count everything.  If no\n"
     "\tcount is specified, flows are counted.  Def. No"),
    ("Sum bytes in each bin; optionally choose to print\n"
     "\tbins whose total is in given range; range is MIN or MIN-MAX. Def. No"),
    ("Sum packets in each bin; optionally choose to print\n"
     "\tbins whose total is in given range; range is MIN or MIN-MAX. Def. No"),
    ("Count flow records in each bin; optionally choose to print\n"
     "\tbins whose count is in given range; range is MIN or MIN-MAX. Def. No"),
    "Print earliest time flow was seen in each bin. Def. No",
    "Print latest time flow was seen  in each bin. Def. No",
    ("Count distinct sIPs in each bin; optionally choose to\n"
     "\tprint bins whose count is in range; range is MIN or MIN-MAX. Def. No"),
    ("Count distinct dIPs in each bin; optionally choose to\n"
     "\tprint bins whose count is in range; range is MIN or MIN-MAX. Def. No"),
    "Use given dynamic library to add fields. Def. None",
    ("When using 'sTime' or 'eTime' as a key, adjust time(s) to\n"
     "\tto appear in N-second bins (floor of time is used). Def. No, "
     DEFAULT_TIME_BIN_STR),
    "Print times in UNIX epoch seconds. Def. No",
    "Print IP numbers as integers. Def. Dotted decimal",
    "Print IP numbers as zero-padded dotted decimal. Def. No",
    "Print sensor as an integer. Def. Sensor name",
    "Do not print column titles. Def. Print titles",
    "Disable fixed-width columnar output. Def. Columnar",
    "Use specified character between columns. Def. '|'",
    "Shortcut for --no-columns --column-sep=CHAR",
    "Print names of input files as they are opened. Def. No",
    "Copy all input SiLK Flows to given pipe or file. Def. No",
    "Send output to given file path. Def. stdout",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    ("Timestamp format. Choices:\n"
     "\t0==(new)\"YYYY/MM/DDThh:mm:ss\"; 1==(legacy)\"MM/DD/YYYY hh:mm:ss\".\n"
     "\tDef. "
#ifdef SK_ENABLE_LEGACY_TIMESTAMP
     "1"
#else
     "0"
#endif
     " when switch not provided; 1 when switch provided with no value"),
#ifdef ENABLE_BINARY
    "Output in binary (hash serialization) format. Def. ASCII",
#endif
    (char *)NULL
};



/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); optionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
    FILE *fh = USAGE_FH;
    int i;
    int j;

#define USAGE_MSG                                                             \
    ("--fields=N [SWITCHES] [FILES]\n"                                        \
     "\tSummarize SiLK Flow records into user-defined keyed bins specified\n" \
     "\twith the --fields switch.  For each keyed bin, print byte, packet,\n" \
     "\tand/or flow counts and/or the time window when key was active.\n"     \
     "\tWhen no files are given on command line, flows are read from STDIN.\n")

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);

    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);
    sksiteOptionsUsage(fh);
    for (i = app_option_first_index; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch ((appOptionsEnum)i) {
          case OPT_FIELDS:
            /* Dynamically build the help */
            fprintf(fh, "Field(s) to use as key:\n");
            rwAsciiFieldMapPrintUsage(fh, field_map);
            break;
          default:
            /* Simple help text from the appHelp array */
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }

    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibOptionsUsage(app_dynlib_list[j].dlisp, fh);
    }

    fprintf(fh,
            "\nNOTES:\t* The --fields switch is required.\n"
            "\t* If --bytes or --packets is specified, rwuniq will no longer\n"
            "\t  count flows unless --flows is also specified.\n");
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
static void appSetup(int argc, char **argv)
{
    int using_pager = 0;
    size_t i;
    int j;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* check that TOTAL_COUNTS is correct */
    assert(TOTAL_COUNTS == _TOTAL_COUNTS_);

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize globals */
    outputs = NULL;
    output_count = 0;
    uniq_io.outFPath = "stdout";
    uniq_io.outF = stdout;
    value_byte_width = 0;
    value_num_fields = 0;
    memset(value_fields, 0, sizeof(value_fields));

    /* allow for 0 input and 0 output pipes */
    uniq_io.ISP = iochecksSetup(1, 0, argc, argv);

    /* initialize string-map of field identifiers: add default fields,
     * then remove millisec fields, since unique-ing over them makes
     * little sense.
     *
     * Note that although we remove the MSEC fields from the available
     * fields here, the remainder of the code still supports MSEC
     * fields---which are mapped onto the non-MSEC versions of the
     * fields. */
    if (rwAsciiFieldMapAddDefaultFields(&field_map)) {
        skAppPrintErr("unable to setup fields stringmap");
        exit(EXIT_FAILURE);
    }
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_STIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ETIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ELAPSED_MSEC);
    max_field_id = RWREC_PRINTABLE_FIELD_COUNT - 1;

    /* register the options */
    if (optionsRegister(appOptions, (optHandler)appOptionsHandler, NULL)
        || sksiteOptionsRegister(SK_SITE_FLAG_CONFIG_FILE))
    {
        skAppPrintErr("unable to register options");
        exit(EXIT_FAILURE);
    }

    /* try to load run-time dynamic libraries */
    for (j = 0; app_dynlib_names[j]; ++j) {
        appAddDynlib(app_dynlib_names[j], 0);
    }

    /* parse options */
    uniq_io.ISP->firstFile = optionsParse(argc, argv);
    if (uniq_io.ISP->firstFile < 0) {
        skAppUsage();/* never returns */
    }

    /* try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names */
    sksiteConfigure(0);

    /* Use a particular timestamp if requested */
    if (legacy_timestamps == 0) {
        sktimestampSetMode(SK_TIMESTAMP_YYYYMMDD, SK_TIME_FRACSEC_OFF);
    } else if (legacy_timestamps == 1) {
        sktimestampSetMode(SK_TIMESTAMP_MMDDYYYY, SK_TIME_FRACSEC_OFF);
    } else {
        sktimestampSetMode(SK_TIMESTAMP_DEFAULT, SK_TIME_FRACSEC_OFF);
    }

    /* Not having site config file is allowed */
    sksiteConfigure(0);

    /* Set the value fields and parse the user's minima/maxima */
    if (setValueFields()) {
        exit(EXIT_FAILURE);
    }
    /* If epoch time requested, reduce width of COUNT_STIME and
     * COUNT_ETIME fields. */
    if (uniq_flags.epoch_time) {
        for (i = 0; i < value_num_fields; ++i) {
            if ((value_fields[i].vf_id == COUNT_STIME)
                || (value_fields[i].vf_id == COUNT_ETIME))
            {
                value_fields[i].vf_text_len = 10;
            }
        }
    }

    /* Make sure the user specified at least one key field, parse the
     * fields */
    if (fields_arg == NULL || fields_arg[0] == '\0') {
        skAppPrintErr("the --%s switch is required",
                      appOptions[OPT_FIELDS].name);
        skAppUsage();         /* never returns */
    }
    if (parseKeyFields(fields_arg)) {
        exit(EXIT_FAILURE);
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(uniq_io.ISP)
        || iochecksInputs(uniq_io.ISP, 0))
    {
        skAppUsage();
    }

#ifdef ENABLE_BINARY
    /* if binary is selected, check that uniq_io.outF is not a tty */
    if (uniq_flags.binary_output) {
        if (FILEIsATty(uniq_io.outF)) {
            skAppPrintErr("Will not print binary output to tty");
            exit(EXIT_FAILURE);
        }
    }
#endif

    /* final check. See if stdout is being used for both --copy-input
     * and as the destination for the uniq data.  */
    if (uniq_io.ISP->stdoutUsed && uniq_io.outF == stdout) {
        skAppPrintErr("stdout used for both --copy-input and ascii output");
        exit(EXIT_FAILURE);
    }

    /* Invoke the pager */
    using_pager = skOpenPagerWhenStdoutTty(&(uniq_io.outF), &pager);
    if (using_pager < 0) {
        exit(EXIT_FAILURE);
    }
    if (using_pager) {
        uniq_io.outFPath = pager;
        uniq_io.outIsPipe = 1;
    }

    /* looks good, open the --copy-input destination */
    if (iochecksOpenCopyDest(uniq_io.ISP)) {
        exit(EXIT_FAILURE);
    }

    /* Finally, set the output handle properties */
    outStreamsSetProperties();

    /* We don't need this any more.  Destroy it */
    if (field_map != NULL) {
        skStringMapDestroy(field_map);
        field_map = NULL;
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                       /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
static void appTeardown(void)
{
    static int teardownFlag = 0;
    int j;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(uniq_io.ISP);

    /* Dynamic library teardown */
    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibTeardown(app_dynlib_list[j].dlisp);
        app_dynlib_list[j].dlisp = NULL;
    }
    app_dynlib_count = 0;

    /* close output */
    if (uniq_io.outF != stdout) {
        if (uniq_io.outIsPipe) {
            pclose(uniq_io.outF);
        } else {
            if (EOF == fclose(uniq_io.outF)) {
                skAppPrintErr("Error closing output file '%s': %s",
                              uniq_io.outFPath, strerror(errno));
            }
        }
    }

    /* destroy output */
    if (outputs) {
        size_t i;
        out_stream_t *ostream;

        for (i = 0; i < output_count; ++i) {
            ostream = outputs[i];
            if (ostream->a_stream) {
                rwAsciiStreamDestroy(&ostream->a_stream);
            }
            free(ostream);
        }

        free(outputs);
        outputs = NULL;
    }

    /* destroy field map */
    if (field_map != NULL) {
        skStringMapDestroy(field_map);
        field_map = NULL;
    }

    /* destroy key */
    if (key_fields) {
        free(key_fields);
        key_fields = NULL;
    }

    skAppUnregister();
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    Called by optionsParse(), this handles a user-specified switch
 *    that the application has registered, typically by setting global
 *    variables.  Returns 1 if the switch processing failed or 0 if it
 *    succeeded.  Returning a non-zero from from the handler causes
 *    optionsParse() to return a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    size_t i;

    switch ((appOptionsEnum)opt_index) {
      case OPT_FIELDS:
        if (fields_arg) {
            skAppPrintErr("--%s option given multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        fields_arg = opt_arg;
        break;

      case OPT_ALL_COUNTS:
        for (i = 0; i < TOTAL_COUNTS; ++i) {
            if (value_col[i].vc_include_in_all) {
                value_col[i].vc_iscounted = 1;
            }
        }
        break;

      case OPT_THRESHOLD:
#if 0
        /* print a warning if user is running interactively */
        if (FILEIsATty(stdout)) {
           skAppPrintErr("--threshold is deprecated.  use --flows instead");
        }
        /*@fallthrough@*/
#endif
      case OPT_COUNT_FLOWS:
        value_col[COUNT_FLOWS].vc_iscounted = 1;
        value_col[COUNT_FLOWS].vc_user_param = opt_arg;
        break;

      case OPT_COUNT_PACKETS:
        value_col[COUNT_PACKETS].vc_iscounted = 1;
        value_col[COUNT_PACKETS].vc_user_param = opt_arg;
        break;

      case OPT_COUNT_BYTES:
        value_col[COUNT_BYTES].vc_iscounted = 1;
        value_col[COUNT_BYTES].vc_user_param = opt_arg;
        break;

      case OPT_COUNT_STIME:
        /* earlieset stime */
        value_col[COUNT_STIME].vc_iscounted = 1;
        value_col[COUNT_STIME].vc_user_param = opt_arg;
        break;

      case OPT_COUNT_ETIME:
        /* latest etime */
        value_col[COUNT_ETIME].vc_iscounted = 1;
        value_col[COUNT_ETIME].vc_user_param = opt_arg;
        break;

      case OPT_COUNT_SIP:
        value_col[COUNT_SIP].vc_iscounted = 1;
        value_col[COUNT_SIP].vc_user_param = opt_arg;
        break;

      case OPT_COUNT_DIP:
        value_col[COUNT_DIP].vc_iscounted = 1;
        value_col[COUNT_DIP].vc_user_param = opt_arg;
        break;

      case OPT_DYNAMIC_LIB_VALUE:
        /* dynamic library */
        if (appAddDynlib(opt_arg, 1)) {
            return 1;
        }
        break;

      case OPT_BIN_TIME:
        if (opt_arg == NULL || opt_arg[0] == '\0') {
            /* no time given; use default */
            time_bin_size = DEFAULT_TIME_BIN;
        } else {
            /* parse user's time */
            if (skStringParseUint32(&time_bin_size, opt_arg, 1, 0)) {
                skAppPrintErr("Cannot parse %s value '%s'",
                              appOptions[opt_index].name, opt_arg);
                return 1;
            }
        }
        break;

      case OPT_PRESORTED_INPUT:
        uniq_flags.presorted_input = 1;
        break;

      case OPT_EPOCH_TIME:
        uniq_flags.epoch_time = 1;
        break;

      case OPT_INTEGER_IPS:
        uniq_flags.integer_ips = 1;
        break;

      case OPT_ZERO_PAD_IPS:
        uniq_flags.zero_pad_ips = 1;
        break;

      case OPT_INTEGER_SENSORS:
        uniq_flags.integer_sensors = 1;
        break;

      case OPT_NO_TITLES:
        uniq_flags.no_titles = 1;
        break;

      case OPT_NO_COLUMNS:
        uniq_flags.no_columns = 1;
        break;

      case OPT_COLUMN_SEPARATOR:
        g_delimiter = opt_arg[0];
        break;

      case OPT_DELIMITED:
        if (opt_arg) {
            g_delimiter = opt_arg[0];
        }
        uniq_flags.no_columns = 1;
        break;

      case OPT_PRINT_FILENAMES:
        uniq_flags.print_filenames = 1;
        break;

      case OPT_COPY_INPUT:
        if (iochecksAllDestinations(uniq_io.ISP, opt_arg)) {
            return 1;
        }
        break;

      case OPT_OUTPUT_PATH:
        uniq_io.outFPath = opt_arg;
        if (openFile(uniq_io.outFPath, 1 /* write */, &uniq_io.outF,
                      &uniq_io.outIsPipe))
        {
            return 1;
        }
        break;

      case OPT_PAGER:
        pager = opt_arg;
        break;

      case OPT_LEGACY_TIMESTAMPS:
        if (opt_arg == NULL) {
            legacy_timestamps = 1;
        } else if (opt_arg[0] == '1') {
            legacy_timestamps = 1;
        } else if (opt_arg[0] == '0') {
            legacy_timestamps = 0;
        } else {
            skAppPrintErr("Invalid value for %s: '%s'.",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

#ifdef ENABLE_BINARY
      case OPT_ENABLE_BINARY:
        skAppPrintErr("Enabling binary output!");
        uniq_flags.binary_output = 1;
        break;
#endif

    }

    return 0;                     /* OK */
}


/*
 *  keyfield = addKeyField(map_id, byte_width, member_addr, func_ptr, func_id);
 *
 *    Add a new field to the attributes that will make up the key.
 *    'map_id' is the ID that this field has in the global field_map
 *    (RWREC_FIELD_SIP for source IP).
 *
 *    'byte_width' is the number of bytes required to hold this
 *    field's value; if 'byte_width' is 0, the function given in
 *    'func_ptr' will be invoked to get the byte-width.
 *
 *    'member_addr', if non-NULL, is a pointer to the member of the
 *    global rwRec 'g_rwrec' that the binary value of this key-field
 *    will be mapped back to during output.  If 'func_ptr' is NULL,
 *    the 'member_addr' will also be read to get the key field's
 *    value while reading the SiLK flow records.
 *
 *    If 'func_ptr' is non-NULL, it will used to convert the fields on
 *    an rwRec to a binary key-field value.  If 'member_addr' is NULL,
 *    the function will also be used to convert the binary value back
 *    to a textual value for output.
 *
 *    'fn_id' is the value that must be passed to 'func_ptr' to
 *    represent this key-field.
 *
 *    Returns a pointer into the global 'key_fields' array.
 */
static key_field_t *addKeyField(
    uint32_t        map_id,
    size_t          width,
    uint8_t        *addr,
    uniqf_t         uniqf,
    unsigned int    fn_id)
{
    key_field_t *key = &(key_fields[key_num_fields]);

    key->kf_id     = map_id;
    key->kf_offset = key_byte_width;
    key->kf_width  = width;
    key->kf_addr   = addr;
    key->kf_fxn    = uniqf;
    key->kf_fxn_id = fn_id;

    if (uniqf != NULL && addr == NULL) {
        uint8_t dummy_byte;
        rwRec dummy_rec;

        if (width == 0) {
            key->kf_width = (size_t)uniqf(fn_id, &dummy_byte, NULL, 0, NULL);
        }
        key->kf_text_len = (uniqf(fn_id, NULL, NULL, 0, &dummy_rec)
                            - 1);
    }

    /* compute new key width */
    key_byte_width += key->kf_width;
    ++key_num_fields;

    return key;
}


/*
 *  ok = setValueFields();
 *
 *    Loop over all the entries in the value_col[] array; for those
 *    value columns the user wants to count, parse the parameter to
 *    the switch.  The parameter should be NULL to use the default
 *    limits, a single value ("10") or a single value followed by a
 *    single hyphen ("10-") to be used as the minimum threshold, or
 *    the minimum and maximum limits separated by a hyphen ("10-20").
 *
 *    If the user didn't provide a value field, default to counting
 *    flows.
 *
 *    Figure out all the offsets into the values byte-array for each
 *    of the value fields and the overall length of the byte-array.
 *
 *    Returns 0 on success, or 1 if the opt_arg could not be parsed.
 */
static int setValueFields(void)
{
    uint64_t min_val = 1;
    uint64_t max_val = UINT64_MAX;
    uint64_t temp;
    int rv;
    const char *sp;
    size_t offset;
    uint8_t p_order;
    uint8_t target_order;
    uint8_t seen_min;
    size_t seen_idx;
    size_t i;
    int opt_index;

    for (i = 0; i < TOTAL_COUNTS; ++i) {
        min_val = 1;
        max_val = UINT64_MAX;

        if (0 == value_col[i].vc_iscounted) {
            continue;
        }
        opt_index = i + OPT_COUNT_BYTES;

        /* If an argument is provided, parse it as min-max.  If a
         * single number or a single number followed by a hyphen,
         * treat it as the minimum with no maximum. */
        sp = value_col[i].vc_user_param;
        if (sp && sp[0]) {
            rv = skStringParseUint64(&temp, sp, 0, 0);
            if ((rv < 0) || ((rv > 0) && (sp[rv-1] != '-'))) {
                skAppPrintErr("Error parsing argument to --%s: '%s'",
                              appOptions[opt_index].name,
                              value_col[i].vc_user_param);
                return 1;
            }
            if (temp > 0) {
                min_val = temp;
            }

            if ((rv > 0) && (sp[rv] != '\0')) {
                sp += rv;
                if (skStringParseUint64(&temp, sp, 0, 0)) {
                    skAppPrintErr("Error parsing argument to --%s: '%s'",
                                  appOptions[opt_index].name,
                                  value_col[i].vc_user_param);
                    return 1;
                }
                if (temp < min_val) {
                    skAppPrintErr("Bad limits on --%s: max < min: '%s'",
                                  appOptions[opt_index].name,
                                  value_col[i].vc_user_param);
                    return 1;
                }
                max_val = temp;
            }
        }

        value_fields[value_num_fields].vf_min = min_val;
        value_fields[value_num_fields].vf_max = max_val;
        value_fields[value_num_fields].vf_offset = 0;
        value_fields[value_num_fields].vf_width = value_col[i].vc_byte_width;
        value_fields[value_num_fields].vf_text_len = value_col[i].vc_text_len;
        value_fields[value_num_fields].vf_id = i;

        value_byte_width += value_fields[value_num_fields].vf_width;
        ++value_num_fields;
    }

    /* Make certain the user specified at least one value field.  If
     * not, count by records (flows) for legacy compatibility. */
    if (value_num_fields == 0) {
        value_fields[value_num_fields].vf_min = min_val;
        value_fields[value_num_fields].vf_max = max_val;
        value_fields[value_num_fields].vf_offset = 0;
        value_fields[value_num_fields].vf_width
            = value_col[COUNT_FLOWS].vc_byte_width;
        value_fields[value_num_fields].vf_text_len
            = value_col[COUNT_FLOWS].vc_text_len;
        value_fields[value_num_fields].vf_id = COUNT_FLOWS;

        value_byte_width += value_fields[value_num_fields].vf_width;
        ++value_num_fields;
        return 0;
    }

    /* compute all the offsets; since the packing order is different
     * than the printing order, and values are in the value_fields[]
     * array in printing order. */
    offset = 0;
    target_order = 0;
    while (offset != value_byte_width) {
        assert(offset < value_byte_width);
        assert(target_order < TOTAL_COUNTS);
        seen_idx = seen_min = TOTAL_COUNTS;
        for (i = 0; i < value_num_fields; ++i) {
            p_order = value_col[value_fields[i].vf_id].vc_pack_order;
            if ((p_order < seen_min) && (p_order >= target_order)) {
                seen_idx = i;
                seen_min = p_order;
                if (p_order == target_order) {
                    /* there is no need to look further */
                    break;
                }
            }
        }
        assert(seen_min != TOTAL_COUNTS);
        value_fields[seen_idx].vf_offset = offset;
        offset += value_fields[seen_idx].vf_width;
        target_order = seen_min + 1;
    }

    return 0;
}



/*
 *  status = parseKeyFields(field_string);
 *
 *    Parse the string that represents the columns the user wishes to
 *    "uniq" over and fill in the key_fields[] and outputs[] arrays.
 *    Return 0 on success or non-zero on error.
 */
static int parseKeyFields(const char *field_string)
{
    /* the fields (columns) to print and the order to print them; each
     * value is an ID from field_map */
    uint32_t *field_list;

    /* number of fields to print; length of field_list */
    uint32_t field_count;

    /* return value; assume failure */
    int rv = 1;

    /* whether we've added the dPort key.  Both DPORT and
     * ICMP_TYPE_CODE map to the dPort. */
    unsigned int saw_dport = 0;

    /* which of sTime, eTime, and elapsed were seen */
    unsigned int saw_time = 0;

    /* which of class and type were seen */
    unsigned int saw_class_type = 0;

    /* when time-binning is active and all time fields---sTime, eTime,
     * elapased---are requested, we must adjust 'elapsed' so that it
     * is equal to eTime-sTime.  this is non-0 if we must adjust */
    int adjust_elapsed = 0;

    /* rwRec fields are printed by an rwAsciiStream_t; plug-in fields
     * must be printed by called the uniqDL.  ostream is wrapper
     * around these two ways of printing a field. */
    out_stream_t *ostream;

    /* used to build an array of out_stream_t's that will eventually
     * fill the outputs[] array. */
    sk_vector_t *outputs_vec = NULL;

    /* we can map multiple rwRec fields into a single rwAsciiStream_t,
     * but we must know the range of start and end. This gives start
     * of the range, as an index into the field_list[] array. */
    int first_position = 0;

    /* whether the previous key-field is part of an rwAsciiStream */
    int previous_field_builtin;

    /* number of bytes required to pad the key to a word boundary */
    uint32_t pad;
    size_t j;
    uint32_t i;

    /* parse the field-list into an array of fields */
    if (rwAsciiFieldMapParseFields(&field_list, &field_count, field_string,
                                   field_map, RW_ASCII_DUPES_ERROR))
    {
        /* library printed the error */
        return 1;
    }

    /* set the number of key-fields and the key-byte-width to zero.
     * create the array to hold the information about each
     * key-field. */
    key_byte_width = 0;
    key_num_fields = 0;
    key_fields = calloc(field_count, sizeof(key_field_t));
    if (key_fields == NULL) {
        goto END;
    }

    /* create vector to hold output streams */
    outputs = NULL;
    outputs_vec = skVectorNew(sizeof(out_stream_t*));
    if (outputs_vec == NULL) {
        goto END;
    }

    /*
     * We make two passes over the fields is the field_list[].  For
     * this first pass, create an ostream for every key-field that
     * comes from a plug-in, and for each run of consecutive built-in
     * (rwAsciiStream) key-fields.  We will fill in the ostream's on
     * the second pass.
     *
     * Also, for each built-in key-field, add it to the key_fields[]
     * array unless we have to compute it (eTime), in which case we'll
     * add it later.  Note that we do not have to preserve the order
     * of the keys, since we only care about the key's uniqueness.
     */
    previous_field_builtin = 0;
    for (i = 0; i < field_count; ++i) {
        if (field_list[i] >= RWREC_PRINTABLE_FIELD_COUNT) {
            /* field is not built-in; create a new ostream which we
             * will fill in later. */
            ostream = calloc(1, sizeof(out_stream_t));
            if (ostream == NULL) {
                goto END;
            }
            if (0 != skVectorAppendValue(outputs_vec, &ostream)) {
                free(ostream);
                goto END;
            }
            previous_field_builtin = 0;
        } else {
            /* field is built-in */

            /* handle the output stream */
            if (previous_field_builtin == 0) {
                /* must create a new AsciiStream and a new ostream to
                 * hold it. */
                ostream = calloc(1, sizeof(out_stream_t));
                if (ostream == NULL) {
                    goto END;
                }
                if (0 != rwAsciiStreamCreate(&(ostream->a_stream))) {
                    skAppPrintErr("cannot create output stream");
                    goto END;
                }
                if (0 != skVectorAppendValue(outputs_vec, &ostream)) {
                    goto END;
                }
                ostream = NULL;
            }
            previous_field_builtin = 1;

            /* add this field to the key_fields[]. */
            switch ((rwrec_printable_fields_t)field_list[i]) {
              case RWREC_FIELD_SIP:
                /* Counting distinct IPs when the IP is part of the
                 * key doesn't make sense. */
                for (j = 0; j < value_num_fields; ++j) {
                    if (value_fields[j].vf_id == COUNT_SIP) {
                        skAppPrintErr(("Error: Cannot count distinct sIPs"
                                       " when sIP is part of key."));
                        goto END;
                    }
                }
                addKeyField(field_list[i], sizeof(g_rwrec.sIP.ipnum),
                            (uint8_t*)&g_rwrec.sIP.ipnum, NULL, 0);
                break;

              case RWREC_FIELD_DIP:
                /* Counting distinct IPs when the IP is part of the
                 * key doesn't make sense. */
                for (j = 0; j < value_num_fields; ++j) {
                    if (value_fields[j].vf_id == COUNT_DIP) {
                        skAppPrintErr(("Error: Cannot count distinct dIPs"
                                       " when dIP is part of key."));
                        goto END;
                    }
                }
                addKeyField(field_list[i], sizeof(g_rwrec.dIP.ipnum),
                            (uint8_t*)&g_rwrec.dIP.ipnum, NULL, 0);
                break;

              case RWREC_FIELD_SPORT:
                addKeyField(field_list[i], sizeof(g_rwrec.sPort),
                            (uint8_t*)&g_rwrec.sPort, NULL, 0);
                break;

              case RWREC_FIELD_DPORT:
              case RWREC_FIELD_ICMP_TYPE_CODE:
                if (saw_dport == 0) {
                    addKeyField(field_list[i], sizeof(g_rwrec.dPort),
                                (uint8_t*)&g_rwrec.dPort, NULL, 0);
                }
                saw_dport = 1;
                break;

              case RWREC_FIELD_PROTO:
                addKeyField(field_list[i], sizeof(g_rwrec.proto),
                            (uint8_t*)&g_rwrec.proto, NULL, 0);
                break;

              case RWREC_FIELD_PKTS:
                addKeyField(field_list[i], sizeof(g_rwrec.pkts),
                            (uint8_t*)&g_rwrec.pkts, NULL, 0);
                break;

              case RWREC_FIELD_BYTES:
                addKeyField(field_list[i], sizeof(g_rwrec.bytes),
                            (uint8_t*)&g_rwrec.bytes, NULL, 0);
                break;

              case RWREC_FIELD_FLAGS:
                addKeyField(field_list[i], sizeof(g_rwrec.flags),
                            (uint8_t*)&g_rwrec.flags, NULL, 0);
                break;

              case RWREC_FIELD_STIME:
              case RWREC_FIELD_STIME_MSEC:
                if (0 == (saw_time & 2)) {
                    saw_time |= 2;
                    if (time_bin_size == 0) {
                        addKeyField(field_list[i], sizeof(g_rwrec.sTime),
                                    (uint8_t*)&g_rwrec.sTime, NULL, 0);
                    } else {
                        /* must use a function to convert the sTime to a
                         * key, but can use the address to convert from
                         * key back to sTime  */
                        addKeyField(field_list[i], sizeof(g_rwrec.sTime),
                                    (uint8_t*)&g_rwrec.sTime,
                                    &setKeyStimeFloor, 0);
                    }
                }
                break;

              case RWREC_FIELD_ELAPSED:
              case RWREC_FIELD_ELAPSED_MSEC:
                /* this could be a pointer or a function; handle below */
                saw_time |= 1;
                break;

              case RWREC_FIELD_ETIME:
              case RWREC_FIELD_ETIME_MSEC:
                /* etime is always a function; handle below */
                saw_time |= 4;
                break;

              case RWREC_FIELD_SID:
                addKeyField(field_list[i], sizeof(g_rwrec.sID),
                            (uint8_t*)&g_rwrec.sID, NULL, 0);
                break;

              case RWREC_FIELD_INPUT:
                addKeyField(field_list[i], sizeof(g_rwrec.input),
                            (uint8_t*)&g_rwrec.input, NULL, 0);
                break;

              case RWREC_FIELD_OUTPUT:
                addKeyField(field_list[i], sizeof(g_rwrec.output),
                            (uint8_t*)&g_rwrec.output, NULL, 0);
                break;

              case RWREC_FIELD_NHIP:
                addKeyField(field_list[i], sizeof(g_rwrec.nhIP.ipnum),
                            (uint8_t*)&g_rwrec.nhIP.ipnum, NULL, 0);
                break;

              case RWREC_FIELD_INIT_FLAGS:
                addKeyField(field_list[i], sizeof(g_rwrec.init_flags),
                            (uint8_t*)&g_rwrec.init_flags, NULL, 0);
                break;

              case RWREC_FIELD_REST_FLAGS:
                addKeyField(field_list[i], sizeof(g_rwrec.rest_flags),
                            (uint8_t*)&g_rwrec.rest_flags, NULL, 0);
                break;

              case RWREC_FIELD_TCP_STATE:
                addKeyField(field_list[i], sizeof(g_rwrec.tcp_state),
                            (uint8_t*)&g_rwrec.tcp_state, NULL, 0);
                break;

              case RWREC_FIELD_APPLICATION:
                addKeyField(field_list[i], sizeof(g_rwrec.application),
                            (uint8_t*)&g_rwrec.application, NULL, 0);
                break;

              case RWREC_FIELD_FTYPE_CLASS:
              case RWREC_FIELD_FTYPE_TYPE:
                if (saw_class_type == 0) {
                    addKeyField(field_list[i], sizeof(g_rwrec.flow_type),
                                (uint8_t*)&g_rwrec.flow_type, NULL, 0);
                }
                saw_class_type = 1;
                break;
            }
        }
    }

    /* convert the output_vec to an array */
    output_count = skVectorGetCount(outputs_vec);
    outputs = calloc(output_count, sizeof(out_stream_t*));
    if (outputs == NULL) {
        goto END;
    }
    skVectorToArray(outputs, outputs_vec);
    skVectorDestroy(outputs_vec);
    outputs_vec = NULL;

    /* Warn if --bin-time requested but time not part of key.  Warn
     * that duration field will be modified when all time fields are
     * given. */
    if (time_bin_size != 0) {
        if (saw_time < 2) {
            if (FILEIsATty(stderr)) {
                skAppPrintErr("No time fields in key; %s switch ignored",
                              appOptions[OPT_BIN_TIME].name);
            }
        } else if (saw_time == 7) {
            /* must adjust elapsed to be eTime-sTime */
            adjust_elapsed = 1;
            if (FILEIsATty(stderr)) {
                skAppPrintErr("Warning: modifying duration field "
                              "to be eTime-sTime");
            }
        }
    }

    /* We skipped the elapsed field above; add it now that we know
     * whether it should be a function or not. */
    if (saw_time & 1) {
        if (adjust_elapsed) {
            /* use function to convert elapsed to key; use address to
             * convert key back to elapsed. */
            addKeyField(RWREC_FIELD_ELAPSED, sizeof(g_rwrec.elapsed),
                        (uint8_t*)&g_rwrec.elapsed, &setKeyAdjustElapsed, 0);
        } else {
            addKeyField(RWREC_FIELD_ELAPSED, sizeof(g_rwrec.elapsed),
                        (uint8_t*)&g_rwrec.elapsed, NULL, 0);
        }
    }

    /* We skipped the etime field above; add it now that sTime and
     * elapsed have been processed.  This ensures that we will have
     * what we need when we unpack the key during printing. */
    if (saw_time & 4) {
        addKeyField(RWREC_FIELD_ETIME, sizeof(g_rwrec.sTime),
                    NULL, &setKeyEtime, 0);
    }

    /*
     * Now make a second pass through the field_list[].  For any
     * non-built-in fields, find the dynamic library that provides
     * them, initialize that plug-in, add the plug-in's uniq()
     * function to the key-fields, and have the output stream point to
     * the key-field.
     *
     * For built-in fields, fill the rwAsciiStream that we created in
     * the first pass.
     *
     * The variable 'j' is an index into the outputs[] array.
     */
    previous_field_builtin = 0;
    j = 0;
    for (i = 0; i < field_count; ++i) {
        if (field_list[i] >= RWREC_PRINTABLE_FIELD_COUNT) {
            /* field is not built-in */

            app_dynlib_t *app_dl;
            key_field_t *keyfield;
            unsigned int id;

            if (previous_field_builtin) {
                /* found a consecutive run of built-in fields, add
                 * them to the rwAsciiStream. */
                rwAsciiSetFields(outputs[j]->a_stream,
                                 &field_list[first_position],
                                 i - first_position);
                ++j;
            }
            previous_field_builtin = 0;

            /* map the field_id to the dynamic library and initialize
             * the plug-in; find the id that the plug-in uses for this
             * field. */
            app_dl = fieldNum2Dynlib(field_list[i]);
            if (app_dl == NULL) {
                /* could not initialize the plug-in */
                goto END;
            }
            id = field_list[i] - app_dl->offset;

            /* add the key-field */
            keyfield = addKeyField(field_list[i], 0, NULL, app_dl->fxn, id);
            outputs[j]->k_field = keyfield;
            ++j;
        } else {
            /* field is built-in.  We just need to keep a track of
             * consecutive runs of these */
            if (previous_field_builtin == 0) {
                first_position = i;
            }
            previous_field_builtin = 1;
        }
    }

    /* handle the case where the final field(s) are built-in */
    if (previous_field_builtin) {
        /* set the fields on the ascii stream */
        rwAsciiSetFields(outputs[j]->a_stream, &field_list[first_position],
                         i - first_position);
        ++j;
    }
    assert(j == output_count);

    /* check size of key */
    if (key_byte_width > MAX_KEY_BYTE_WIDTH) {
        skAppPrintErr("Fields make too large a key: %u bytes > %u max",
                      (unsigned int)key_byte_width, MAX_KEY_BYTE_WIDTH);
        goto END;
    }

    /* pad key width to a 32-bit boundary to avoid bus errors on Sparc */
    pad = 4 - (key_byte_width % 4);
    if (pad != 4) {
        key_byte_width += pad;
    }

    /* successful */
    rv = 0;

  END:
    if (rv != 0) {
        /* something went wrong.  clean up */
        if (key_fields) {
            free(key_fields);
            key_fields = NULL;
        }
        if (outputs_vec != NULL) {
            /* clean up all the out_stream_t's on error */
            output_count = skVectorGetCount(outputs_vec);
            for (i = 0; i < output_count; ++i) {
                out_stream_t *o_stream;
                skVectorGetValue(&o_stream, outputs_vec, i);
                if (NULL != o_stream->a_stream) {
                    rwAsciiStreamDestroy(&(o_stream->a_stream));
                }
                free(o_stream);
            }
        }
        if (outputs != NULL) {
            for (i = 0; i < output_count; ++i) {
                out_stream_t *o_stream = outputs[i];
                if (NULL != o_stream->a_stream) {
                    rwAsciiStreamDestroy(&(o_stream->a_stream));
                }
                free(o_stream);
            }
            free(outputs);
            outputs = NULL;
        }
    }
    /* do standard clean-up */
    if (outputs_vec != NULL) {
        skVectorDestroy(outputs_vec);
    }
    if (field_list != NULL) {
        free(field_list);
    }

    return rv;
}


/*
 *  outStreamsSetProperties();
 *
 *    Use the switches entered by the user to set properties on all
 *    the ascii-streams that appear in the global 'outputs' array.
 */
static void outStreamsSetProperties(void)
{
    uint32_t i;
    rwAsciiStream_t *a_stream;
    int ip_warning_printed = 0;

    for (i = 0; i < output_count; ++i) {
        if (NULL == outputs[i]->a_stream) {
            continue;
        }

        /* only ascii streams make it here */
        a_stream = outputs[i]->a_stream;

        /* properties for every stream */
        rwAsciiSetNoNewline(a_stream);
        rwAsciiSetDelimiter(a_stream, g_delimiter);
        rwAsciiSetOutputHandle(a_stream, uniq_io.outF);

        /* simple properties */
        if (uniq_flags.no_titles) {
            rwAsciiSetNoTitles(a_stream);
        }
        if (uniq_flags.no_columns) {
            rwAsciiSetNoColumns(a_stream);
        }
        if (uniq_flags.epoch_time) {
            rwAsciiSetEpochTime(a_stream);
        }
        if (uniq_flags.integer_sensors) {
            rwAsciiSetIntegerSensors(a_stream);
        }

        /* conflicting properties */
        if (uniq_flags.zero_pad_ips && !uniq_flags.integer_ips) {
            rwAsciiSetZeroPadIps(a_stream);
        }
        if (uniq_flags.integer_ips) {
            if ((uniq_flags.zero_pad_ips) && (ip_warning_printed == 0)) {
                ip_warning_printed = 1;
                skAppPrintErr("--integer-ips option overrides"
                              " --zero-pad-ips\n"
                              "\tWill print IPs as integers.");
            }
            rwAsciiSetIntegerIps(a_stream);
        }
#if 0
        if ((uniq_flags.icmpTandC) && (icmp_t_and_c == 0)) {
            icmp_t_and_c = 1;
            /* would prefer to do this in the ascii-stream where the
             * user has used sport/dport columns */
            rwAsciiSetIcmpTypeCode(a_stream);
        }
#endif
    }
}


/*
 *  writeColTitles();
 *
 *    Print the column titles to the global 'uniq_io.outF'.
 */
static void writeColTitles(void)
{
    size_t i;
    char buf[VALUE_OUT_BUF_SIZE];
    out_stream_t *ostream;

    if (uniq_flags.no_titles) {
        return;
    }

    /* print key titles */
    for (i = 0; i < output_count; ++i) {
        ostream = outputs[i];
        if (ostream->a_stream) {
            rwAsciiPrintTitles(ostream->a_stream);
        } else {
            ostream->k_field->kf_fxn(ostream->k_field->kf_fxn_id,
                                     0, buf, sizeof(buf), NULL);
            if (uniq_flags.no_columns) {
                fprintf(uniq_io.outF, "%s%c", buf, g_delimiter);
            } else {
                fprintf(uniq_io.outF, "%*.*s%c",
                        ostream->k_field->kf_text_len,
                        ostream->k_field->kf_text_len,
                        buf, g_delimiter);
            }
        }
    }

    /* print value titles */
    for (i = 0; i < value_num_fields; ++i) {
        if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, "%s%c",
                    value_col[value_fields[i].vf_id].vc_title, g_delimiter);
        } else {
            fprintf(uniq_io.outF, "%*.*s%c",
                    value_fields[i].vf_text_len,
                    value_fields[i].vf_text_len,
                    value_col[value_fields[i].vf_id].vc_title, g_delimiter);
        }
    }
    fprintf(uniq_io.outF, "\n");
}


/*
 *  writeAsciiRecord(key, value);
 *
 *    Unpacks the fields from 'key' into the global 'g_rwrec' and
 *    prints the key fields and the value fields to the global output
 *    stream 'uniq_io.outF'.
 *
 *    Note that the indexes into values[] are determined by
 *    value_fields[].vf_id.
 *
 *    The global 'g_rwrec' will be modified.
 */
static void writeAsciiRecord(
    uint8_t    *key,
    uint64_t    values[TOTAL_COUNTS])
{
    size_t i;
    out_stream_t *ostream;
    struct timeval tv;
    char buf[VALUE_OUT_BUF_SIZE];

    /* Zero out global rwRec to avoid display errors---specifically
     * with msec fields and eTime. */
    memset(&g_rwrec, 0, sizeof(rwRec));

    /* Initialize the protocol to 1, so that if the user has requested
     * ICMP without the protocol, we get ICMP values. */
    g_rwrec.proto = 1;

    tv.tv_usec = 0;

    /* unpack the key into the global 'g_rwrec' */
    for (i = 0; i < key_num_fields; ++i) {
        if (key_fields[i].kf_addr) {
            memcpy(key_fields[i].kf_addr, key+key_fields[i].kf_offset,
                   key_fields[i].kf_width);
        } else {
            key_fields[i].kf_fxn(key_fields[i].kf_fxn_id,
                                 key+key_fields[i].kf_offset,
                                 key_fields[i].kf_textbuf,
                                 sizeof(key_fields[i].kf_textbuf),
                                 &g_rwrec);
        }
    }

    /* print key fields */
    for (i = 0; i < output_count; ++i) {
        ostream = outputs[i];
        if (ostream->a_stream) {
            rwAsciiPrintRec(ostream->a_stream, &g_rwrec);
        } else if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, "%s%c",
                    ostream->k_field->kf_textbuf, g_delimiter);
        } else {
            fprintf(uniq_io.outF, "%*s%c",
                    ostream->k_field->kf_text_len,
                    ostream->k_field->kf_textbuf, g_delimiter);
        }
    }

    /* print value fields */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_BYTES:
          case COUNT_PACKETS:
          case COUNT_FLOWS:
          case COUNT_SIP:
          case COUNT_DIP:
            sprintf(buf, ("%" PRIu64), values[i]);
            break;

          case COUNT_STIME:
          case COUNT_ETIME:
            if (uniq_flags.epoch_time) {
                sprintf(buf, ("%" PRIu64), values[i]);
            } else {
                tv.tv_sec = (time_t)values[i];
                sktimestamp_r(buf, &tv, 1);
            }
            break;

          case _TOTAL_COUNTS_:
            assert(0);
            abort();
        }

        if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, "%s%c",
                    buf, g_delimiter);
        } else {
            fprintf(uniq_io.outF, "%*s%c",
                    value_fields[i].vf_text_len,
                    buf, g_delimiter);
        }
    }

    fprintf(uniq_io.outF, "\n");
}


/*
 *  writeAsciiHash();
 *
 *    Loop through all the elements in the global hash table
 *    'g_hash_counter' and print those where the values meet the
 *    thresholds.  Output is written to the global 'uniq_io.outF'.
 */
static void writeAsciiHash(void)
{
    HASH_ITER it;
    uint8_t *key;
    uint8_t *val_ptr; /* the count */
    uint64_t values[TOTAL_COUNTS];
#ifdef GIVE_META_INFO
    size_t i;
    uint64_t meta_entries = 0;    /* table entries */
    uint64_t meta_printed = 0;    /* records printed */

    /* Print some statistics information */
    fprintf(uniq_io.outF, "Accumulated %u records.\n",
            hashlib_count_entries(g_hash_counter));
    fprintf(uniq_io.outF, "Limits:\n");
    for (i = 0; i < value_num_fields; ++i) {
        fprintf(uniq_io.outF, ("\t%" PRIu64 " <= %s <= %" PRIu64 "\n"),
                value_fields[i].vf_min,
                value_col[value_fields[i].vf_id].vc_title,
                value_fields[i].vf_max);
    }
    fprintf(uniq_io.outF, "\n");
#endif

    /* Write out the headings */
    writeColTitles();

    /* create the iterator */
    it = hashlib_create_iterator(g_hash_counter);

    /* Write out the records */
    while (hashlib_iterate(g_hash_counter, &it, &key, &val_ptr)
           != ERR_NOMOREENTRIES)
    {
#ifdef GIVE_META_INFO
        meta_entries++;
#endif
        unpackValues(values, val_ptr);
        if (checkValues(values)) {
            writeAsciiRecord(key, values);
#ifdef GIVE_META_INFO
            ++meta_printed;
#endif
        }
    }

#ifdef GIVE_META_INFO
    if (meta_printed == 0) {
        fprintf(uniq_io.outF, "\nNo records with values within the limits\n");
    } else {
        fprintf(uniq_io.outF,
                ("\n%" PRIu64 " of %" PRIu64 " records met the limits\n"),
                meta_printed, meta_entries);
    }
#endif

    return;
}


#ifdef  ENABLE_BINARY
/* Serialize the hash table data in binary form */
void writeBinary(FILE *fd)
{
    genericHeader header;

    /* Build the header */
    PREPHEADER(&header);
    header.type = FT_RWUNIQ_DATA;
    header.version = RWUNIQ_FILE_VERSION;

    /* Write the data, prefixed with the header */
    hashlib_serialize_table(g_hash_counter, df,
                            (uint8_t*)&header, sizeof(header));
}
#endif


/*
 *  len = setKeyStimeFloor(id, bin_value, text_value, text_len, rwrec);
 *
 *    Get the sTime field from the 'rwrec' and compute the floor of it
 *    relative to the global 'time_bin_size'.  Put that value into
 *    'bin_value' and return the length of 'bin_value'.
 *
 *    This function is only used when converting rwRec's into the
 *    key-byte-array.  For turning the key-byte-array back into an
 *    rwRec, just use the address of the sTime field.
 */
static int setKeyStimeFloor(
    unsigned int    UNUSED(id),
    uint8_t        *bin_value,
    char            UNUSED(*text_value),
    size_t          UNUSED(text_len),
    rwRec          *rwrec)
{
    uint32_t sTime;

    /* find the floor of the time */
    sTime = rwrec->sTime - (rwrec->sTime % time_bin_size);
    memcpy(bin_value, &sTime, sizeof(sTime));
    return (int)sizeof(sTime);
}


/*
 *  len = setKeyAdjustElapsed(id, bin_value, text_value, text_len, rwrec);
 *
 *    Get the elapsed field from the 'rwrec' and adjust it so that the
 *    computed 'sTime' and 'eTime' fields fall into the proper bins
 *    with respect to the global 'time_bin_size'.  Put that computed
 *    elapsed value into 'bin_value' and return the length of
 *    'bin_value'.
 *
 *    This function is only used when converting rwRec's into the
 *    key-byte-array.  For turning the key-byte-array back into an
 *    rwRec, just use the address of the elapsed field.
 */
static int setKeyAdjustElapsed(
    unsigned int    UNUSED(id),
    uint8_t        *bin_value,
    char            UNUSED(*text_value),
    size_t          UNUSED(text_len),
    rwRec          *rwrec)
{
    uint32_t elapsed;

    elapsed = (rwrec->elapsed
               - ((rwrec->sTime + rwrec->elapsed) % time_bin_size)
               + (rwrec->sTime % time_bin_size));
    memcpy(bin_value, &elapsed, sizeof(elapsed));

    return (int)sizeof(elapsed);
}


/*
 *  len = setKeyEtime(id, bin_value, text_value, text_len, rwrec);
 *
 *    When text_value is null (input), use the sTime and elapsed fields on the
 *    'rwrec' to compute the eTime, put the value into 'bin_value',
 *    and return the length of 'bin_value'.  Adjust the eTime relative
 *    to the global 'time_bin_size' if required.
 *
 *    When text_value is non-null (output), adjust the sTime and
 *    elapsed fields on the rwrec so that the eTime will have the
 *    value present in the key-byte-array.
 */
static int setKeyEtime(
    unsigned int    UNUSED(id),
    uint8_t        *bin_value,
    char           *text_value,
    size_t          UNUSED(text_len),
    rwRec          *rwrec)
{
    uint32_t eTime;

    if (NULL == bin_value) {
        return 0;
    }
    if (NULL == text_value) {
        /* compute eTime and put it into bin_value */
        eTime = rwrec->sTime + rwrec->elapsed;
        if (time_bin_size != 0) {
            /* compute floor */
            eTime -= (eTime % time_bin_size);
        }
        memcpy(bin_value, &eTime, sizeof(eTime));

        return (int)sizeof(eTime);
    }

    /* given an eTime value, find the sTime and elapsed values to set
     * on the rwrec that will add to give eTime.  This assumes that
     * the rwrec's sTime and elapsed have already been set. */
    memcpy(&eTime, bin_value, sizeof(eTime));

    if (rwrec->sTime) {
        if (rwrec->elapsed == 0) {
            /* set elapsed based on start and end times */
            assert(rwrec->sTime <= eTime);
            rwrec->elapsed = eTime - rwrec->sTime;
        }
#ifndef NDEBUG
        else {
            /* nothing to do to the records; verify that the times all
             * make sense. */
            assert(eTime == (rwrec->sTime + rwrec->elapsed));
        }
#endif
    } else if (rwrec->elapsed) {
        /* set start time based on end time and elapsed */
        assert(eTime > rwrec->elapsed);
        rwrec->sTime = eTime - rwrec->elapsed;
    } else {
        /* Just set sTime to eTime and elapsed is 0 */
        rwrec->sTime = eTime;
    }

    return 0;
}


/*
 *  status = packValue(out_val_ptr, rec_ptr);
 *
 *    For each value field that is being computed (value_fields[]),
 *    add the values from the current 'rec_ptr' to the values that are
 *    packed into the byte-array 'out_val_ptr'.  Return 0 on success or
 *    -1 on a memory allocation error.
 */
static int packValue(
    uint8_t        *out_val_ptr,
    const rwRec    *rec_ptr)
{
    skIPTree_t **iptree;
    uint64_t val64;
    uint32_t *val32;
    size_t i;

    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_BYTES:
            memcpy(&val64, out_val_ptr + value_fields[i].vf_offset,
                   value_fields[i].vf_width);
            val64 += rec_ptr->bytes;
            memcpy(out_val_ptr + value_fields[i].vf_offset, &val64,
                   value_fields[i].vf_width);
            break;

          case COUNT_PACKETS:
            *((uint32_t *)(out_val_ptr + value_fields[i].vf_offset))
                += rec_ptr->pkts;
            break;

          case COUNT_FLOWS:
            ++*((uint32_t *)(out_val_ptr + value_fields[i].vf_offset));
            break;

          case COUNT_STIME:
            /* earliest sTime */
            val32 = ((uint32_t*)(out_val_ptr + value_fields[i].vf_offset));
            if (*val32 == 0) {
                *val32 = rec_ptr->sTime;
            } else if (rec_ptr->sTime < *val32) {
                *val32 = rec_ptr->sTime;
            }
            break;

          case COUNT_ETIME:
            /* latest eTime */
            val32 = ((uint32_t*)(out_val_ptr + value_fields[i].vf_offset));
            if (*val32 < (rec_ptr->sTime + rec_ptr->elapsed)) {
                *val32 = (rec_ptr->sTime + rec_ptr->elapsed);
            }
            break;

          case COUNT_SIP:
            /* unique source ips */
            iptree = (skIPTree_t**)(out_val_ptr + value_fields[i].vf_offset);
            if (*iptree == NULL) {
                if (skIPTreeCreate(iptree)) {
                    return -1;
                }
            }
            if (skIPTreeAddAddress(rec_ptr->sIP.ipnum, *iptree)) {
                return -1;
            }
            break;

          case COUNT_DIP:
            /* unique destination ips */
            iptree = (skIPTree_t**)(out_val_ptr + value_fields[i].vf_offset);
            if (*iptree == NULL) {
                if (skIPTreeCreate(iptree)) {
                    return -1;
                }
            }
            if (skIPTreeAddAddress(rec_ptr->dIP.ipnum, *iptree)) {
                return -1;
            }
            break;

          case _TOTAL_COUNTS_:
            assert(0);
            abort();
        }
    }

    return 0;
}


/*
 *  bzeroValue(val);
 *
 *    Zero out the value 'val', deleting any IPsets that exist.
 */
static void bzeroValue(
    uint8_t        *packed_val)
{
    skIPTree_t **iptree;
    size_t i;

    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_SIP:
          case COUNT_DIP:
            iptree = (skIPTree_t**)(packed_val + value_fields[i].vf_offset);
            if (*iptree) {
                skIPTreeDelete(iptree);
            }
            break;

          default:
            break;
        }
    }

    memset(packed_val, 0, value_byte_width);
}


/*
 *  unpackValues(out_val, packed_val);
 *
 *    Does reverse of the packValues() subroutine: Unpacks the
 *    'packed_val' values into 'out_val', an array of TOTAL_COUNTS
 *    elements.  Values unpacked depend on the elements of the
 *    value_fields[] array and occur in that order.
 */
static void unpackValues(
    uint64_t        out_val[TOTAL_COUNTS],
    uint8_t        *packed_val)
{
    skIPTree_t **iptree;
    size_t i;

    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_BYTES:
            memcpy(&out_val[i], (packed_val + value_fields[i].vf_offset),
                   value_fields[i].vf_width);
            break;

          case COUNT_PACKETS:
          case COUNT_FLOWS:
          case COUNT_STIME:
          case COUNT_ETIME:
            out_val[i] = *((uint32_t*)(packed_val+value_fields[i].vf_offset));
            break;

          case COUNT_SIP:
          case COUNT_DIP:
            /* the packed value is a pointer to an IPset.  set
             * 'out_val' to the number of IPs in that set. */
            iptree = (skIPTree_t**)(packed_val + value_fields[i].vf_offset);
            if (*iptree) {
                out_val[i] = skIPTreeCountIPs(*iptree);
#if UNPACK_VALUES_DESTROY_IPSET
                skIPTreeDelete(iptree);
#endif
            } else {
                out_val[i] = 0;
            }
            break;

          case _TOTAL_COUNTS_:
            assert(0);
            abort();
        }
    }
}


/*
 *  meets_thresholds = checkValues(values);
 *
 *    Checks the values against the mininum and maximum thresholds;
 *    returns 1 if the values are within the limits, and 0 otherwise.
 *
 *    Values unpacked depend on the elements of the value_fields[]
 *    array and occur in that order.
 */
static int checkValues(
    const uint64_t      values[TOTAL_COUNTS])
{
    size_t i;

    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_BYTES:
          case COUNT_PACKETS:
          case COUNT_FLOWS:
          case COUNT_SIP:
          case COUNT_DIP:
            if ((values[i] < value_fields[i].vf_min)
                || (values[i] > value_fields[i].vf_max))
            {
                return 0;
            }
            break;

          case COUNT_STIME:
          case COUNT_ETIME:
            break;

          case _TOTAL_COUNTS_:
            assert(0);
            abort();
        }
    }

    return 1;
}


#if !UNPACK_VALUES_DESTROY_IPSET
/*
 *  destroyIPTrees();
 *
 *    Destroy the IPTrees that were created in counting unique source
 *    or destination IPs.
 */
static void destroyIPTrees(void)
{
    HASH_ITER it;
    skIPTree_t **iptree;
    uint8_t *key;
    uint8_t *val_ptr;
    size_t i;
    size_t sip_pos = TOTAL_COUNTS;
    size_t dip_pos = TOTAL_COUNTS;

    if (!g_hash_counter) {
        return;
    }

    /* figure out whether COUNT_SIP and/or COUNT_DIP were specified;
     * if so, determine their position in the value_fields array. */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_SIP:
            sip_pos = i;
            break;
          case COUNT_DIP:
            dip_pos = i;
            break;
          default:
            break;
        }
    }
    for (i = 0; i < TOTAL_COUNTS; ++i) {
        if (COUNT_SIP == value_fields[i].vf_id) {
            sip_pos = i;
        }
        if (COUNT_DIP == value_fields[i].vf_id) {
            dip_pos = i;
        }
    }
    if (dip_pos == sip_pos) {
        return;
    }

    /* create the iterator */
    it = hashlib_create_iterator(g_hash_counter);

    while (hashlib_iterate(g_hash_counter, &it, &key, &val_ptr)
           != ERR_NOMOREENTRIES)
    {
        if (sip_pos != TOTAL_COUNTS) {
            iptree = (skIPTree_t**)(val_ptr + value_fields[sip_pos].vf_offset);
            skIPTreeDelete(iptree);
        }
        if (dip_pos != TOTAL_COUNTS) {
            iptree = (skIPTree_t**)(val_ptr + value_fields[dip_pos].vf_offset);
            skIPTreeDelete(iptree);
        }
    }
}
#endif /* !UNPACK_VALUE_DESTROY_IPSET */


/*
 *  appAddDynlib(dlpath, warn_on_error);
 *
 *    Load the dynamic library at 'dlpath', store its information in
 *    the global 'app_dynlib_list', and increment the global counter
 *    'app_dynlib_count'.  Return 0 on success.
 *
 *    Return 1 if the library cannot be loaded or if too many dynamic
 *    libraries have been loaded.  In addition, if 'warn_on_error' is
 *    non-zero, print a message to stderr that the library could not
 *    be loaded.
 */
static int appAddDynlib(const char *dlpath, int warn_on_error)
{
    dynlibInfoStruct *dlisp = NULL;
    uniqf_t uniqf;
    sk_stringmap_entry_t *entries = NULL;
    int i;
    uint32_t offset;
    int rv = 1; /* return value; assume error */

    /* fields from the plug-in are added above this value.  All fields
     * from the plug-in must have an ID > 0. */
    offset = max_field_id;

    /* too many plug-ins? */
    if (app_dynlib_count == APP_MAX_DYNLIBS) {
        skAppPrintErr("Too many dynlibs.  Only %u allowed",
                      APP_MAX_DYNLIBS);
        goto END;
    }

    /* load it */
    dlisp = dynlibCreate(DYNLIB_UNIQ);
    if (dynlibLoad(dlisp, dlpath)) {
        /* failed to find library, missing symbols, or setup failed */
        if (warn_on_error) {
            skAppPrintErr("Error loading dynamic library '%s': %s",
                          dlpath, dynlibLastError(dlisp));
        }
        goto END;
    }
    uniqf = (uniqf_t)dynlibGetRWProcessor(dlisp);
    /* dynlibCheck() should have made this check already */
    assert(uniqf != NULL);

    /* get the fields for the application */
    if (dynlibConfigure(dlisp, (void*)(&entries)) != 0) {
        skAppPrintErr("Error getting fields from plug-in %s",
                      dynlibGetPath(dlisp));
        goto END;
    }

    if (entries == NULL) {
        /* We didn't get any fields from dynlibConfigure().  That's
         * the only way this application supports fields, so ignore
         * this dynlib. */
        if (warn_on_error) {
            skAppPrintErr(("Did not get fields from configure() method in\n"
                           "plug-in '%s'"),
                          dynlibGetPath(dlisp));
        }
        goto END;
    }

    /* the dynlibConfigure() routine gave us the fields */
    for (i = 0; entries[i].name != NULL; ++i) {
        /* check that value is not too small or too big */
        if (entries[i].id == 0) {
            skAppPrintErr("Invalid id '0' for field '%s' in %s plugin",
                          entries[i].name, dynlibGetPath(dlisp));
            goto END;
        }
        if (offset >= (UINT32_MAX - entries[i].id)) {
            skAppPrintErr("Invalid id '%u' for field '%s' in %s plugin",
                          entries[i].id, entries[i].name,
                          dynlibGetPath(dlisp));
            goto END;
        }

        /* add it */
        if (SKSTRINGMAP_OK != skStringMapAddID(field_map, entries[i].name,
                                               offset + entries[i].id))
        {
            skAppPrintErr("Cannot add entries to map");
            goto END;
        }
        /* keep track of the maximum field id */
        if (max_field_id < (offset + entries[i].id)) {
            max_field_id = offset + entries[i].id;
        }
    }

    /* success, add the dynamic library */
    app_dynlib_list[app_dynlib_count].dlisp  = dlisp;
    app_dynlib_list[app_dynlib_count].fxn    = uniqf;
    app_dynlib_list[app_dynlib_count].offset = offset;
    ++app_dynlib_count;

    rv = 0;

  END:
    if ((rv != 0) && (dlisp != NULL)) {
        sk_link_err_t link_rv;
        sk_link_item_t *node;
        sk_stringmap_entry_t *entry;

        dynlibTeardown(dlisp);

        /* roll back the field_map to those whose IDs are <= offset */
        link_rv = skLinkGetHead(&node, field_map);
        while (link_rv == SKLINK_OK) {
            skLinkGetData((void**)&entry, node);
            link_rv = skLinkGetNext(&node, node);
            if (entry->id > offset) {
                /* remove the entry */
                skStringMapRemoveByName(field_map, entry->name);
            }
        }
        max_field_id = offset;
    }

    return rv;
}


/*
 *  fieldNum2Dynlib(field_id);
 *
 *    Given a field number 'field_id' from the global 'field_map',
 *    return a pointer to a app_dynlib_t from the app_dynlib_list[]
 *    that points to the dynamic library used to print the field_id
 *    field.  Initializes the dynamic library.
 *
 *    Return NULL if the field_id is not related to any dynamic
 *    library or if the initialization of dynamic library failed.
 */
static app_dynlib_t* fieldNum2Dynlib(uint32_t field_id)
{
    int j;
    app_dynlib_t *app_dl;

    if (!app_dynlib_count || (field_id <= app_dynlib_list[0].offset)) {
        return NULL;
    }

    for (j = 1; j < app_dynlib_count; ++j) {
        if (field_id <= app_dynlib_list[j].offset) {
            break;
        }
    }
    app_dl = &(app_dynlib_list[j-1]);

    /* since we will be using the dynamic library, initialize it */
    if (dynlibInitialize(app_dl->dlisp)) {
        /* did not find the field; go through a lot of trouble to
         * print an error message, then return */
        sk_vector_t *entries_vec;
        sk_stringmap_entry_t *entry = NULL;

        entries_vec = skVectorNew(sizeof(sk_stringmap_entry_t*));
        if (NULL == entries_vec) {
            skAppPrintErr(("field not available--cannot initialize %s"),
                          dynlibGetPath(app_dl->dlisp));
            return NULL;
        }
        skStringMapGetNames(entries_vec, field_map, field_id);
        skVectorGetValue(&entry, entries_vec, 0);
        if (NULL != entry) {
            skAppPrintErr(("field '%s' not available--cannot initialize %s"),
                          entry->name, dynlibGetPath(app_dl->dlisp));
        }
        skVectorClear(entries_vec);
        return NULL;
    }

    return app_dl;
}


/*
 *  status = fillRecordAndKey(key);
 *
 *    Fills in the global 'g_rwrec' with a flow read from the input.
 *    Processes stdin or the files listed on the command line.
 *    Computes the key based on the key_fields[] settings and fills in
 *    the parameter 'key' with that value.
 */
static int fillRecordAndKey(uint8_t *key)
{
    static int ctr = -1;
    static rwIOStruct_t *rwIOS = NULL;
    size_t i;

    while (ctr < uniq_io.ISP->fileCount) {
        while (rwIOS) {
            /* get next record */
            if (rwRead(rwIOS, &g_rwrec)) {
                /* create the key */
                for (i = 0; i < key_num_fields; ++i) {
                    if (key_fields[i].kf_fxn) {
                        key_fields[i].kf_fxn(key_fields[i].kf_fxn_id,
                                             key+key_fields[i].kf_offset,
                                             NULL, 0, &g_rwrec);
                    } else {
                        memcpy(key+key_fields[i].kf_offset,
                               key_fields[i].kf_addr,
                               key_fields[i].kf_width);
                    }
                }

                ++g_record_count;
                return 1;
            }

            /* rwRead failed; close file */
            rwioDestroy(&rwIOS);
        }

        /* goto next file */
        if (ctr == -1) {
            ctr = uniq_io.ISP->firstFile;
        } else {
            ++ctr;
        }
        if (ctr < uniq_io.ISP->fileCount) {
            /* open the file */
            const char *curFName;
            curFName = uniq_io.ISP->fnArray[ctr];
            rwIOS = rwOpenFile(curFName, uniq_io.ISP->inputCopyFD);
            if (!rwIOS) {
                /* rwOpenFile prints error */
                ctr = uniq_io.ISP->fileCount;
                break;
            }

            /* print its name */
            if (uniq_flags.print_filenames) {
                fprintf(PRINT_FILENAMES_FH, "%s\n", rwGetFileName(rwIOS));
            }
        }
    }

    /* out of files or error */
    return 0;
}


/*
 *  uniqmain_hash();
 *
 *    Main control function that creates a hash table, processes the
 *    input (files or stdin), and prints the results.
 */
static int uniqmain_hash(void)
{
    uint8_t *no_val_ptr;
    uint8_t *key = NULL;
    uint8_t *val_ptr;
    int rv = 0;

    /* allocate the memory for the key */
    key = calloc(key_byte_width, sizeof(uint8_t));
    if (!key) {
        skAppPrintErr("Unable to allocate key");
        exit(EXIT_FAILURE);
    }

    /* Create the value that the hash table uses to indicate a key
     * without a value */
    no_val_ptr = (uint8_t *)malloc(value_byte_width);
    if (!no_val_ptr) {
        skAppPrintErr("Unable to allocate no-value default");
        exit(EXIT_FAILURE);
    }
    memset(no_val_ptr, 0xFF, value_byte_width);

    /* Create the hash table */
    g_hash_counter = hashlib_create_table(key_byte_width,
                                          value_byte_width,
                                          HTT_INPLACE,
                                          no_val_ptr,
                                          NULL,
                                          0,
                                          HASH_INITIAL_SIZE,
                                          DEFAULT_LOAD_FACTOR);
    if (NULL == g_hash_counter) {
        skAppPrintErr("Unable to create hash table");
        rv = 1;
        goto END;
    }

    /* Fill in the global 'g_rwrec' and specified key */
    while (fillRecordAndKey(key)) {

        /* the 'insert' will return memory for the values. either
         * fresh memory or the existing value(s). */
        rv = hashlib_insert(g_hash_counter, key, &val_ptr);
        switch (rv) {
          case OK:
            /* new value */
            memset(val_ptr, 0, value_byte_width);
            /* FALLTHROUGH */

          case OK_DUPLICATE:
            /* existing value */
            if (packValue(val_ptr, &g_rwrec)) {
                if (rv == OK) {
                    /* problem creating new entry; set the value to 0;
                     * this assumes min threshold is 1. */
                    bzeroValue(val_ptr);
                }
                rv = RWUNIQ_NO_MEMORY_EXIT_CODE;
                goto GENERATE_OUTPUT;
            }
            break;

          case ERR_OUTOFMEMORY:
          case ERR_NOMOREBLOCKS:
            /* out of memory */
            --g_record_count;
            rv = RWUNIQ_NO_MEMORY_EXIT_CODE;
            goto GENERATE_OUTPUT;

          default:
            skAppPrintErr(("Unexpected return code '%d'"
                           " from hash table insert"),
                          rv);
            exit(EXIT_FAILURE);
        }
    }

    rv = 0;

  GENERATE_OUTPUT:
    if (rv == RWUNIQ_NO_MEMORY_EXIT_CODE) {
        skAppPrintErr(("Out of memory after processing %" PRIu64 " records for"
                       "\n\t%u hash entries.  Printing partial results"),
                      g_record_count, hashlib_count_entries(g_hash_counter));
    }

    /* Generate output */
#ifdef  ENABLE_BINARY
    if (uniq_flags.binary_output) {
        writeBinary(uniq_io.outF);
    } else {
        writeAsciiHash();
    }
#else
    writeAsciiHash();
#endif

  END:
#if !UNPACK_VALUES_DESTROY_IPSET
    /* If we created IPTrees, destroy them---unless they were
     * destroyed by unpackValues(). */
    destroyIPTrees();
#endif

    if (g_hash_counter) {
        hashlib_free_table(g_hash_counter);
    }
    free(no_val_ptr);
    free(key);

    return rv;
}


/*
 *  uniqmain_presorted();
 *
 *    Main control function that reads presorted flow records from
 *    files or stdin and prints the results.
 */
static int uniqmain_presorted(void)
{
    uint64_t values[TOTAL_COUNTS];
    uint8_t *key = NULL;
    uint8_t *old_key = NULL;
    rwRec new_rec;
    size_t i;
    int rv = 0;
    size_t sip_pos = 0;
    size_t dip_pos = 0;
    skIPTree_t *sip_set = NULL;
    skIPTree_t *dip_set = NULL;

    /* figure out whether COUNT_SIP and/or COUNT_DIP were specified;
     * if so, determine their index in the value_fields array and
     * create IPsets for them. */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case COUNT_SIP:
            sip_pos = i;
            if (skIPTreeCreate(&sip_set)) {
                skAppPrintErr("Unable to create IPset");
                exit(EXIT_FAILURE);
            }
            break;

          case COUNT_DIP:
            dip_pos = i;
            if (skIPTreeCreate(&dip_set)) {
                skAppPrintErr("Unable to create IPset");
                exit(EXIT_FAILURE);
            }
            break;

          default:
            break;
        }
    }


    /* zero out the values */
    memset(values, 0, sizeof(values));

    /* create two keys, current and old */
    key = calloc(key_byte_width, sizeof(uint8_t));
    if (!key) {
        skAppPrintErr("Unable to allocate key");
        exit(EXIT_FAILURE);
    }
    old_key = calloc(key_byte_width, sizeof(uint8_t));
    if (!old_key) {
        skAppPrintErr("Unable to allocate key");
        exit(EXIT_FAILURE);
    }

    /* Write out the headings */
    writeColTitles();

    /* Fill in the global 'g_rwrec' and specified key */
    while (fillRecordAndKey(key)) {
        /* compare keys unless this is the initial record */
        if (g_record_count == 1) {
            memcpy(old_key, key, key_byte_width);
        } else if (0 != memcmp(key, old_key, key_byte_width)) {
            /* keys differ, print the old key and the values */
            if (sip_set) {
                values[sip_pos] = skIPTreeCountIPs(sip_set);
            }
            if (dip_set) {
                values[dip_pos] = skIPTreeCountIPs(dip_set);
            }
            if (checkValues(values)) {
                /* must cache the global 'g_rwrec' */
                memcpy(&new_rec, &g_rwrec, sizeof(rwRec));
                writeAsciiRecord(old_key, values);
                memcpy(&g_rwrec, &new_rec, sizeof(rwRec));
            }

            /* remember the key and reset values */
            memcpy(old_key, key, key_byte_width);
            memset(values, 0, sizeof(values));
            if (sip_set) {
                skIPTreeRemoveAll(sip_set);
            }
            if (dip_set) {
                skIPTreeRemoveAll(dip_set);
            }
        }

        /* add this entry's values */
        for (i = 0; i < value_num_fields; ++i) {
            switch ((value_types_t)value_fields[i].vf_id) {
              case COUNT_BYTES:
                values[i] += g_rwrec.bytes;
                break;

              case COUNT_PACKETS:
                values[i] += g_rwrec.pkts;
                break;

              case COUNT_FLOWS:
                ++values[i];
                break;

              case COUNT_STIME:
                /* earliest sTime */
                if ((values[i] == 0) || (g_rwrec.sTime < values[i])) {
                    values[i] = g_rwrec.sTime;
                }
                break;

              case COUNT_ETIME:
                /* latest eTime */
                if (values[i] < (g_rwrec.sTime + g_rwrec.elapsed)) {
                    values[i] = (g_rwrec.sTime + g_rwrec.elapsed);
                }
                break;

              case COUNT_SIP:
                /* unique source ips */
                skIPTreeAddAddress(g_rwrec.sIP.ipnum, sip_set);
                break;

              case COUNT_DIP:
                /* unique destination ips */
                skIPTreeAddAddress(g_rwrec.dIP.ipnum, dip_set);
                break;

              case _TOTAL_COUNTS_:
                assert(0);
                abort();
            }
        }
    }

    /* print final entry */
    if (g_record_count > 0) {
        if (sip_set) {
            values[sip_pos] = skIPTreeCountIPs(sip_set);
        }
        if (dip_set) {
            values[dip_pos] = skIPTreeCountIPs(dip_set);
        }
        if (checkValues(values)) {
            writeAsciiRecord(key, values);
        }
    }

    /* cleanup */
    if (key) {
        free(key);
    }
    if (old_key) {
        free(old_key);
    }
    if (sip_set) {
        skIPTreeDelete(&sip_set);
    }
    if (dip_set) {
        skIPTreeDelete(&dip_set);
    }

    return rv;
}


int main(int argc, char **argv)
{
    int rv;

    /* Global setup */
    appSetup(argc, argv);

    if (uniq_flags.presorted_input) {
        rv = uniqmain_presorted();
    } else {
        rv = uniqmain_hash();
    }

    /* Done, do cleanup */
    appTeardown();

    return rv;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
