/*
** Copyright (C) 2003-2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: append.c 6898 2007-04-13 20:02:41Z mthomas $");

#include "sendrcv.h"
#include "sksite.h"
#include "sklog.h"


/* Exported variables */
append_options_t append_opt = {APPEND_ENDIAN_AS_IS, SK_INVALID_COMPMETHOD, 0};


/*
 *  status = create_hourly(out_ios, hourly_file, in_ios)
 *
 *    Given the open RW stream 'in_ios' connected to an incremental
 *    file, create a new hourly file 'hourly_file' of the same type
 *    and version (RWSPLIT, etc) to hold the data in the incremental
 *    file.  The new RW stream is put into the value pointed at by
 *    'out_ios'.
 *
 *    The version of the new file is determined by the
 *    'append_opt.mustPad' variable; the endianness of the new file
 *    is determined by the 'append_opt.endian' variable.
 *
 *    Returns 0 on success.  On error, prints a message to the log and
 *    returns -1.
 */
static int create_hourly(
    rwIOStruct        **out_ios,
    const char         *hourly_file,
    const rwIOStruct   *in_ios)
{
    int rv = LIBRW_OK;
    silk_endian_t out_endian;
    fileVersion_t out_version;
    sk_compmethod_t out_compmethod;

    /* Determine the byte order for the new file */
    switch (append_opt.endian) {
      case APPEND_ENDIAN_AS_IS:
        if (rwGetIsBigEndian(in_ios)) {
            out_endian = SILK_ENDIAN_BIG;
        } else {
            out_endian = SILK_ENDIAN_LITTLE;
        }
        break;
      case APPEND_ENDIAN_LITTLE:
        out_endian = SILK_ENDIAN_LITTLE;
        break;
      case APPEND_ENDIAN_BIG:
        out_endian = SILK_ENDIAN_BIG;
        break;
      case APPEND_ENDIAN_NATIVE:
      default:
        logMsg("Unexpected value '%d' in endian switch", append_opt.endian);
        return -1;
    }

    if (append_opt.comp_method == SK_INVALID_COMPMETHOD) {
        /* if user didn't set the value, use that from the file we
         * received */
        out_compmethod = rwGetFileCompMethod(in_ios);
    } else {
        /* use the value the user specified */
        out_compmethod = append_opt.comp_method;
    }

    /* Determine the version for the new file.  Bump the version
     * number from 1 to 2 to get header padding if required. */
    out_version = rwGetFileVersion(in_ios);
    if ((append_opt.must_pad == 1) && (out_version == 1)) {
        out_version = 2;
    }

    /* Create the output stream object and set its attributes. */
    if ((rv = rwioCreate(out_ios, hourly_file, SK_RWIO_WRITE))
        || (rv = rwioMakeDirectory(*out_ios))
        || (rv = rwioSetFileType(*out_ios, rwGetFileType(in_ios)))
        || (rv = rwioSetCompression(*out_ios, out_compmethod))
        || (rv = rwioSetFileVersion(*out_ios, out_version))
        || (rv = rwioSetFileByteorder(*out_ios, out_endian))
        || (rv = rwioSetFileSTime(*out_ios, rwGetFileSTime(in_ios), 0)))
    {
        goto ERROR;
    }

    /* open the file */
    rv = rwioOpen(*out_ios);
    if (rv) {
        if ((rv == LIBRW_ERR_UNSUPPORT_COMPRESS)
            && (out_compmethod != SK_COMPMETHOD_NONE))
        {
            /* Assume we attempted to set the compression of an
             * earlier file version.  Bump the version and try
             * again. */
            logMsg(("Modifying version number of file '%s' (from %u)"
                    " to support compression"),
                   hourly_file, out_version);
            if (LIBRW_OK == rwioSetFileVersion(*out_ios, SK_FILE_VERSION_ANY)){
                rv = rwioOpen(*out_ios);
            }
        }
        if (rv) {
            goto ERROR;
        }
    }

    /* Write the header to avoid zero-length files in the data
     * store. */
    rv = rwioWriteHeader(*out_ios);
    if (rv) {
        goto ERROR;
    }

    /* success */
    return 0;

  ERROR:
    if (rv != LIBRW_OK) {
        rwioPrintLastErr(*out_ios, rv, &logMsg);
    }
    rwioDestroy(out_ios);
    return -1;
}


/*
 *  status = append_data(fnames, &out_pos);
 *
 *    Given the names of an incremental source file and an hourly
 *    destination file in the 'fnames' structure, append the data in
 *    the incremental file to the destination file, creating a new
 *    destination file if it does not exist.
 *
 *    The function will set 'out_pos' to the position in the hourly
 *    file where thew new incremental data begins (i.e., the size of
 *    the file before the append); this will be 0 for a newly created
 *    file.
 *
 *    If all goes well, return 0.  If an error occurs, write an error
 *    to the log buffer and return non-zero.  A return value of -1
 *    indicates that no data was appended; 1 indicates that data was
 *    appended at 'out_pos'---indicating that the data file may now be
 *    corrupt.
 */
int append_data(
    file_names_t   *fnames,
    off_t          *out_pos)
{
    uint32_t long_duration = 0;
    rwIOStruct *in_ios = NULL;
    rwIOStruct *out_ios = NULL;
    int ret_val = -1;
    int rwio_rv;
    rwRec rwrec;

    *out_pos = 0;

    /* Open input file */
    if ((rwio_rv = rwioCreate(&in_ios, fnames->input_path, SK_RWIO_READ))
        || (rwio_rv = rwioOpen(in_ios)))
    {
        rwioPrintLastErr(in_ios, rwio_rv, &logMsg);
        goto END;
    }

    /* The code assumes we have a fileSTime in the header; reject
     * files that do not have the correct header. */
    if ((rwGetFileType(in_ios) == FT_RWFILTER)
        || (rwGetFileType(in_ios) == FT_RWGENERIC))
    {
        logMsg("Type of input file '%s' (0x%02x) is not supported.",
               fnames->input_path, rwGetFileType(in_ios));
        goto END;
    }

    /* Open or create a new file as necessary.  Having zero-length
     * files in the data store should be a rare occurrence, but note
     * that fileSize() also returns 0 for non-existent files. */
    if (0 == fileSize(fnames->output_path)) {
        if (fileExists(fnames->output_path)) {
            /* Remove the zero-length file noisily */
            logMsg("Removing zero length file '%s'", fnames->output_path);
            if (unlink(fnames->output_path) == -1) {
                logMsg("Cannot remove '%s': %s",
                       fnames->output_path, strerror(errno));
                goto END;
            }
        }
        /* Create new output file */
        if (0 != create_hourly(&out_ios, fnames->output_path, in_ios)) {
            goto END;
        }
    } else {
        /* Open existing output file */
        if ((rwio_rv = rwioCreate(&out_ios,fnames->output_path,SK_RWIO_APPEND))
            || (rwio_rv = rwioOpen(out_ios)))
        {
            rwioPrintLastErr(out_ios, rwio_rv, &logMsg);
            goto END;
        }
        *out_pos = lseek(rwGetFD(out_ios), 0, SEEK_CUR);

        /* Sanity check that the times in the headers of the two files
         * are the same. */
        if (rwGetFileSTime(in_ios) != rwGetFileSTime(out_ios)) {
            logMsg("StartTime mismatch in files: incremental %u, output %u",
                   rwGetFileSTime(in_ios), rwGetFileSTime(out_ios));
            goto END;
        }
    }

    /* read the first record */
    if ( !rwRead(in_ios, &rwrec)) {
        /* check for no records versus a bad read */
        if (fileSize(fnames->input_path) == (off_t)rwGetHeaderLength(in_ios)) {
            logMsg("Input has no records");
            ret_val = 0;
        } else {
            logMsg("Could not read first record from '%s'",
                   fnames->input_path);
        }
        goto END;
    }

    /* we're about to modify the output file; adjust return value */
    ret_val = 1;

    do {
        rwio_rv = rwWrite(out_ios, &rwrec);
        if (rwio_rv != LIBRW_OK) {
            if (LIBRW_ERROR_IS_FATAL(rwio_rv)) {
                rwioPrintLastErr(out_ios, rwio_rv, &logMsg);
                goto END;
            }

            if ((rwio_rv == LIBRW_ERR_ELPSD_OVRFLO)&&(rwrec.elapsed > 0x03FF))
            {
                /* try to handle case of an overflow in the elapsed
                 * time---our formats changed from 11 bits to 12 bits,
                 * but we may be trying to write new 12b into old 11b.
                 */
                ++long_duration;
                rwrec.elapsed = 0x03FF;
                rwio_rv = rwWrite(out_ios, &rwrec);
            }
            if (rwio_rv != LIBRW_OK) {
                /* either error wasn't elapsed overflow, or the above
                 * fix didn't help */
                rwioPrintLastErr(out_ios, rwio_rv, &logMsg);
            }
        }
    } while (rwRead(in_ios, &rwrec));

    /* log about truncating the elapsed time field */
    if (long_duration > 0) {
        logMsg("Truncated elapsed time for %u records in file %s",
               long_duration, rwGetFileName(in_ios));
    }

    /* close output */
    if (out_ios) {
        /* close the output file.  If we encountered an error
         * previously, leave rwio_rv alone, else set it to the status
         * of the close. */
        int close_rv = rwioDestroy(&out_ios);
        if (close_rv) {
            rwioPrintLastErr(out_ios, close_rv, &logMsg);
            /* make this RV the exit status if we were o.k. to here */
            if (rwio_rv == LIBRW_OK) {
                rwio_rv = close_rv;
            }
        }
    }

    /* close input */
    if (in_ios) {
        int close_rv = rwioDestroy(&in_ios);
        if (close_rv) {
            rwioPrintLastErr(in_ios, close_rv, &logMsg);
        }
    }

    /* everything worked; set output status to success */
    if (rwio_rv == LIBRW_OK) {
        ret_val = 0;
    }

  END:
    if (out_ios) {
        (void)rwioDestroy(&out_ios);
    }
    if (in_ios) {
        (void)rwioDestroy(&in_ios);
    }

    return ret_val;
}


/*
** Local variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
